# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import time

import collections
import logging

from address_book_app.address_book import _errors
from address_book_app.address_book import data
from autopilot import introspection

from autopilot.exceptions import StateNotFoundError
from ubuntu_system_tests.helpers.ubuntu_keyboard import get_keyboard
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._textfield import TextField  # NOQA
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._toolbar import Toolbar
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import (
    CustomProxyObjectBase,
    UiToolsException
)

from ubuntu_system_tests.helpers.autopilot import validate_dbus_object

APP_PATH_ROOT = b'AddressBookApp'
logger = logging.getLogger(__name__)

_TEXT_FIELD_OBJECT_NAMES = {
    'first_name': 'firstName',
    'last_name': 'lastName',
    'street': 'streetAddress_{}',
    'locality': 'localityAddress_{}',
    'region': 'regionAddress_{}',
    'postal_code': 'postcodeAddress_{}',
    'country': 'countryAddress_{}'
}


class PageWithHeader(CustomProxyObjectBase):

    def get_header(self, main_window_name='MainWindow'):
        """Return the Header custom proxy object of the Page."""
        return self.get_root_instance().select_single(
            main_window_name).get_header()


class ContactEditorPage(PageWithHeader):
    """Custom proxy object for the Contact Editor."""

    _DETAIL_ALIAS = {
        'phones': 'Phone',
        'emails': 'Email',
        'ims': 'Social',
        'addresses': 'Address',
        'professionalDetails': 'Professional Details'
    }

    def add_field(self, detail_name):
        """Create a new field into the edit contact form.

        :param detail_name: The detail field name

        """

        add_field_button = self.select_single(
            'ComboButtonAddField', objectName='addNewFieldButton')
        add_field_button.swipe_into_view()

        self.pointing_device.click_object(add_field_button)
        add_field_button.height.wait_for(add_field_button.expandedHeight)
        self.wait_to_stop_moving()

        options_list = add_field_button.select_single(
            "QQuickListView",
            objectName="listViewOptions")
        new_field_item = options_list.select_single(
            "Standard",
            objectName=self._DETAIL_ALIAS[detail_name])
        new_field_item.swipe_into_view()

        self.pointing_device.click_object(new_field_item)
        add_field_button.height.wait_for(add_field_button.collapsedHeight)

    def fill_form(self, contact_information):
        """Fill the edit contact form.
        :param contact_information: Values of the contact to fill the form.
        :type contact_information: data object with the attributes first_name,
            last_name, phones, emails, social_aliases, addresses and
             professional_details
        """

        self._fill_full_name(contact_information)

        groups = collections.OrderedDict()
        groups['phones'] = contact_information.phones
        groups['emails'] = contact_information.emails
        groups['ims'] = contact_information.social_aliases
        groups['addresses'] = contact_information.addresses
        groups['professionalDetails'] = (
            contact_information.professional_details)

        for key, information in groups.items():
            if information:
                self._fill_detail_group(
                    object_name=key, details=information)

    def _fill_full_name(self, contact_information):
        if contact_information.first_name is not None:
            self._fill_first_name(contact_information.first_name)
        if contact_information.last_name is not None:
            self._fill_last_name(contact_information.last_name)

    def _fill_first_name(self, first_name):
        text_field = _get_text_field(self, 'first_name')
        text_field.write(first_name)

    def _fill_last_name(self, last_name):
        text_field = _get_text_field(self, 'last_name')
        text_field.write(last_name)

    def _fill_detail_group(self, object_name, details):
        editor = self.select_single(
            ContactDetailGroupWithTypeEditor, objectName=object_name)
        editor.fill(self, details)

    def _get_form_values(self):
        first_name = _get_text_field(self, 'first_name').text
        last_name = _get_text_field(self, 'last_name').text
        phones = self._get_values_from_detail_group(object_name='phones')
        emails = self._get_values_from_detail_group(object_name='emails')
        social_aliases = self._get_values_from_detail_group(object_name='ims')
        addresses = self._get_values_from_detail_group(object_name='addresses')
        professional_details = self._get_values_from_detail_group(
            object_name='professionalDetails')

        return data.Contact(
            first_name=first_name, last_name=last_name, phones=phones,
            emails=emails, social_aliases=social_aliases, addresses=addresses,
            professional_details=professional_details)

    def _get_values_from_detail_group(self, object_name):
        editor = self.select_single(
            ContactDetailGroupWithTypeEditor, objectName=object_name)
        return editor.get_values(object_name)

    def wait_to_stop_moving(self):
        flickable = self.select_single(
            'QQuickFlickable', objectName='scrollArea')
        flickable.flicking.wait_for(False)

    def wait_get_focus(self, section_name):
        editor = self.select_single(
            ContactDetailGroupWithTypeEditor, objectName=section_name)
        editor.activeFocus.wait_for(True)


class ABContactEditorPage(ContactEditorPage):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ABContactEditorPage',
            objectName='contactEditorPage')


class ABContactListPage(PageWithHeader):

    def open_contact(self, index):
        """Open the page with the contact information.

        :param index: The index of the contact to open.
        :return: The page with the contact information.

        """
        contact_delegate = self._get_contact_delegate(index)
        self.pointing_device.click_object(contact_delegate)
        # WORKAROUND: give some time to the view became available
        time.sleep(5.0)
        return self.get_root_instance().select_single(
            ABContactViewPage, objectName='contactViewPage', active=True)

    def _get_contact_delegate(self, index):
        contact_delegates = self._get_sorted_contact_delegates()
        return contact_delegates[index]

    def select_contacts(self, indices):
        """ Select contacts corresponding to the list of index in indices

        :param indices: List of integers

        """
        self._deselect_all()
        if indices > 0:
            view = self._get_list_view()
            if not view.isInSelectionMode:
                self._start_selection(indices[0])
                indices = indices[1:]

            for index in indices:
                contact = self._get_contact_delegate(index)
                self.pointing_device.click_object(contact)

    def _deselect_all(self):
        """Deselect all contacts."""
        view = self._get_list_view()
        if view.isInSelectionMode:
            contacts = self.select_many('ContactDelegate', visible=True)
            for contact in contacts:
                if contact.selected:
                    logger.info('Deselect {}.'.format(contact.objectName))
                    self.pointing_device.click_object(contact)
        else:
            logger.debug('The page is not in selection mode.')

    def _start_selection(self, index):
        # TODO change this for click_object once the press duration
        # parameter is added. See http://pad.lv/1268782
        contact = self._get_contact_delegate(index)
        self.pointing_device.move_to_object(contact)
        self.pointing_device.press()
        time.sleep(2.0)
        self.pointing_device.release()
        view = self._get_list_view()
        view.isInSelectionMode.wait_for(True)

    def _get_list_view(self):
        return self.wait_select_single(
            'ContactListView', objectName='contactListView')

    def delete_selected_contacts(self, main_window):
        main_window.delete()
        main_window.wait_bottom_edge(False)
        dialog = self.get_root_instance().wait_select_single(
            RemoveContactsDialog, objectName='removeContactsDialog')
        dialog.confirm_removal()

    def get_contacts(self):
        """Return a list with the names of the contacts."""
        contact_delegates = self._get_sorted_contact_delegates()
        name_labels = [
            delegate.select_single('UCLabel', objectName='nameLabel') for
            delegate in contact_delegates
        ]
        return [label.text for label in name_labels]

    def get_button(self, buttonName):
        try:
            return self.get_header()._get_action_button(buttonName)
        except UiToolsException:
            return None

    def is_import_from_sim_button_visible(self):
        import_from_sim_button = self.select_single(
            'ContactListButtonDelegate',
            objectName='contactListView.importFromSimCardButton')
        return import_from_sim_button.visible

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ABContactListPage')

    def expand_contact(self, index):
        """Click the contact to show more information.

        :param index: The index of the contact to expand.

        """
        contact_delegate = self._get_contact_delegate(index)
        self.pointing_device.click_object(contact_delegate)
        contact_delegate.wait_until_destroyed()

    def open_contact_by_name(self, display_name):
        """
        Open the contact specified by display_name
        :param display_name: Name of contact to open
        :return: Page with contact information
        """
        return self.open_contact(self.get_contacts().index(display_name))

    def select_contact_by_name(self, display_name):
        """Select a contact in list view by name."""
        self.select_contacts([self.get_contacts().index(display_name)])

    def confirm_selection(self):
        """Press confirm selection button in picker mode."""
        self.pointing_device.click_object(self._get_share_button())

    def cancel_selection(self):
        """Press cancel selection button in picker mode."""
        self.pointing_device.click_object(self._get_cancel_button())

    def _get_cancel_button(self):
        return self.select_single(objectName='cancel_button')

    def _get_share_button(self):
        return self.select_single(objectName='share_button')

    def _get_sorted_contact_delegates(self):
        """Over-ride base helper to use wait_select_many."""
        contact_delegates = self.wait_select_many(
            'ContactDelegate', visible=True)
        return sorted(
            contact_delegates, key=lambda delegate: delegate.globalRect.y)


class ABContactViewPage(PageWithHeader):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ABContactViewPage')

    def get_contact_mobile_phone_number(self):
        """Return the first mobile phone listed for the contact"""
        mobile_fields = self.select_many('BasicFieldView',
                                         typeLabel='Mobile')
        if not mobile_fields:
            raise Exception('Mobile fields not found')

        return mobile_fields[0].\
            select_single('UCLabel', objectName='label_phoneNumber_0.0').text

    def _get_contact_phone_number(self, phone_number):
        """
        Find the contact using phone number.

        :param phone_number: Phone number for required contact
        :return: ContactDetailPhoneNumberView matching search requirements
                 None if match is not found

        """
        contacts = self.select_many(ContactDetailPhoneNumberView, visible=True)
        for contact in contacts:
            try:
                # try to find matching phone number
                contact.select_single(
                    'UCLabel', objectName='label_phoneNumber_0.0',
                    text=phone_number, visible=True)
            except StateNotFoundError:
                # move onto the next one
                pass
            else:
                # found the required contact
                return contact

        # No matching contact found
        raise Exception('Contact with phone number {n} not found'.format(
            n=phone_number))

    def call(self, phone_number):
        """
        Press the call button for the required phone number
        :param phone_number: The specified phone number to search for
        """
        phone_number = self._get_contact_phone_number(phone_number)
        phone_number.click_call_button()

    def message(self, phone_number):
        """
        Press the message button for the required phone number
        :param phone_number: The specified phone number to search for
        """
        phone_number = self._get_contact_phone_number(phone_number)
        phone_number.click_message_button()


class ContactDetailPhoneNumberView(CustomProxyObjectBase):
    """Class to represent a phone number view with call and message buttons"""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ContactDetailPhoneNumberView')

    def _get_call_button(self):
        """Return the call button ActionButton object"""
        return self.select_single('ActionButton', iconName='call-start')

    def _get_message_button(self):
        """Return the message button ActionButton object"""
        return self.select_single('ActionButton', iconName='message')

    def click_call_button(self):
        """Click on the call button for this phone number"""
        self.pointing_device.click_object(self._get_call_button())

    def click_message_button(self):
        """Click on the message button for this phone number"""
        self.pointing_device.click_object(self._get_message_button())


class ContactViewPage(PageWithHeader):
    """Autopilot helper for the ContactView page."""
    pass


class AddressBookApp(CustomProxyObjectBase):

    """Autopilot custom proxy object for the address book app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        name = introspection.get_classname_from_path(path)
        return (name == b'AddressBookApp' and
                state['applicationName'][1] == 'AddressBookApp')

    @property
    def main_window(self):
        return self.select_single(objectName='addressBookAppMainWindow')


class AddressBookAppMainWindow(CustomProxyObjectBase):

    """An emulator class that makes it easy to interact with the app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        name = introspection.get_classname_from_path(path)
        return (name == b'MainWindow' and
                state['objectName'][1] == 'addressBookAppMainWindow')

    def get_main_title(self):
        """ Retrieve the app title """
        return self.wait_select_single('Page').title

    def get_contact_list_page(self):
        """Call base class helper and wait until page is ready."""
        page = self.wait_select_single(
            ABContactListPage,
            objectName='contactListPage',
            pickMode=False, active=True)
        page.wait_select_single(
            'ContactSimpleListView', visible=True, moving=False, loading=False)
        page.wait_select_single(objectName='bottomEdge', visible=True)
        return page

    def get_contact_edit_page(self):
        """We can have two contact editor page because of bottom edge page
        but we will return only the active one
        """
        return self.wait_select_single(ABContactEditorPage,
                                       objectName="contactEditorPage",
                                       active=True)

    def get_contact_view_page(self):
        return self.wait_select_single(ABContactViewPage,
                                       objectName="contactViewPage",
                                       active=True)

    def get_contact_list_pick_page(self):
        try:
            return self.select_single(ABContactListPage,
                                      objectName='contactListPage',
                                      pickMode=True)
        except StateNotFoundError:
            return None

    def get_share_page(self):
        return self.wait_select_single("ContactSharePage",
                                       objectName="contactSharePage",
                                       active=True)

    def start_import_contacts(self):
        view = self.get_contact_list_view()
        if view.count > 0:
            self.click_action_button("importFromSimHeaderButton")
        else:
            import_buttom = self.select_single(
                'ContactListButtonDelegate',
                objectName='contactListView.importFromSimCardButton')
            self.pointing_device.click_object(import_buttom)

        return self.wait_select_single(SIMCardImportPage,
                                       objectName="simCardImportPage",
                                       active=True)

    def get_contact_list_view(self):
        """
        Returns a ContactListView object for the current window
        """
        return self.wait_select_single("ContactListView",
                                       objectName="contactListView")

    def get_action(self, action_name):
        actionbars = self.select_many('ActionBar',
                                      objectName='headerActionBar')
        for actionbar in actionbars:
            object_name = action_name + "_action_button"
            try:
                button = actionbar.select_single(objectName=object_name)
                if button:
                    return button
            except introspection.dbus.StateNotFoundError:
                continue
        return None

    def click_action_button(self, action_name):
        actions = self.select_many(objectName='%s_button' % action_name)
        for action in actions:
            if action.enabled:
                self.pointing_device.click_object(action)
                return

        raise StateNotFoundError(action_name)

    def open_header(self):
        header = self.get_header()
        if header.y != 0:
            edit_page = self.get_contact_edit_page()
            flickable = edit_page.wait_select_single(
                "QQuickFlickable",
                objectName="scrollArea")

            while header.y != 0:
                globalRect = flickable.globalRect
                start_x = globalRect.x + (globalRect.width * 0.5)
                start_y = globalRect.y + (flickable.height * 0.1)
                stop_y = start_y + (flickable.height * 0.1)

                self.pointing_device.drag(
                    start_x, start_y, start_x, stop_y, rate=5)
                # wait flicking stops to move to the next field
                flickable.flicking.wait_for(False)

        return header

    def wait_bottom_edge(self, opened):
        """wait bottom edge to fully appear"""
        mainStack = self.wait_select_single(objectName='mainStack')
        mainStack.bottomEdgeOpened.wait_for(opened)

    def reveal_bottom_edge_page(self):
        flickable = self.get_contact_list_view()

        globalRect = flickable.globalRect
        start_x = globalRect.x + (globalRect.width * 0.5)
        start_y = globalRect.y + (flickable.height * 0.98)
        stop_y = globalRect.y + (flickable.height * 0.4)

        self.pointing_device.drag(
            start_x, start_y, start_x, stop_y, rate=5)

        # wait bottom edge to fully appear
        self.wait_bottom_edge(True)

    def cancel(self):
        """Press the 'Cancel' button"""
        buttons = self.select_many(objectName='customBackButton')
        for button in buttons:
            if button.enabled and button.visible:
                self.pointing_device.click_object(button)
                return

        self.click_action_button('cancel')

    def save(self):
        """Press the 'Save' button"""
        self.click_action_button("save")

    def edit(self):
        """Press the 'Edit' button"""
        self.click_action_button("edit")

    def delete(self):
        """Press the 'Delete' button"""
        self.click_action_button("delete")

    def confirm_import(self):
        """Press the 'confirm' button"""
        self.click_action_button("confirmImport")

    def get_toolbar(self):
        """Override base class so we get our expected Toolbar subclass."""
        return self.select_single(Toolbar)

    def go_to_add_contact(self):
        """Press the 'Add' button and return the contact editor page"""
        self.reveal_bottom_edge_page()

        return self.get_contact_edit_page()


class SIMCardImportPage(PageWithHeader):
    """Autopilot helper for the SIMCardImportPage page."""

    def _get_sorted_contact_delegates(self):
        # FIXME this returns only the contact delegates that are loaded in
        # memory. The list might be big, so not all delegates might be in
        # memory at the same time.
        contact_delegates = self.select_many('ContactDelegate', visible=True)
        return sorted(
            contact_delegates, key=lambda delegate: delegate.globalRect.y)

    def _get_contact_delegate(self, index):
        contact_delegates = self._get_sorted_contact_delegates()
        return contact_delegates[index]

    def _start_selection(self, index):
        # TODO change this for click_object once the press duration
        # parameter is added. See http://pad.lv/1268782
        contact = self._get_contact_delegate(index)
        self.pointing_device.move_to_object(contact)
        self.pointing_device.press()
        time.sleep(2.0)
        self.pointing_device.release()
        view = self._get_list_view()
        view.isInSelectionMode.wait_for(True)

    def _get_list_view(self):
        return self.wait_select_single(
            'ContactListView', objectName='contactListViewFromSimCard')

    def _deselect_all(self):
        """Deselect all contacts."""
        view = self._get_list_view()
        if view.isInSelectionMode:
            contacts = self.select_many('ContactDelegate',
                                        visible=True,
                                        selected=True)
            for contact in contacts:
                logger.info('Deselect {}.'.format(contact.objectName))
                self.pointing_device.click_object(contact)
        else:
            logger.debug('The page is not in selection mode.')

    def get_contacts(self):
        """Return a list with the names of the contacts."""
        contact_delegates = self._get_sorted_contact_delegates()
        return [delegate.select_single('UCLabel',
                                       objectName='nameLabel').text
                for delegate in contact_delegates]

    def select_contacts(self, indices):
        """ Select contacts corresponding to the list of index in indices

        :param indices: List of integers

        """
        contacts = []
        self._deselect_all()
        if len(indices) > 0:
            view = self._get_list_view()
            if not view.isInSelectionMode:
                self._start_selection(indices[0])
                indices = indices[1:]

            for index in indices:
                contact = self._get_contact_delegate(index)
                self.pointing_device.click_object(contact)
                contacts.append(contact.select_single(
                    'UCLabel', objectName='nameLabel').text)

        return contacts


class ContactDetailGroupWithTypeEditor(CustomProxyObjectBase):
    """Custom proxy object for the ContactDetailGroupWithTypeEditor widget."""

    _DETAIL_EDITORS = {
        'phones': 'base_phoneNumber_{}',
        'emails': 'base_email_{}',
        'ims': 'base_onlineAccount_{}',
        'addresses': 'base_address_{}',
        # FIXME fix the unknown. --elopio - 2014-03-01
        'professionalDetails': 'base_unknown_{}'
    }

    def fill(self, editor, details):
        """Fill a contact detail group."""
        for index, detail in enumerate(details):
            if self.detailsCount <= index:
                editor.add_field(self.objectName)
            self._fill_detail(index, detail)

    def _fill_detail(self, index, detail):
        detail_editor = self._get_detail_editor_by_index(index)
        detail_editor.fill(field=self.objectName, index=index, detail=detail)

    def _get_detail_editor_by_index(self, index):
        object_name = self._get_contact_detail_editor_object_name(index)
        return self.select_single(
            ContactDetailWithTypeEditor, objectName=object_name)

    def _get_contact_detail_editor_object_name(self, index):
        return self._DETAIL_EDITORS[self.objectName].format(index)

    def _add_detail(self):
        # TODO --elopio - 2014-03-01
        raise NotImplementedError('Add extra details not yet implemented.')

    def get_values(self, object_name):
        """Return the values of a contact detail group."""
        values = []
        for index in range(self.detailsCount):
            detail_editor = self._get_detail_editor_by_index(index)
            value = detail_editor.get_values(field=object_name, index=index)
            if (value):
                values.append(value)

        return values


class ContactDetailWithTypeEditor(CustomProxyObjectBase):
    """Custom proxy object for the ContactDetailWithTypeEditor widget."""

    VALID_FIELDS = ['phones', 'emails', 'ims', 'addresses',
                    'professionalDetails']

    def fill(self, field, index, detail):
        self._fill_value(field, index, detail)
        self._select_type(detail)

    def _select_type(self, detail):
        type_index = detail.TYPES.index(detail.type)
        value_selector = self.select_single('ValueSelector')

        while(value_selector.currentIndex != type_index):
            get_keyboard().press_and_release("Shift+Right")
            time.sleep(0.1)

    def _get_selected_type_index(self):
        value_selector = self.select_single('ValueSelector')
        return value_selector.currentIndex

    def _fill_value(self, field, index, detail):
        single_values = {
            'phones': 'number',
            'emails': 'address',
            'ims': 'alias'
        }
        if field in single_values:
            self._fill_single_field(getattr(detail, single_values[field]))
        elif field == 'addresses':
            self._fill_address(index, detail)
        elif field == 'professionalDetails':
            self._fill_professional_details(index, detail)
        else:
            raise _errors.AddressBookAppError(
                'Unknown field: {}.'.format(field))

    def _fill_single_field(self, value):
        text_field = self.select_single(TextInputDetail)
        self._make_field_visible_and_write(text_field, value)

    def _make_field_visible_and_write(self, text_field, value):
        text_field.swipe_into_view()
        text_field.write(value)

    def _fill_address(self, index, address):
        fields = collections.OrderedDict()
        fields['street'] = address.street
        fields['locality'] = address.locality
        fields['region'] = address.region
        fields['postal_code'] = address.postal_code
        fields['country'] = address.country
        for key, value in fields.items():
            text_field = _get_text_field(self, key, index)
            self._make_field_visible_and_write(text_field, value)

    def _fill_professional_details(self, index, address):
        # TODO --elopio - 2014-03-01
        raise NotImplementedError('Not yet implemented.')

    def get_values(self, field, index):

        ContactDetailWithTypeEditor._assert_valid_field(field)

        return {
            'phones': self.__get_phones,
            'emails': self.__get_emails,
            'ims': self.__get_ims,
            'addresses': self.__get_addresses,
            'professionalDetails': self.__get_professional_details
        }.get(field)(index=index)

    def __get_phones(self, **kwargs):
        return data.Phone(
            type_=data.Phone.TYPES[self._get_selected_type_index()],
            number=self._get_single_field_value())

    def __get_emails(self, **kwargs):
        return data.Email(
            type_=data.Email.TYPES[self._get_selected_type_index()],
            address=self._get_single_field_value())

    def __get_ims(self, **kwargs):
        data.SocialAlias(
            type_=data.SocialAlias.TYPES[self._get_selected_type_index()],
            alias=self._get_single_field_value())

    def __get_addresses(self, **kwargs):
        return self._get_address_value(kwargs.get('index'))

    def __get_professional_details(self, **kwargs):
        raise NotImplementedError('The field professional '
                                  'details is not implemented yet')

    @staticmethod
    def _assert_valid_field(field):
        if field not in ContactDetailWithTypeEditor.VALID_FIELDS:
            raise _errors.AddressBookAppError(
                'Unknown field: {}.'.format(field)
            )

    def _get_single_field_value(self):
        return self.select_single(TextInputDetail).text

    def _get_address_value(self, index):
        street = _get_text_field(self, 'street', index).text
        locality = _get_text_field(self, 'locality', index).text
        region = _get_text_field(self, 'region', index).text
        postal_code = _get_text_field(self, 'postal_code', index).text
        country = _get_text_field(self, 'country', index).text
        return data.Address(
            type_=data.Address.TYPES[self._get_selected_type_index()],
            street=street, locality=locality, region=region,
            postal_code=postal_code, country=country)


def _get_text_field(parent, field, index=None):
    if field not in _TEXT_FIELD_OBJECT_NAMES:
        raise _errors.AddressBookAppError('Unknown field: {}.'.format(field))

    object_name = _TEXT_FIELD_OBJECT_NAMES[field]
    if index is not None:
        object_name = object_name.format(index)
    return parent.select_single(TextInputDetail, objectName=object_name)


class TextInputDetail(TextField):
    """Custom proxy object for the Text Input Detail field."""


class AddressBookAppError(UiToolsException):
    """Exception raised when there is an error with the emulator."""


class RemoveContactsDialog(CustomProxyObjectBase):

    """Autopilot helper for the Remove Contacts dialog."""

    def confirm_removal(self):
        button = self.select_single(
            'Button', objectName='removeContactsDialog.Yes')
        self.pointing_device.click_object(button)
        self.wait_until_destroyed()
