# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from abc import abstractmethod, ABCMeta

import subprocess
from autopilot.exceptions import ProcessSearchError
from psutil import NoSuchProcess
from retrying import retry

from ubuntu_system_tests.helpers import autopilot
from ubuntu_system_tests.helpers import processes
from ubuntu_system_tests.helpers.processes import ensure_application_closed
from ubuntu_system_tests.helpers.scopes.apps import (
    launch_application_from_apps_scope)
from ubuntu_system_tests.helpers.unity8 import (
    launch_application_from_launcher)


class Application:
    __metaclass__ = ABCMeta

    def __init__(self, app_name, app_click, app_id=None):
        self.app_name = app_name
        self.app_click = app_click
        self.app_id = app_id
        self.proxy_object = None

    def launch_from_scope(self, return_proxy=True):
        """
        Launch app from apps scope
        """
        launch_application_from_apps_scope(self.app_name)
        if return_proxy:
            return self.get_proxy_object()

    def launch_from_launcher(self, return_proxy=True):
        """Drag out the launcher and tap the application icon to start it."""
        launch_application_from_launcher(self.app_id)
        if return_proxy:
            return self.get_proxy_object()

    def close(self):
        pass

    def stop(self):
        ensure_application_closed(self.app_id)

    @abstractmethod
    def get_pid(self):
        pass

    @abstractmethod
    def is_running(self):
        pass

    @retry(stop_max_delay=30000,
           wait_exponential_multiplier=1000,
           wait_exponential_max=10000,
           retry_on_exception=lambda exception: (
               isinstance(exception, ProcessSearchError) or
               isinstance(exception, processes.PidNotFoundError) or
               isinstance(exception, NoSuchProcess)))
    def get_proxy_object(self):
        if self.proxy_object is None:
            self.proxy_object = autopilot.get_proxy_object(
                self.get_pid())
        return self.proxy_object

    @abstractmethod
    def get_main_view(self):
        pass

    def is_installed(self, _id=None):
        try:
            _id = _id or self.app_id
            apps = subprocess.check_output(
                ['ubuntu-app-launch-appids'],
                universal_newlines=True,
                stderr=subprocess.DEVNULL).splitlines()
            return any(app.startswith(_id) for app in apps)
        except subprocess.CalledProcessError:
            return False


class WebApplication(Application):

    def get_proxy_object(self):
        """To be implemented"""
        pass

    @abstractmethod
    def get_main_window(self, proxy_object):
        pass

    @abstractmethod
    def get_main_view(self):
        pass


class QmlApplication(Application):

    def __init__(self, app_name, app_click, app_process_name, app_id=None):
        self.app_process_name = app_process_name
        super().__init__(app_name, app_click, app_id)

    def get_pid(self):
        """Return the QML application process id"""
        return processes.get_process_id(self.app_process_name)

    def is_running(self):
        """Return bool representing if the app is running."""
        return processes.is_process_running(self.app_process_name)

    @abstractmethod
    def get_main_view(self):
        pass


class QmlSceneApplication(Application):

    def __init__(self, app_name, app_click, qml_file, app_id=None):
        self.qml_file = qml_file
        super().__init__(app_name, app_click, app_id)

    def get_pid(self):
        """Return the QML Scene application process id"""
        return processes.get_qmlscene_process_id(self.qml_file)

    @abstractmethod
    def get_main_view(self):
        pass
