# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from autopilot import introspection
from autopilot.exceptions import ProcessSearchError
from autopilot.introspection import _xpathselect as xpathselect
from retrying import retry

from ubuntu_system_tests.helpers.timeout import timeout

from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA


def validate_dbus_object(path, state, root_name, class_name, **kwargs):
    """Return True if dbus object matches required parameters.
    :param path: Path of object.
    :param state: Properties of object.
    :param root_name: The required root name for the object.
    :param class_name: The required class name for the object.
    :param kwargs: Any other properties to match.
    :return: True if object matches required path root, name and properties,
             False otherwise.
    """
    if (xpathselect.get_path_root(path) == root_name and
            introspection.get_classname_from_path(path) == class_name):
        for key in kwargs.keys():
            if state[key][1] != kwargs[key]:
                return False
        return True
    return False


def get_patched_registry(class_object, class_name):
    """
    Return a patched copy of autopilot's object registry.

    Using the specified class name and class object reference patch autopilot's
    object registry. This can be used in conjunction with autopilot's
    patch_registry() function to ensure that the correct CPO is returned for
    a specified type.

    :param class_object: Object type to create.
    :param class_name: The name of the class to patch.

    """
    registry = introspection._object_registry._object_registry.copy()
    for base_uuid in registry.keys():
        if class_name in registry[base_uuid]:
            registry[base_uuid][class_name] = class_object
    return registry


def _patched_select(wait, root, class_object, class_name=None, **kwargs):
    """
    Patch the autopilot object registry and select an object from root.

    This method is used to avoid autopilot CPO errors by patching the autopilot
    object registry before calling the required select method on the specified
    root object.

    :param wait: Whether to call wait_select_single or select_single.
    :param root: Root introspection object.
    :param class_object: Instance of object to patch the registry with.
    :param class_name: Name of registry entry to patch. If None then the class
                       name of class_object will be used.
    :return: Intropsection object of type class_object.

    """
    if not class_name:
        class_name = class_object.__class__.__name__
    registry = get_patched_registry(class_object, class_name)
    with introspection._object_registry.patch_registry(registry):
        if wait:
            return root.wait_select_single(class_name, **kwargs)
        else:
            return root.select_single(class_name, **kwargs)


def patched_select_single(root, class_object, class_name=None, **kwargs):
    """
    Perform select_single on root object using patched autopilot registry.

    :param root: Root introspection object.
    :param class_object: Instance of object to patch the registry with.
    :param class_name: Name of registry entry to patch. If None then the class
                       name of class_object will be used.
    :return: Intropsection object of type class_object.

    """
    return _patched_select(False, root, class_object, class_name, **kwargs)


def patched_wait_select_single(root, class_object, class_name=None, **kwargs):
    """
    Perform wait_select_single on root object using patched autopilot registry.

    :param root: Root introspection object.
    :param class_object: Instance of object to patch the registry with.
    :param class_name: Name of registry entry to patch. If None then the class
                       name of class_object will be used.
    :return: Intropsection object of type class_object.

    """
    return _patched_select(True, root, class_object, class_name, **kwargs)


@timeout(60)
@retry(stop_max_attempt_number=5, wait_fixed=3000,
       retry_on_exception=lambda exception: isinstance(
           exception, ProcessSearchError))
def get_proxy_object(pid, emulator_base='default'):
    """ Return the proxy object for the given pid
    :param pid: The PID of the application to search for.
    :param emulator_base: The custom emulator to use when creating the
        resulting proxy object.
    """
    assert pid, 'Please specify required process id.'
    if emulator_base == 'default':
        emulator_base = CustomProxyObjectBase
    return introspection.get_proxy_object_for_existing_process(
        pid=pid,
        emulator_base=emulator_base
    )
