# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from retrying import retry
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._mainview import MainView as Mv  # NOQA
from ubuntu_system_tests.helpers.autopilot import validate_dbus_object

APP_PATH_ROOT = b'comubuntucalculator'


class CalculatorApp(CustomProxyObjectBase):
    """Autopilot helper object for clock app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, APP_PATH_ROOT)

    @property
    def main_view(self):
        return self.wait_select_single(MainView)


class MainView(Mv):
    """Calculator MainView Autopilot emulator."""

    BUTTONS = {'delete': 'deleteButton', '*': 'multiplyButton',
               '/': 'divideButton', '.': 'pointButton',
               '=': 'equalsButton', '-': 'minusButton', '+': 'plusButton',
               '0': 'zeroButton', '1': 'oneButton', '2': 'twoButton',
               '3': 'threeButton', '4': 'fourButton', '5': 'fiveButton',
               '6': 'sixButton', '7': 'sevenButton', '8': 'eightButton',
               '9': 'nineButton', 'bracket': 'universalBracketButton',
               'square': 'squareButton', 'cube': 'cubeButton',
               'power': 'powerButton', 'log': 'logarithmButton',
               'e': 'eNumberButton', '!': 'factorialNumberButton',
               'i': 'iButton', 'sin': 'sinusButton', 'cos': 'cosButton'}

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MainView', objectName='calculator')

    def __init__(self, *args):
        super(MainView, self).__init__(*args)
        self.visible.wait_for(True)

    @retry(stop_max_attempt_number=10,
           retry_on_exception=lambda exception: isinstance(
               exception, KeyNotPressedError))
    def press(self, button):
        """Press the requested key on the calculator."""
        logging.info("Pressing button: {}".format(button))
        button = self.wait_select_single('KeyboardButton',
                                         objectName=MainView.BUTTONS[button],
                                         visible=True)
        self.pointing_device.click_object(button)
        result = self.wait_select_single('TextField',
                                         objectName='textInputField',
                                         visible=True
                                         ).displayText
        if button.text not in result or result == '':
            raise KeyNotPressedError

    @retry(stop_max_attempt_number=5,
           retry_on_exception=lambda exception: isinstance(
               exception, EmptyResultError))
    def get_result(self):
        """Return the current calculation result visible on screen."""
        result = self.wait_select_single('TextField',
                                         objectName='textInputField'
                                         ).displayText
        if result == '':
            raise EmptyResultError
        return result

    def enter_operation(self, operation):
        for key in operation:
            self.press(key)

    def calculate(self):
        """Make a calculation of the request."""
        button = self.wait_select_single('KeyboardButton',
                                         objectName=MainView.BUTTONS["="])
        self.pointing_device.click_object(button)
        return self.get_result()


class KeyNotPressedError(RuntimeError):
    """"Key not pressed error."""
    pass


class EmptyResultError(RuntimeError):
    """"Empty result error."""
    pass
