# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import autopilot
import logging
import time
from random import randint
from autopilot import logging as autopilot_logging
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA
from camera_app.emulators.main_window import MainWindow
from retrying import retry

from ubuntu_system_tests.helpers.autopilot.list import KEY_ORDER_BY_X
from ubuntu_system_tests.helpers.autopilot import validate_dbus_object
from ubuntu_system_tests.helpers import wait_until

logger = logging.getLogger(__name__)

APP_PATH_ROOT = b'comubuntucamera'


class MainWindow(MainWindow):
    """Class to extend camera applications MainWindow controls."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MainWindow')

    def get_main_title(self):
        """ Retrieve the app title """
        return self.app.wait_select_single(objectName='main').title

    @autopilot_logging.log_action(logger.info)
    def take_photo(self, photoroll_hint=True):
        """ Check camera is ready and take a photo, return photoroll hint if
        needed.
        :param photoroll_hint: there are scenarios when the hint is not
        displayed and it should not be retrieved
        :return: the photoroll hint if it was requested, None otherwise
        """
        self.ensure_camera_ready()
        self.app.pointing_device.click_object(
            self.get_exposure_button())
        if photoroll_hint:
            return self._get_photoroll_hint()

    def take_photo_and_select(self):
        """Capture a photo and tap the validate button."""
        validate_button = self._take_photo_for_export()
        self.app.pointing_device.click_object(validate_button)

    def _take_photo_for_export(self):
        """Capture a photo and return the validate button."""
        self.ensure_camera_ready()
        exposure_button = self.get_exposure_button()
        self.app.pointing_device.click_object(exposure_button)
        return self._get_validate_button()

    def _get_validate_button(self):
        """Returns the "validate" photo button"""
        return self.app.wait_select_single(
            objectName='validateButton', visible=True)

    def select_photo_from_roll(self):
        """Picks the first visible photo in camera roll."""
        validate_button = self._select_first_photo_in_roll()
        self.app.pointing_device.click_object(validate_button)

    def _select_first_photo_in_roll(self):
        """Long presses the first photo in the roll to select."""
        photo = self._get_photo_thumbnail_by_index(0)
        # Do a long press to select.
        self.app.pointing_device.click_object(photo, press_duration=1)
        return self._get_toolbar_validate_button()

    def _get_photo_thumbnail_by_index(self, index):
        """Returns photo thumbnail of the given index."""
        gallery = self.get_gallery()
        return gallery.wait_select_single(
            objectName="mediaItem{}".format(index))

    def _get_toolbar_validate_button(self):
        """Returns the validate/pick button on the toolbar."""
        return self.app.wait_select_single(
            objectName='validationButton', visible=True)

    def ensure_camera_ready(self):
        """Ensure that the camera is ready and exposure button displayed."""
        self.get_exposure_button().enabled.wait_for(True)

    def open_bottom_edge(self):
        """Swipe up to display the bottom edge."""
        self.ensure_camera_ready()
        bottom_edge = self.get_bottom_edge()
        bottom_edge.open()
        self._wait_for_fastblur(True)
        return bottom_edge

    def _get_fast_blur(self):
        """Get the fastblur object."""
        return self.app.wait_select_single('FastBlur')

    def _wait_for_fastblur(self, expected):
        """Wait for fastblur visibility to be required state."""
        self._get_fast_blur().visible.wait_for(expected)

    def _get_photoroll_hint(self):
        """Return the photoroll hint."""
        self.ensure_camera_ready()
        try:
            photoroll = self.app.wait_select_single(
                PhotoRollHint, visible=True)
        except autopilot.exceptions.StateNotFoundError:
            raise Exception('Photo roll not displayed after taking photo')
        return photoroll

    def get_flash_icon_name(self):
        """Get the icon name for the flash button."""
        return self.get_flash_button().get_properties()['iconName']

    def set_flash_option(self, value):
        """Set the flash option to specified value."""
        self._press_flash_option()
        self._press_option(value)

    @autopilot_logging.log_action(logger.info)
    def _press_flash_option(self):
        """Press the flash option button."""
        self.app.pointing_device.click_object(self.get_flash_button())

    def set_encoding_quality_option(self, value):
        """Set the quality option to specified value."""
        self._press_encoding_quality_option()
        self._press_option(value)

    @autopilot_logging.log_action(logger.info)
    def _press_encoding_quality_option(self):
        """Press the quality option button."""
        self.app.pointing_device.click_object(
            self.get_encoding_quality_button())

    def set_hdr_option(self, value):
        """Set the hdr option to specified value."""
        self._press_hdr_option()
        self._press_option(value)

    @autopilot_logging.log_action(logger.info)
    def _press_hdr_option(self):
        """Press the hdr option button."""
        self.app.pointing_device.click_object(self.get_hdr_button())

    def get_sd_card_button(self):
        return self.get_option_button('preferRemovableStorage')

    def _press_sd_card_option(self):
        """Press the flash option button."""
        self.app.pointing_device.click_object(self.get_sd_card_button())

    def set_sd_card_option(self, value):
        """Set the flash option to specified value."""
        self._press_sd_card_option()
        self._press_option(value)

    def get_gps_button(self):
        return self.get_option_button('gpsEnabled')

    def _press_gps_option(self):
        """Press the flash option button."""
        self.app.pointing_device.click_object(self.get_gps_button())

    def set_gps_option(self, value):
        """Set the gps option to specified value."""
        self._press_gps_option()
        self._press_option(value)

    @autopilot_logging.log_action(logger.info)
    def _press_option(self, value):
        """Press the specified option button."""
        self._wait_option_selectable(value)
        self.app.pointing_device.click_object(
            self.get_option_value_button(value))

    def _wait_option_selectable(self, label):
        """ Wait until the label is ready to be selected """
        self.app.wait_select_single('OptionValueButton', label=label,
                                    visible=True)

    def switch_record_control(self):
        """ Switch between camera and video modes """
        record_control = self.get_record_control()
        record_control.visible.wait_for(True)
        self.app.pointing_device.click_object(record_control)

    def is_camera_mode(self):
        return self.get_record_control().iconName == 'camcorder'

    def is_video_mode(self):
        return self.get_record_control().iconName == 'camera-symbolic'

    def get_viewfinder(self):
        return self.app.select_single('ViewFinderView')

    def start_playback(self):
        """ Start playing the video """
        play = self.app.select_many(
            ap_result_sort_keys=KEY_ORDER_BY_X, name='media-playback-start',
            visible=True)[0]
        self.app.pointing_device.click_object(play)

    def click_exposure_button(self):
        """ Click on the button to either take a picture or record a video """
        exposure_button = self.get_exposure_button()
        exposure_button.opacity.wait_for(1.0)
        self.app.pointing_device.click_object(exposure_button)

    def swap_camera(self):
        """ Swap between rear and front cameras """
        swap_camera = self.get_swap_camera_button()
        swap_camera.visible.wait_for(True)
        self.app.pointing_device.click_object(swap_camera)

    def _change_config(self, method, val):
        """ Generic method used to configure one camera setting
        :param method: the method that will make the configuration
        :param val: The value to set
        """
        bottom_edge = self.open_bottom_edge()
        method(val)
        bottom_edge.close()

    def config_flash(self, val):
        """ Set the flash to the val. This method will open and close the
        configuration screen
        :param val: The value to set
        """
        self._change_config(self.set_flash_option, val)

    def config_quality(self, val):
        """ Set the encoding quality to the status. This method will open and
        close the configuration screen
        :param val: The value to set
        """
        self._change_config(self.set_encoding_quality_option, val)

    def config_hdr(self, val):
        """ Set the hdr to the val. This method will open and close the
        configuration screen
        :param val: The value to set
        """
        self._change_config(self.set_hdr_option, val)

    def config_sd_card(self, val):
        """ Set the config to the val. This method will open and close the
        configuration screen
        :param val: The value to set
        """
        self._change_config(self.set_sd_card_option, val)

    def config_gps(self, val):
        """ Set the gps to the val. This method will open and close the
        configuration screen
        :param val: The value to set
        """
        self._change_config(self.set_gps_option, val)

    def record_video(self, duration, random_focus=True):
        """ Start the video record, wait for the timeout and stop the video
        record
        :param duration: The duration of the video to record
        :param random_focus: If True during the video recording it is triggered
        autofocus by touching randomly in the screen
        """
        self.click_exposure_button()
        if random_focus:
            wait_until(self.make_random_focus, timeout=duration)
        else:
            time.sleep(duration)
        self.click_exposure_button()

    def make_random_focus(self):
        """ Click randomly in the screen to force a change in the camera
        focus and based on that, create movement during a video is recorded
        """
        view_finder = self.get_viewfinder()
        (x, y, w, h) = view_finder.globalRect
        from_x = from_y = 0.1
        to_x = 0.9
        to_y = 0.75

        pos_x = randint(x + int((w - x) * from_x), x + int((w - x) * to_x))
        pos_y = randint(y + int((h - y) * from_y), y + int((h - y) * to_y))

        self.app.pointing_device.move(pos_x, pos_y)
        self.app.pointing_device.click()

    def config_camera(self, config):
        """ Config the camera based on the configuration values
        :param config: A CameraConfig object with the desired configuration
        """
        hasattr(config, 'mode') and self._config_mode(config.mode)
        if hasattr(config, 'swap_camera'):
            self.swap_camera()
        hasattr(config, 'hdr') and self.config_hdr(config.hdr)
        hasattr(config, 'flash') and self.config_flash(config.flash)
        hasattr(config, 'quality') and self.config_quality(
            config.quality)
        hasattr(config, 'sd') and self.config_sd_card(config.sd)
        hasattr(config, 'gps') and self.config_gps(config.gps)

    def _config_mode(self, mode):
        if (mode == 'camera' and self.is_video_mode() or
                mode == 'video' and self.is_camera_mode()):
            self.switch_record_control()


class Panel(CustomProxyObjectBase):
    """Panel Autopilot emulator."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'Panel')

    @retry(stop_max_attempt_number=5,
           retry_on_exception=lambda exception: (
               isinstance(exception, AssertionError)))
    def open(self):
        """Open the panel if it's not already opened.
        :return: The panel.
        """
        if not self.opened:
            self._drag_to_open()
            self.opened.wait_for(True)
        return self

    def _drag_to_open(self):
        view_switcher = self.get_root_instance().wait_select_single(
            objectName='viewSwitcher')

        x, y, width, height = view_switcher.globalRect
        center_x = width * 0.50

        # FIXME: a rate higher than 1 does not always make panel move
        if view_switcher.state == "PORTRAIT":
            view_switcher.pointing_device.drag(
                center_x, height, center_x, height * 0.8)
        elif view_switcher.state == "LANDSCAPE":
            self.pointing_device.drag(0, y, width * 0.2, y)
        elif view_switcher.state == "INVERTED_LANDSCAPE":
            self.pointing_device.drag(width, height, width * 0.8, height)
        else:
            self.pointing_device.drag(center_x, y, center_x, height * 0.2)

    @retry(stop_max_attempt_number=5,
           retry_on_exception=lambda exception: (
               isinstance(exception, AssertionError)))
    def close(self):
        """Close the panel if it's not already closed.
        :return: The panel.
        """
        if self.opened:
            self._drag_to_close()
            self.opened.wait_for(False)
        return self

    @autopilot_logging.log_action(logger.info)
    def _drag_to_close(self):
        """
        This method is required due to bug pad.lv/1428090.
        This should be removed as soon as this bug is fixed.
        """
        x, y, _, _ = self.globalRect
        line_x = x + self.width * 0.25
        start_y = y
        stop_y = y + self.height - 1
        self.pointing_device.drag(line_x, start_y, line_x, stop_y)


class PhotoRollHint(CustomProxyObjectBase):
    """Class to represent the photoroll hint shown after taking first photo."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'PhotoRollHint')

    @autopilot_logging.log_action(logger.info)
    def swipe_to_dismiss(self):
        """Swipe left to dismiss hint and open gallery view."""
        x, y, w, h = self.globalRect
        x_mid = x + w // 2
        y_mid = y + h // 2
        self.pointing_device.drag(x_mid, y_mid, x, y_mid)
        self.visible.wait_for(False)
        return self._get_gallery_view()

    def _get_gallery_view(self):
        """Get the visible gallery view."""
        return self.get_root_instance().select_single(
            GalleryView, visible=True)


class GalleryView(CustomProxyObjectBase):
    """Class to represent the photoroll gallery view."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'GalleryView')

    def get_photo(self, path):
        """Look for the photo with the specified file path."""
        url_path = 'image://photo/file://{}'.format(path)
        return self.wait_select_single(
            'QQuickItem', photoUrl=url_path, visible=True)

    def go_back(self):
        """Go back to the camera app view."""
        self.header.go_back()

    @property
    def header(self):
        """Get the header object."""
        return self.select_single(GalleryViewHeader, visible=True)


class GalleryViewHeader(CustomProxyObjectBase):
    """Class to represent the photoroll gallery view toolbar header."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'GalleryViewHeader')

    @autopilot_logging.log_action(logger.info)
    def go_back(self):
        """Press the back button."""
        self.pointing_device.click_object(self._get_back_button())

    def _get_back_button(self):
        """Get the back button from the header."""
        return self.select_single(
            'IconButton', objectName='backButton', visible=True)


class OptionButton(CustomProxyObjectBase):
    """Class to represent the setting buttons on camera overlay panel."""

    @property
    def globalRect(self):
        """Due to http://launchpad.net/bugs/1520164 the globalRect property
        of the button is not correct, so over-ride this property and return
        the correct rectangle co-ordinates.
        Get the incorrect globalRect from child QQuickMouseArea object and
        offset the y co-ordinate by half the height to bring it back to the
        top edge rather than center of the button.
        """
        rect = self.select_single('QQuickMouseArea').globalRect
        y = rect.y - (rect.height // 2)
        return autopilot.introspection.types.Rectangle(
            rect.x, y, rect.width, rect.height)


class OptionValueButton(OptionButton):
    """Class to represent option value buttons on camera overlay panel."""
    pass


class CircleButton(OptionButton):
    """Class to represent top-level option buttons on camera overlay panel."""
    pass


class ShootButton(OptionButton):
    """Class to represent the shoot button on camera overlay panel."""
    pass
