# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import glob
import os

import warlock

from ubuntu_system_tests.helpers import file_system as fs
from ubuntu_system_tests.helpers import wait_until

config_schema = {
    'name': 'CameraConfig',
    'properties': {
        'flash': {'enum': ['Off', 'On', 'Auto']},
        'quality': {
            'enum': ['Fine Quality', 'Normal Quality', 'Basic Quality']
        },
        'hdr': {'enum': ['Off', 'On']},
        'mode': {
            'enum': ['camera', 'video']
        },
        'swap_camera': {'type': 'boolean'},
        'sd': {
            'enum': ['Save to SD Card', 'Save internally']
        },
        'gps': {'enum': ['Off', 'On']},
    },
    'additionalProperties': False
}


def get_config(**kwargs):
    return warlock.model_factory(config_schema)(kwargs)


def config_camera(camera_app, config):
    """ Config the camera based on the configuration values
    :param camera_app: The cpo for the camera app
    :param config: A CameraConfig object with the desired configuration
    """
    hasattr(config, 'mode') and _config_mode(camera_app, config.mode)
    hasattr(config, 'swap_camera') and camera_app.swap_camera()
    hasattr(config, 'hdr') and camera_app.config_hdr(config.hdr)
    hasattr(config, 'flash') and camera_app.config_flash(config.flash)
    hasattr(config, 'quality') and camera_app.config_quality(
        config.quality)
    hasattr(config, 'sd') and camera_app.config_sd_card(config.sd)
    hasattr(config, 'gps') and camera_app.config_gps(config.gps)


def _config_mode(camera_app, mode):
    if (mode == 'camera' and camera_app.is_video_mode() or
            mode == 'video' and camera_app.is_camera_mode()):
        camera_app.switch_record_control()


def get_last_picture_saved(num=1):
    """ Return filepath for last picture taken.
    :param num: The number of pictures that are already saved in the
    camera pictures directory
    :return: The path to the last picture taken
    """
    return _get_last_file_saved(fs.DIR_HOME_CAMERA_PICTURES, num)


def get_last_video_saved(num=1):
    """ Return filepath for last video recorded.
    :param num: The number of videos that are already saved in the
    camera videos directory
    :return: The path to the last video recorded
    """
    return _get_last_file_saved(fs.DIR_HOME_CAMERA_VIDEOS, num)


def _get_last_file_saved(dir, num=1):
    """ Return filepath for last file saved recorded. """
    wait_until(lambda: len(fs.files_in_dir(dir)) == num)
    return os.path.join(dir, fs.files_in_dir(dir)[-1])


def get_recorded_video(index=0):
    """ Get the first video in the videos home dir """
    files = glob.glob(fs.DIR_HOME_CAMERA_VIDEOS + '/video*')
    return files[index]
