# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2016 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from autopilot.exceptions import StateNotFoundError

from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA


from ubuntu_system_tests.helpers import dekko
from ubuntu_system_tests.helpers.autopilot import validate_dbus_object


APP_PATH_ROOT = b'dekkodekkoproject'


class DekkoApp(CustomProxyObjectBase):
    """Autopilot custom proxy object for the dekko app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'dekkodekkoproject',
            applicationName=dekko.APP_CLICK_ID)

    @property
    def no_accounts_page(self):
        return self.wait_select_single(NoAccountsPage)


class NoAccountsPage(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'NoAccountsPage',
            objectName='noAccountsPage')

    def _get_add_account_button(self):
        return self.wait_select_single('Button', objectName='addAccountButton')

    def add_new_account(self):
        """
        Tap 'add account' button on the 'no accounts' screen and return
        the supported accounts types list proxy.
        """
        self.pointing_device.click_object(self._get_add_account_button())
        new_accounts_list = self.get_root_instance().wait_select_single(
            NewAccountTypeListView)
        new_accounts_list.visible.wait_for(True)
        return new_accounts_list


class NewAccountTypeListView(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'NewAccountTypeListView',
            objectName='accountTypeList')

    def _get_accounts_list_item(self, name):
        return self.wait_select_single(
            'AccountsDrawerDelegate', objectName='newAccountListItem_' + name)

    def add_email_account(self, name):
        """
        Add email accounts with basic authentication.

        :param name: Name of the email provider to add
        """
        self.pointing_device.click_object(
            self._get_accounts_list_item(name))
        user_input = self.get_root_instance().wait_select_single(UserInput)
        user_input.visible.wait_for(True)
        return user_input


class UserInput(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'UserInput',
            objectName='usernamePasswordArea')

    def _get_email_field(self):
        """Return the email address entry."""
        email_field_area = self.select_single(
            'TitledTextField', objectName='emailField')
        return email_field_area.select_single('TextField', visible=True)

    def _get_password_field(self):
        """Return the password entry."""
        password_field_area = self.select_single(
            'TitledTextField', objectName='passwordField')
        return password_field_area.select_single('TextField', visible=True)

    def enter_email(self, email):
        """Type in the email address."""
        self._get_email_field().write(email)

    def enter_password(self, password):
        """Type in the password for the relevant email address."""
        self._get_password_field().write(password)

    def login(self, email, password, timeout=120):
        """
        Enter email address and password for the account to login and
        wait for authentication to complete.

        :param email: Email address to login
        :param password: Password for the email to login
        :param timeout: the max time to wait for authentication to complete
        """
        self._enter_credentials_and_proceed(email=email, password=password)
        validation_overlay = self._get_credentials_verification_overlay()
        validation_overlay.wait_until_destroyed(timeout=timeout)
        try:
            user_identity_page = self.get_root_instance().select_single(
                IdentityInput)
            user_identity_page.visible.wait_for(True)
            return user_identity_page
        except StateNotFoundError:
            failure_dialog = self.get_root_instance().wait_select_single(
                'InfoDialog')
            failure_dialog.visible.wait_for(True)
            return None

    def login_imap(self, username, password):
        self._enter_credentials_and_proceed(email=username, password=password)
        details_page = self._get_server_details_input_screen()
        details_page.visible.wait_for(True)
        return details_page

    def _enter_credentials_and_proceed(self, email, password):
        self.enter_email(email=email)
        self.enter_password(password=password)
        self.pointing_device.click_object(self._get_next_button())

    def _get_next_button(self):
        """Tap the 'next' button in the header to proceed to next screen."""
        return self.get_root_instance().select_single(
            'HeaderButton', iconSource='image://theme/next')

    def _get_credentials_verification_overlay(self):
        """Return the credentials validation screen."""
        return self.get_root_instance().wait_select_single(
            CredentialValidation)

    def _get_server_details_input_screen(self):
        return self.get_root_instance().wait_select_single(ServerInput)


class CredentialValidation(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'CredentialValidation',
            objectName='credentialValidation')


class IdentityInput(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'IdentityInput',
            objectName='userIdentityInputScreen')

    def _get_next_button(self):
        return self.get_root_instance().select_single(
            'HeaderButton', iconSource='image://theme/next')

    def go_to_details_page(self):
        self.pointing_device.click_object(self._get_next_button())
        details_input_page = self.get_root_instance().wait_select_single(
            DetailsInput)
        details_input_page.visible.wait_for(True)
        return details_input_page


class DetailsInput(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'DetailsInput',
            objectName='accountDetailExtraScreen')

    def _get_next_button(self):
        return self.get_root_instance().select_single(
            'HeaderButton', iconSource='image://theme/next')

    def finish_wizard(self):
        self.pointing_device.click_object(self._get_next_button())
        message_list_page = self.get_root_instance().wait_select_single(
            MessageListPage)
        message_list_page.visible.wait_for(True)
        return message_list_page


class MessageListPage(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MessageListPage',
            objectName='emailListPage')


class ServerInput(CustomProxyObjectBase):

    """Autopilot CPO for email server details screen."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ServerInput',
            objectName='serverDetailsPage')

    def _get_credentials_verification_overlay(self):
        """Return the credentials validation screen."""
        return self.get_root_instance().wait_select_single(
            CredentialValidation)

    def _get_input_field(self, object_name):
        entry_area = self.select_single(
            'TitledTextField', objectName=object_name)
        return entry_area.select_single('TextField', visible=True)

    def enter_imap_host_name(self, hostname):
        hostname_field = self._get_input_field(object_name='imapServerField')
        hostname_field.write(hostname)

    def enter_imap_port_number(self, port_number):
        port_number_field = self._get_input_field(object_name='imapPortField')
        port_number_field.write(port_number)

    def select_imap_encryption_type(self, encryption_type):
        selector = self.select_single(
            EncryptionSelector, objectName='imapEncryptionSelector')
        selector.select_encryption(encryption_type=encryption_type)

    def _get_next_button(self):
        return self.get_root_instance().select_single(
            'HeaderButton', iconSource='image://theme/next')

    def _go_to_details_page(self, timeout):
        self.pointing_device.click_object(self._get_next_button())
        validation_overlay = self._get_credentials_verification_overlay()
        validation_overlay.wait_until_destroyed(timeout=timeout)
        try:
            details_page = self.get_root_instance().select_single(
                DetailsInput)
            details_page.visible.wait_for(True)
            return details_page
        except StateNotFoundError:
            failure_dialog = self.get_root_instance().wait_select_single(
                'InfoDialog')
            failure_dialog.visible.wait_for(True)
            return None

    def enter_imap_server_details(self, host, port, encryption, timeout=120):
        """Enter server details and proceed"""
        self.select_imap_encryption_type(encryption_type=encryption)
        self.enter_imap_host_name(hostname=host)
        self.enter_imap_port_number(port_number=port)
        return self._go_to_details_page(timeout=timeout)


class EncryptionSelector(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'EncryptionSelector')

    def _expand(self):
        if self.state != 'expanded':
            self._click_expander()
            self.state.wait_for('expanded')

    def _contract(self):
        if self.state != '':
            self._click_expander()
            self.state.wait_for('')

    def _get_expander(self):
        return self.select_single('UCAbstractButton')

    def _click_expander(self):
        self.pointing_device.click_object(self._get_expander())

    def _get_encryption_list_item(self, encryption_type):
        return self.select_single(
            'CheckboxWithLabel',
            objectName='encryption{}'.format(encryption_type),
            visible=True)

    def _select_encryption(self, encryption_type):
        list_item = self._get_encryption_list_item(
            encryption_type=encryption_type)
        check_box = list_item.select_single('ListItemWithActionsCheckBox')
        self.pointing_device.click_object(check_box)
        check_box.checked.wait_for(True)

    def select_encryption(self, encryption_type):
        """Tap the encryption selector and select the requested type."""
        self._expand()
        self._select_encryption(encryption_type=encryption_type)
        self._contract()
