# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

import autopilot
import time
from autopilot.exceptions import StateNotFoundError

from ubuntu_system_tests.helpers.addressbook._cpo import (
    ContactEditorPage,
    ContactViewPage
    )
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA

from ubuntu_system_tests.helpers.autopilot import (
    patched_select_single, validate_dbus_object)
from ubuntu_system_tests.helpers.dialer_app.utils import (
    unformatted_phone_number)

logger = logging.getLogger(__name__)

APP_PATH_ROOT = b'DialerApp'


class DialerApp(CustomProxyObjectBase):

    """Autopilot custom proxy object for the dialer app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'DialerApp',
            applicationName='DialerApp')

    @property
    def main_view(self):
        return patched_select_single(
            root=self, class_object=MainView, class_name='MainView',
            objectName='mainView')


class MainView(CustomProxyObjectBase):
    def __init__(self, *args):
        super().__init__(*args)
        self.logger = logging.getLogger(__name__)

    @property
    def live_call_page(self):
        # wait until we actually have the calls before returning the live call
        self.hasCalls.wait_for(True)
        return self.wait_select_single(LiveCall, active=True)

    @property
    def contacts_page(self):
        return self._get_page(ContactsPage, 'contactsPage')

    @property
    def contact_editor_page(self):
        return self._get_page(DialerContactEditorPage, 'contactEditorPage')

    @property
    def contact_view_page(self):
        return self._get_page(DialerContactViewPage, 'contactViewPage')

    def get_first_log(self):
        return self.wait_select_single(objectName="historyDelegate0")

    def _click_button(self, button):
        """Generic way to click a button"""
        self.visible.wait_for(True)
        button.visible.wait_for(True)
        self.pointing_device.click_object(button)
        return button

    def _long_press(self, obj):
        """long press on object because press_duration is not honored on touch
        see bug #1268782

        :parameter obj: the object to long press on
        """

        self.pointing_device.move_to_object(obj)
        self.pointing_device.press()
        time.sleep(3)
        self.pointing_device.release()

    def check_ussd_error_dialog_visible(self):
        """Check if ussd error dialog is visible"""
        dialog = None
        try:
            dialog = self.wait_select_single(objectName="ussdErrorDialog")
        except:
            # it is ok to fail in this case
            return False

        return dialog.visible

    def check_ussd_progress_dialog_visible(self):
        """Check if ussd progress dialog is visible"""
        dialog = None
        try:
            dialog = self.wait_select_single(
                objectName="ussdProgressDialog")
        except:
            # it is ok to fail in this case
            return False

        return dialog.visible

    def _get_page(self, page_type, page_name):
        page = self.wait_select_single(
            page_type, objectName=page_name, active=True)
        return page

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MainView', objectName='mainView')

    @property
    def dialer_page(self):
        return self.wait_select_single(
            DialerPage, objectName='dialerPage', visible=True)

    @property
    def contacts_edit_page(self):
        return self.wait_select_single(
            DialerContactEditorPage,
            objectName='contactEditorPage',
            visible=True
        )

    def get_main_title(self):
        """ Retrieve the app title """
        return self.wait_select_single('AppHeader').title

    def get_call_hold_failure_dialog(self):
        """Return dialog displayed to indicate an error activating call hold.
        :return: Dialog object or None if it is not found.
        """
        try:
            return patched_select_single(
                self, Dialog, class_name='Dialog',
                title='Call holding failure', visible=True)
        except autopilot.exceptions.StateNotFoundError:
            return None


class LiveCall(MainView):

    def get_elapsed_call_time(self):
        """Return the elapsed call time"""
        return self.wait_select_single(objectName='stopWatch').elapsed

    def _get_hangup_button(self):
        """Return the hangup button"""
        return self.wait_select_single(objectName='hangupButton')

    def _get_call_hold_button(self):
        """Return the call holding button"""
        return self.wait_select_single(objectName='callHoldButton')

    def _get_swap_calls_button(self):
        """Return the swap calls button"""
        return self._get_call_hold_button()

    def _get_new_call_button(self):
        return self.wait_select_single(objectName='newCallButton')

    def get_multi_call_display(self):
        """Return the multi call display panel"""
        return self.wait_select_single(objectName='multiCallDisplay')

    def get_multi_call_item_for_number(self, number):
        """Return the multi call display item for the given number"""
        return self.wait_select_single(objectName='callDelegate',
                                       phoneNumber=number)

    def click_hangup_button(self):
        """Click and return the hangup page"""
        return self._click_button(self._get_hangup_button())

    def click_call_hold_button(self):
        """Click the call holding button"""
        return self._click_button(self._get_call_hold_button())

    def click_swap_calls_button(self):
        """Click the swap calls button"""
        return self._click_button(self._get_swap_calls_button())

    def click_new_call_button(self):
        """Click the new call button"""
        return self._click_button(self._get_new_call_button())

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'LiveCall', objectName='pageLiveCall')

    def get_stop_watch(self):
        """Return the stop watch object"""
        return self.wait_select_single(
            StopWatch, objectName='stopWatch', visible=True)

    def hangup_call(self):
        """Hangup the active call and wait for state to change."""
        self.click_hangup_button()
        try:
            self.active.wait_for(False)
        except autopilot.exceptions.StateNotFoundError:
            # live call page has closed so call has ended
            pass

    def hold_state(self):
        """Return hold state of current call."""
        return self.onHold

    def _get_hold_icon_name(self, hold_state):
        """Return the expected icon name for hold button in specified state."""
        if hold_state:
            return 'media-playback-start'
        else:
            return 'media-playback-pause'

    def toggle_hold(self):
        """Toggle the hold state of the current call by pressing hold button.
        :return: True if the hold state was toggled successfully,
                 False otherwise.
        """
        current_state = self.onHold
        next_state = not current_state
        hold_button = self._get_call_hold_button()
        hold_button.iconSource.wait_for(
            self._get_hold_icon_name(current_state), timeout=0)
        self.click_call_hold_button()
        try:
            self.onHold.wait_for(next_state, timeout=1)
            hold_button.iconSource.wait_for(
                self._get_hold_icon_name(next_state))
            return True
        except AssertionError:
            return False


class StopWatch(CustomProxyObjectBase):

    def elapsed_time(self):
        """Return elapsed time as string"""
        return self.elapsed


class DialerPage(MainView):

    def reveal_bottom_edge_page(self):
        """Bring the bottom edge page to the screen"""
        try:
            start_x = (
                self.globalRect.x +
                (self.globalRect.width * 0.5))
            # Start swiping from the top of the component because after some
            # seconds it gets almost fully hidden. The center will be out of
            # view.
            start_y = self.globalRect.y + self.height - 3
            stop_y = start_y - (self.height * 0.7)
            self.pointing_device.drag(start_x, start_y, start_x, stop_y,
                                      rate=2)
            self.bottomEdgeCommitted.wait_for(True)
        except StateNotFoundError:
            self.logger.error('BottomEdge element not found.')
            raise

    def _get_keypad_entry(self):
        return self.wait_select_single("KeypadEntry")

    def _get_keypad_keys(self):
        return self.select_many("KeypadButton")

    def _get_keypad_key(self, number):
        buttons_dict = {
            "0": "buttonZero",
            "1": "buttonOne",
            "2": "buttonTwo",
            "3": "buttonThree",
            "4": "buttonFour",
            "5": "buttonFive",
            "6": "buttonSix",
            "7": "buttonSeven",
            "8": "buttonEight",
            "9": "buttonNine",
            "*": "buttonAsterisk",
            "#": "buttonHash",
        }
        return self.wait_select_single("KeypadButton",
                                       objectName=buttons_dict[number])

    def _get_erase_button(self):
        """Return the erase button"""
        return self.wait_select_single("CustomButton",
                                       objectName="eraseButton")

    def click_call_button(self):
        """Click and return the call button"""
        return self._click_button(self._get_call_button())

    def click_erase_button(self):
        """Click the erase button"""
        self._click_button(self._get_erase_button())

    def click_keypad_button(self, keypad_button):
        """click the keypad button

        :param keypad_button: the clicked keypad_button
        """
        self._click_button(keypad_button)

    def trigger_copy_and_paste(self):
        """Invoke the copy and paste popup"""
        self._long_press(self._get_keypad_entry())

    def trigger_select_all(self):
        """Trigger select all"""
        button = self.get_root_instance().wait_select_single(
            'UCAbstractButton',
            objectName="select_all_button")
        self._click_button(button)

    def trigger_copy(self):
        """Trigger copy"""
        button = self.get_root_instance().wait_select_single(
            'UCAbstractButton',
            objectName="copy_button")
        self._click_button(button)

    def trigger_paste(self):
        """Trigger paste"""
        button = self.get_root_instance().wait_select_single(
            'UCAbstractButton',
            objectName="paste_button")
        self._click_button(button)

    def trigger_cut(self):
        """Trigger cut"""
        button = self.get_root_instance().wait_select_single(
            'UCAbstractButton',
            objectName="cut_button")
        self._click_button(button)

    def dial_number(self, number, formattedNumber):
        """Dial given number (string) on the keypad and return keypad entry

        :param number: the number to dial
        """
        for digit in number:
            button = self._get_keypad_key(digit)
            self.click_keypad_button(button)

        entry = self._get_keypad_entry()
        entry.value.wait_for(formattedNumber)
        return entry

    def call_number(self, number, formattedNumber):
        """Dial number and call return call_button"""
        self.dial_number(number, formattedNumber)
        self.click_call_button()
        return self.get_root_instance().wait_select_single(LiveCall)

    def click_header_action(self, action):
        """Click the action 'action' on the header"""
        action = self.wait_select_single(objectName='%s_button' % action)
        self.pointing_device.click_object(action)

    def click_contacts_button(self):
        self.click_header_action('contacts')

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'DialerPage', objectName='dialerPage')

    def _get_phone_number_field(self):
        """Return PhoneNumberInput object"""
        return self.select_single('PhoneNumberField', visible=True)

    def get_entered_phone_number(self, unformatted=False):
        """Return the phone number entered into the dialer"""
        number = self._get_phone_number_field().displayText
        if unformatted:
            number = unformatted_phone_number(number)
        return number

    def _get_call_button(self):
        """Over-ride to ensure the button is visible and enabled."""
        return self.wait_select_single(
            objectName='callButton', visible=True, enabled=True)

    def _get_add_new_contact_button(self):
        """Return the 'add to contact' button."""
        return self.wait_select_single(
            'CustomButton',
            icon='contact-new',
            visible=True
        )

    def dial_number_add_new_contact(self, number, formatted_number):
        """Type in a phone number and click the 'add to contact' button."""
        self.dial_number(number, formatted_number)
        self.pointing_device.click_object(self._get_add_new_contact_button())


class DialerContactEditorPage(ContactEditorPage):

    def click_action_button(self, action_name):
        actions = self.select_many(objectName='%s_button' % action_name)
        for action in actions:
            if action.enabled:
                self.pointing_device.click_object(action)
                return

        raise StateNotFoundError(action_name)

    def save(self):
        """
        Press the 'Save' button
        """
        self.click_action_button('save')

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'DialerContactEditorPage',
            objectName='contactEditorPage')

    def _get_new_contact_number_field(self):
        return self.wait_select_single(
            'TextInputDetail', objectName='phoneNumber_0')

    def _get_new_contact_first_name_field(self):
        return self.wait_select_single(
            'TextInputDetail', objectName='firstName')

    def _get_new_contact_last_name_field(self):
        return self.wait_select_single(
            'TextInputDetail', objectName='lastName')

    def _get_contact_detail_phone_number(self):
        return self.select_single(objectName='base_phoneNumber_0')

    def _wait_for_number_field_to_focus(self):
        """Ensure the phone number field is focused."""
        self._get_contact_detail_phone_number().activeFocus.wait_for(True)

    def _write_contact_number(self, phone_number):
        """Type the contact number into the field."""
        phone_number_field = self._get_new_contact_number_field()
        phone_number_field.write(phone_number)

    def _write_first_name(self, first_name):
        """Type first name of the contact to save."""
        first_name_field = self._get_new_contact_first_name_field()
        first_name_field.write(first_name)

    def _write_last_name(self, last_name):
        """Type last name of the contact to save."""
        first_name_field = self._get_new_contact_last_name_field()
        first_name_field.write(last_name)

    def create_new_contact(self, first_name, last_name, phone_number=None):
        """
        Save a new contact by typing in its details.
        :param first_name: first name for the contact
        :param last_name: last name for the contact
        :param phone_number: phone number for the contact
        """
        if phone_number:
            self._write_contact_number(phone_number)
        else:
            self._wait_for_number_field_to_focus()

        self._write_first_name(first_name)
        self._write_last_name(last_name)
        self.save()


class DialerContactViewPage(ContactViewPage):

    def call_phone(self, index):
        phone_group = self.select_single(
            'ContactDetailGroupWithTypeView',
            objectName='phones')

        call_buttons = phone_group.select_many(
            "ActionButton",
            objectName="tel-contact")
        self.pointing_device.click_object(call_buttons[index])

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'DialerContactViewPage',
            objectName='contactViewPage')

    def get_contact_primary_number(self):
        """Return the phone number of the currently opened contact."""
        return self.wait_select_single(objectName='label_phoneNumber_0.0').text

    def get_title(self):
        """Return title of contact view page."""
        return self.select_single('PageHeader').title


class ContactsPage(CustomProxyObjectBase):
    """Autopilot custom proxy object for ContactsPage components."""

    def _click_button(self, button):
        """Generic way to click a button"""
        self.visible.wait_for(True)
        button.visible.wait_for(True)
        self.pointing_device.click_object(button)
        return button

    def _get_add_new_button(self):
        """Return the add-new button"""
        return self.wait_select_single('ContactListButtonDelegate',
                                       objectName='addNewButton')

    def click_add_new(self):
        self._click_button(self._get_add_new_button())

    def click_contact(self, index):
        contact_delegate = self._get_contact_delegate(index)
        self.pointing_device.click_object(contact_delegate)

    def open_contact(self, index):
        self.click_contact(index)
        return self.get_root_instance().select_single(
            DialerContactViewPage, objectName='contactViewPage')

    def _get_contact_delegate(self, index):
        contact_delegates = self._get_sorted_contact_delegates()
        return contact_delegates[index]

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ContactsPage',
            objectName='contactsPage')

    def _get_sorted_contact_delegates(self):
        contact_delegates = self.wait_select_many(
            'ContactDelegate', visible=True)
        return sorted(
            contact_delegates, key=lambda delegate: delegate.globalRect.y)


class Dialog(CustomProxyObjectBase):
    """Generic CPO for dialer-app owned Dialog objects."""

    def _get_close_button(self):
        return self.select_single(objectName='closeDialog', visible=True)

    @autopilot.logging.log_action(logger.info)
    def close(self):
        self.pointing_device.click_object(self._get_close_button())
        self.wait_until_destroyed()
