# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from autopilot import exceptions
from autopilot.introspection.utilities import sort_by_keys
from retrying import retry

from ubuntu_system_tests.helpers.ubuntuuitools._cpo._header import Header
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._mainview import MainView
from ubuntu_system_tests.helpers.autopilot import validate_dbus_object
from ubuntu_system_tests.helpers.autopilot.list import KEY_ORDER_BY_X
from ubuntu_system_tests.helpers import context
from ubuntu_system_tests.helpers import file_system as fs

APP_PATH_ROOT = b'comubuntugallery'


class GalleryApplication(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'GalleryApplication')

    @property
    def picker_screen(self):
        return self.select_single(objectName='pickerMainView', visible=True)

    @property
    def events_view(self):
        return self.select_single(EventsOverview)

    @property
    def photos_view(self):
        return self.select_single(PhotosOverview)

    def get_opened_photo(self):
        """Returns the first visible opened photo."""
        return self.wait_select_single("SingleMediaViewer",
                                       objectName="openedMedia0",
                                       visible=True)

    def _get_image(self):
        """Return the visible (also opaque) QQuickImage."""
        photo = self.get_opened_photo()
        # companion QQuickImage has opacity 0.0
        return photo.wait_select_single(
            'QQuickImage', visible=True, globalRect=photo.globalRect,
            opacity=1.0)

    def get_image_source(self):
        return self._get_image().source


class MainScreen(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MainScreen', objectName='overview')

    def get_header(self):
        """Return the Header emulator of the MainView."""
        return self.select_single(AppHeader, objectName='MainView_Header')


class AppHeader(Header):
    """Header Autopilot helper.

    We override this helper because on the gallery the gesture to show the
    header when it's hidden it's not the default.

    """

    # XXX We are overriding internal methods that may change on the toolkit.
    # The helper on the toolkit needs a public method that will be safe to
    # override. Reported as bug http://pad.lv/1363591 --elopio - 2014-31-08

    def _is_visible(self):
        return self.visible

    def _show(self):
        self.pointing_device.click_object(self._get_top_container())
        self.visible.wait_for(True)


class GalleryAppException(Exception):
    pass


class QQuickView(CustomProxyObjectBase):
    """An emulator class that makes it easy to interact with
       general components of the gallery app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'QQuickView')

    def get_main_photo_viewer_loader(self):
        """Returns the loader item for the PhotoViewer."""
        return self.select_single("QQuickLoader",
                                  objectName="photoViewerLoader")

    def get_main_photo_viewer(self):
        """Returns the MediaListView."""
        return self.wait_select_single("MediaListView",
                                       objectName="mediaListView")

    def get_albums_viewer_loader(self):
        """Returns the loader item for the AlbumsOverview."""
        return self.select_single("QQuickLoader",
                                  objectName="albumsCheckerboardLoader")

    def get_delete_dialog(self):
        """Returns the delete dialog in the events view."""
        return self.wait_select_single("DeleteDialog",
                                       objectName="deleteDialog")

    def delete_dialog_shown(self):
        dialog = self.select_many(
            "DeleteDialog",
            objectName="deleteDialog")
        return len(dialog) >= 1

    def get_delete_dialog_delete_button(self):
        """Returns the delete button of the delete popover."""
        return self.wait_select_single("Button",
                                       objectName="deleteDialogYes",
                                       visible=True)

    def get_delete_dialog_cancel_button(self):
        """Returns the cancel button of the delete popover."""
        return self.wait_select_single("Button",
                                       objectName="deleteDialogNo",
                                       visible=True)

    @retry(stop_max_attempt_number=5, wait_fixed=2000,
           retry_on_result=lambda result: not result)
    def get_all_albums(self):
        """Returns all albums in the albums view"""
        albums = self.select_many("CheckerboardDelegate",
                                  objectName="checkerboardDelegate")
        return albums

    def get_first_album(self):
        """Returns the first album in the albums view."""
        return self.get_album_at(0)

    @retry(stop_max_attempt_number=5, wait_fixed=2000,
           retry_on_result=lambda result: not result)
    def get_album_at(self, position):
        """Returns the albums at this position in the albums view"""
        albums = self.select_many("CheckerboardDelegate",
                                  objectName="checkerboardDelegate")
        albums = sorted(
            albums, key=lambda album: (album.globalRect.y,
                                       album.globalRect.x))
        return albums[position]

    def get_edit_album_button(self):
        """Returns the edit album button in the album popover"""
        return self.select_single("Standard",
                                  objectName="editAlbumListItem")

    def get_cover_menu_item(self, name):
        """Returns the item of the cover menu with the name *name*

        :raises StateNotFoundError: if there is no item with the name *name*

        """
        return self.select_single(
            "Standard",
            objectName="albumCoverMenuItem_" + name
        )

    def _ensure_delete_dialog_visible(self):
        delete_dialog = self.get_delete_dialog()
        delete_dialog.opacity.wait_for(1)

    def _click_dialog_button(self, button):
        self._ensure_delete_dialog_visible()
        self.pointing_device.click_object(button)

    def click_delete_dialog_cancel_button(self):
        button = self.get_delete_dialog_cancel_button()
        self._click_dialog_button(button)

    def click_delete_dialog_delete_button(self):
        button = self.get_delete_dialog_delete_button()
        self._click_dialog_button(button)


class EventsOverview(QQuickView):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'EventsOverview')

    def get_event(self, event_number=0):
        """Return an event in the event view based on index number

        :param event_number: the index number of the organicEventItem to get
        """
        return self.select_single(
            'OrganicMediaList',
            objectName='organicEventItem{}'.format(int(event_number))
        )

    def number_of_events(self):
        """Return the number of events in the model behind the event view"""
        return self._eventCount

    def number_of_photos_in_events(self):
        """Return the number of photos in events"""
        photo_delegates = self.select_many(
            "QQuickItem",
            objectName="eventPhoto"
        )
        return len(photo_delegates)

    def number_of_photos_in_event(self, event):
        """Return the number of photo delgated in an event"""
        photo_delegates = event.select_many(objectName='eventPhoto')
        return len(photo_delegates)

    def get_share_peer_picker(self):
        """Returns the photo viewer share picker."""
        return self.wait_select_single(objectName="sharePickerEvents",
                                       visible=True)

    def _get_image_in_event_view(self, image_name):
        """Return the photo of the gallery based on image name.

        :param image_name: the name of the photo in the event to return"""
        events = self.select_many('OrganicMediaList')
        for event in events:
            photos = event.select_many(
                'QQuickItem',
                objectName='eventPhoto'
            )
            for photo in photos:
                image = EventsOverview._get_image_from_photo(image_name, photo)
                if image:
                    return image
        raise GalleryAppException(
            'Photo with image name {} could not be found'.format(image_name))

    @staticmethod
    def _get_image_from_photo(image_name, photo):
        images = photo.select_many('QQuickImage')
        return EventsOverview._get_image(image_name, images)

    def _get_item_in_event_view(self, image_name, event_index_num=0):
        """Return the item of the gallery based on image name.

        :param image_name: the name of the photo in the event to return"""
        event = self.get_event(event_index_num)
        photos = event.select_many(
            'QQuickItem',
            objectName='eventPhoto'
        )
        for photo in photos:
            images = photo.select_many('QQuickImage')
            if EventsOverview._get_image(image_name, images):
                    return photo
        raise GalleryAppException(
            'Photo with image name {} could not be found'.format(image_name))

    @staticmethod
    def _get_image(image_name, images):
        for image in images:
            src = image.source.split('?')[0]
            if str(src).endswith(image_name):
                return image
        return None

    def click_photo(self, photo_name):
        """Click photo with name and event

        :param photo_name: name of file to click
        """
        photo = self._get_image_in_event_view(photo_name)
        self.pointing_device.click_object(photo)

    def select_photo(self, photo_name, event_index_num=0):
        """Select photo with name and event

        :param photo_name: name of file to click
        :param event_index_num: index of event to click
        """
        photo = self._get_item_in_event_view(photo_name, event_index_num)
        checkbox = photo.select_single(objectName="selectionCheckbox")
        self.pointing_device.click_object(checkbox)

    def is_image_displayed(self, image_name):
        try:
            return self._get_image_in_event_view(image_name) is not None
        except GalleryAppException:
            return False

    def pick_photo(self, file_path):
        """
        Picks the requested photo to be consumed by the requester

        :param file_path: identifier for the photo to pick
        """
        self._select_photo(file_path)
        self._click_pick_button()

    def _select_photo(self, file_path):
        """Selects a photo in EventView by long press."""
        photo = self._get_image_in_event_view(file_path)
        self.pointing_device.click_object(photo, press_duration=1)

    def _get_header_pick_button(self):
        """Returns the validate/pick button on the toolbar."""
        return self.select_single(
            objectName='pickButton_button', visible=True)

    def _click_pick_button(self):
        """Taps the pick button in the header."""
        button = self._get_header_pick_button()
        self.pointing_device.click_object(button)

    def is_image_in_view(self, file_path):
        """ Indicate if the image is currently being displayed  """
        self.wait_select_single('MediaViewer', openCompleted=True)
        current = self.select_single('MainScreen').mediaCurrentlyInView
        return current == fs.get_file_path(file_path)

    @retry(stop_max_delay=context.shared.retry_time,
           retry_on_exception=lambda exception:
           isinstance(exception, exceptions.StateNotFoundError))
    def get_images_in_event_view(self, image_path):
        """ Return all the photos stored in the image path.
        :param image_path: the path to match with the images source
        """
        all_photos = []
        photos = self.select_many('UCUbuntuShape', visible=True)
        for photo in photos:
            images = photo.select_many('QQuickImage', status=1)
            imgs = [image for image in images if image_path in image.source]
            all_photos.extend(imgs)

        if not all_photos:
            raise exceptions.StateNotFoundError(class_name='QQuickImage',
                                                status=1)
        return sort_by_keys(all_photos, KEY_ORDER_BY_X)

    def get_media_list_for_event(self, event_number=0):
        event = self.get_event(event_number)
        return event.select_single('QQuickListView')

    def swipe_to_show_more_on_the_right(self, event_number=0):
        event = self.get_event(event_number)
        list = self.get_media_list_for_event()
        list.swipe_to_show_more_right(containers=[event])

    def swipe_to_show_more_on_the_left(self, event_number=0):
        event = self.get_event(event_number)
        list = self.get_media_list_for_event()
        list.swipe_to_show_more_left(containers=[event])


class PhotosOverview(QQuickView):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'PhotosOverview')

    def get_popup_album_picker(self):
        """Returns the photo viewer album pickers."""
        return self.wait_select_single("PopupAlbumPicker",
                                       objectName="popupAlbumPicker")

    def get_share_peer_picker(self):
        """Returns the photo viewer share picker."""
        return self.wait_select_single(objectName="sharePicker",
                                       visible=True)

    def get_photo_editor(self):
        """Returns the photo edit dialog."""
        return self.wait_select_single("PhotoEditor")

    def get_revert_to_original_dialog(self):
        """Returns the revert to original dialog."""
        return self.wait_select_single("Dialog",
                                       objectName="revertPromptDialog")

    def get_cancel_revert_to_original_button(self):
        """Returns the revert to original cancel button."""
        return self.get_revert_to_original_dialog().wait_select_single(
            "Button",
            objectName="cancelRevertButton",
            visible=True)

    def get_confirm_revert_to_original_button(self):
        """Returns the revert to original confirm button."""
        return self.get_revert_to_original_dialog().wait_select_single(
            "Button",
            objectName="confirmRevertButton",
            visible=True)

    def get_photos_list(self):
        return self.wait_select_single("MediaListView")

    def get_editor_actions_bar(self):
        """Returns the actions bar for the editor."""
        return self.select_single("ActionsBar",
                                  objectName="editorActionsBar")

    def get_editor_action_button_by_text(self, button_text):
        """Returns the action button from the editor by text."""
        actions_bar = self.get_editor_actions_bar()
        buttons = actions_bar.select_many('AbstractButton', text=button_text)
        if buttons:
            return buttons[0]
        raise exceptions.StateNotFoundError(
            'Editor action button {} could not be found'.format(button_text))

    def get_crop_action_button(self):
        """Returns the crop item of the edit dialog."""
        return self.get_editor_action_button_by_text("Crop")

    def get_rotate_action_button(self):
        """Returns the rotate item of the edit dialog."""
        return self.get_editor_action_button_by_text("Rotate")

    def get_undo_menu_item(self):
        """Returns the undo item of the edit dialog."""
        return self.select_single("Standard", objectName="undoListItem")

    def get_redo_menu_item(self):
        """Returns the redo item of the edit dialog."""
        return self.select_single("Standard", objectName="redoListItem")

    def get_revert_action_button(self):
        """Returns the revert to original menu item in the edit dialog."""
        return self.get_editor_action_button_by_text("Revert to Original")

    def get_auto_enhance_menu_item(self):
        """Returns the 'auto enhance' menu item in the edit dialog."""
        return self.select_single("Standard", objectName='enhanceListItem')

    def get_delete_popover_cancel_item(self):
        """Returns the cancel button of the delete popover."""
        return self.wait_select_single("Button",
                                       objectName="deletePhotoDialogNo",
                                       visible=True)

    def get_crop_interactor(self):
        """Returns the crop interactor."""
        return self.wait_select_single("CropInteractor",
                                       objectName="cropInteractor")

    def get_crop_overlay(self):
        """Returns the crop overlay."""
        return self.wait_select_single("CropOverlay",
                                       objectName="cropOverlay")

    def get_top_left_crop_corner(self):
        """Returns the top left corner of the crop overlay for dragging."""
        return self.wait_select_single("CropCorner",
                                       objectName="topLeftCropCorner")

    def get_crop_overlays_crop_icon(self):
        """Returns the crop icon of the crop overlay."""
        return self.wait_select_single("Button",
                                       objectName="centerCropIcon",
                                       visible=True)

    def get_edit_preview(self):
        """Returns the edit preview."""
        return self.wait_select_single("EditPreview",
                                       objectName="editPreview")

    def _click_item(self, item):
        self.pointing_device.click_object(item)

    def click_rotate_button(self):
        rotate_item = self.get_rotate_action_button()
        self._click_item(rotate_item)

    def click_crop_button(self):
        crop_item = self.get_crop_action_button()
        self._click_item(crop_item)

    def click_undo_item(self):
        undo_item = self.get_undo_menu_item()
        self._click_item(undo_item)

    def click_redo_item(self):
        redo_item = self.get_redo_menu_item()
        self._click_item(redo_item)

    def click_revert_button(self):
        revert_item = self.get_revert_action_button()
        self._click_item(revert_item)

    def click_cancel_revert_button(self):
        cancel_item = self.get_cancel_revert_to_original_button()
        self._click_item(cancel_item)

    def click_confirm_revert_button(self):
        confirm_item = self.get_confirm_revert_to_original_button()
        self._click_item(confirm_item)

    def click_enhance_item(self):
        enhance_item = self.get_auto_enhance_menu_item()
        self._click_item(enhance_item)

    @retry(stop_max_delay=context.shared.retry_time,
           retry_on_exception=lambda exception:
           isinstance(exception, exceptions.StateNotFoundError))
    def get_photo_component(self):
        return self.wait_select_single('GalleryPhotoComponent')


class PickerScreen(MainView):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'PickerScreen',
            objectName='pickerMainView')

    def pick_button(self):
        return self.select_single(objectName="pickButton_button")

    def get_photos_tab_button(self):
        """Returns the photos tab."""
        return self.select_single("AbstractButton", buttonIndex=3)

    def events_view(self):
        return self.select_single("OrganicView", objectName="eventSelectView")

    def first_media_in_events_view(self):
        events_view = self.events_view()
        return events_view.select_many("OrganicItemInteraction",
                                       objectName="eventsViewPhoto")[0]

    def grid_view(self):
        return self.select_single("MediaGrid")

    def first_media_in_grid_view(self):
        grid_view = self.grid_view()
        return grid_view.select_many("OrganicItemInteraction")[0]

    def go_to_photos(self):
        """
        Switch to the photos tab on picker screen
        Return the Page object representing the photos
        """
        self.switch_to_tab('photosTab')
        return self.select_single(Page, objectName='photosPage')

    def click_pick_button(self):
        """Click on the pick button"""
        self.pointing_device.click_object(self.pick_button())


class Page(PickerScreen):
    """Class to represent photos page view from picker screen"""

    def _get_named_photo_element(self, photo_name):
        """
        Return the ShapeItem container object for the named photo.
        This object can be clicked to enable the photo to be selected.
        """
        elements = self.grid_view().select_many('QQuickImage')
        for element in elements:
            element.status.wait_for(1)
            src = element.source.split('?')[0]
            if str(src).endswith(photo_name):
                return element.get_parent()
        raise GalleryAppException(
            'Photo with image name {} could not be found'.format(photo_name))

    def select_named_photo(self, photo_name):
        """Select the named photo from the picker view."""
        photo_element = self._get_named_photo_element(photo_name)
        checkbox = photo_element.select_single(objectName="selectionCheckbox")
        self.pointing_device.click_object(checkbox)
