# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
import subprocess
import time

from ubuntu_system_tests.helpers.processes import is_process_running
from ubuntu_system_tests.helpers.testbed import run_command
from ubuntu_system_tests.helpers.utils import wait_until


def get_init_system():
    init_sim_link = os.readlink('/sbin/init')
    return Upstart() if 'upstart' in init_sim_link else SystemD()


class InitSystem:
    """This contains a base implementation for an Init System, which now can
     either be systemd or upstart. The init system is the one responsible for
     start, stop, restart jobs.
     """

    def __init__(self, start_command, stop_command,
                 restart_command, status_command,
                 set_env_command, unset_env_command,
                 get_env_command):

        self.start_command = start_command
        self.stop_command = stop_command
        self.status_command = status_command
        self.restart_command = restart_command
        self.set_env_command = set_env_command
        self.unset_env_command = unset_env_command
        self.get_env_command = get_env_command

    def start(self, job, *args, **kwargs):
        """Start the job """
        command = self.start_command.format(job.name)
        run_command(command, job.sudo, *args, **kwargs)

    def stop(self, job, *args, **kwargs):
        """Stop the job"""
        command = self.stop_command.format(job.name)
        run_command(command, job.sudo, *args, **kwargs)

    def restart(self, job, *args, **kwargs):
        """Restart the job"""
        command = self.restart_command.format(job.name)
        run_command(command, job.sudo, *args, **kwargs)

    def get_status(self, job, *args, **kwargs):
        """Restart the job"""
        command = self.status_command.format(job.name)
        return run_command(command, job.sudo, *args, **kwargs).strip()

    def get_pid(self, job):
        return int(self.get_status(job).split()[-1])

    def is_running(self, job):
        return 'running' in self.get_status(job)

    def ensure_stopped(self, job, *args, **kwargs):
        """Stops the job if it is running and then checks
        that it effectively was stopped
        """
        self.stop(job, *args, **kwargs)
        count = 0
        while is_process_running(job.name) and count < 5:
            time.sleep(1)
            count += 1
        if is_process_running(job.name):
            raise RuntimeError("Process could not be stopped")

    def set_environment_variable(self, key, value, global_scope=True,
                                 sudo=False):
        set_command = self.set_env_command.format(key, value)
        self._manage_environment_variable(set_command, global_scope, sudo)

    def unset_environment_variable(self, key, global_scope=True, sudo=False):
        unset_command = self.unset_env_command.format(key)
        self._manage_environment_variable(unset_command, global_scope, sudo)

    def get_environment_variable(self, key, global_scope=False, sudo=False):
        show_command = self.get_env_command.format(key)
        return self._manage_environment_variable(show_command, global_scope,
                                                 sudo)

    def _manage_environment_variable(self, command, global_scope, sudo):
        if global_scope:
            command += ' --global'
        return run_command(command, sudo)


class Upstart(InitSystem):
    """This class contains the implementation for upstart managed jobs """

    def __init__(self):
        super().__init__(start_command='/sbin/initctl start {}',
                         stop_command='/sbin/initctl stop {}',
                         status_command='/sbin/initctl status {}',
                         restart_command='/sbin/initctl restart {}',
                         set_env_command='/sbin/initctl set-env {}={}',
                         unset_env_command='/sbin/initctl unset-env {}',
                         get_env_command='/sbin/initctl get-env {}={}')

    def wait_for_job_spawned(self, job_name, timeout=10):
        """Wait until the named job is in spawned state."""
        spawned = wait_until(
            lambda: 'start/spawned' in self.get_status(job_name),
            timeout=timeout)
        if not spawned:
            raise RuntimeError(
                'Job {} not in start/spawned state within timeout {}'.format(
                    job_name, timeout))

    def get_env_variable(self, key, global_scope=False):
        try:
            vars = super().get_environment_variable(key=key,
                                                    global_scope=global_scope)
        except subprocess.CalledProcessError:
            return None
        else:
            return vars.rstrip()


class SystemD(InitSystem):
    """This class contains the implementation for systemd managed jobs """

    def __init__(self):
        super().__init__(start_command='systemctl start {}',
                         stop_command='systemctl stop {}',
                         status_command='systemctl status {}',
                         restart_command='systemctl restart {}',
                         set_env_command='systemctl set-environment {}={}',
                         unset_env_command='systemctl unset-environment {}',
                         get_env_command='systemctl show-environment')

    def is_running(self, *args, **kwargs):
        """Return True if the job is running. Otherwise, False."""
        line = run_command(self.status_command, *args, **kwargs)
        return 'active' in line

    def wait_for_job_spawned(self, job_name, timeout=10):
        spawned = wait_until(self.is_running, job_name, timeout=timeout)
        if not spawned:
            raise RuntimeError(
                'Job {} not in start/spwaned state within timeout {}'.format(
                    job_name, timeout))

    def get_env_variable(self, key, global_scope=False, sudo=True):
        vars = self._manage_environment_variable(self.get_env_command,
                                                 global_scope, sudo)
        for var in vars.split('\n'):
            keyval = var.split('=')
            if len(keyval) == 2 and keyval[0] == key:
                return keyval[1]
        return None

    def _manage_environment_variable(self, command, global_scope, sudo=True):
        return super()._manage_environment_variable(command, global_scope,
                                                    True)
