# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import time

import autopilot
from autopilot.input import Keyboard
from autopilot.introspection.utilities import sort_by_keys

from ubuntu_system_tests.helpers.addressbook._cpo import ContactViewPage
from ubuntu_system_tests.helpers.init_systems import get_init_system
from ubuntu_system_tests.helpers.ubuntu_keyboard import get_keyboard
from ubuntu_system_tests.helpers.ubuntu_keyboard.job import MaliitServerJob
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._mainview import MainView as BaseMainView  # NOQA
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import (
    CustomProxyObjectBase,
    get_pointing_device,
    )

from ubuntu_system_tests.helpers.ubuntuuitools._cpo._textfield import TextField  # NOQA

from autopilot.exceptions import StateNotFoundError
from urllib.parse import urlparse

from ubuntu_system_tests.helpers.autopilot import (
    patched_select_single, patched_wait_select_single, validate_dbus_object)
from ubuntu_system_tests.helpers.autopilot.list import KEY_ORDER_BY_Y
from ubuntu_system_tests.helpers.messaging import (
    MEDIA_TYPE_PICTURE, MEDIA_TYPE_VIDEO, MEDIA_TYPE_CONTACT)
from ubuntu_system_tests.helpers.system_settings import SIM_1, SIM_2

logger = logging.getLogger(__name__)

APP_PATH_ROOT = b'MessagingApp'
DEFAULT_SIM_MSG = 'Change all Messaging associations to SIM {sim}?'


class MessagingApp(CustomProxyObjectBase):
    """Autopilot helper object for messaging app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, APP_PATH_ROOT)

    @property
    def main_view(self):
        return patched_wait_select_single(
            root=self, class_object=MainView, class_name='MainView')


class MainView(BaseMainView):
    def __init__(self, *args):
        super(MainView, self).__init__(*args)
        self.pointing_device = get_pointing_device()
        self.keyboard = Keyboard.create()
        self.logger = logging.getLogger(__name__)

    def get_pagestack(self):
        """Return PageStack with objectName mainStack"""
        return self.select_single("PageStack", objectName="mainStack")

    def get_thread_from_number(self, phone_number):
        """Return thread from number

        :parameter phone_number: the phone_number of message thread

        """
        return self.wait_select_single(ThreadDelegate,
                                       phoneNumber=phone_number)

    def get_message(self, text):
        """Return message from text
        :parameter text: the text of the message
        """
        time.sleep(2)  # message is not always found on slow emulator
        for message in self.select_many(MessageDelegateFactory):
            for item in message.select_many('MessageBubble'):
                if self.is_message_in_item(message, item):
                        return message
        raise EmulatorException('Could not find message with the text '
                                '{}'.format(text))

    @staticmethod
    def is_message_in_item(message, item):
        return ('messageText', message) in item.get_properties().items()

    def click_header_action(self, action):
        """Click the action 'action' on the header"""
        action = self.wait_select_single(objectName='%s_button' % action)
        self.pointing_device.click_object(action)

    def get_newmessage_textfield(self):
        """Return TextField with objectName newPhoneNumberField"""

        return self.select_single(objectName="contactSearchInput")

    def get_multiple_selection_list_view(self):
        """Return MultipleSelectionListView from the messages page"""

        page = self.get_messages_page()
        return page.select_single('MessagesListView')

    def get_newmessage_textarea(self):
        """Return TextArea with blank objectName"""

        return self.get_messages_page().select_single(
            objectName='messageTextArea')

    def get_send_button(self):
        """Return Button with text Send"""

        return self.get_messages_page().select_single(objectName='sendButton')

    def get_contact_list_view(self):
        """Returns the ContactListView object"""
        return self.select_single(
            objectName='newRecipientList'
        )

    def get_dialog_buttons(self, visible=True):
        """Return DialogButtons

        :parameter visible: the visible state of the dialog button
        """

        if visible:
            return self.wait_select_single('DialogButtons', visible=True)
        else:
            return self.select_many('DialogButtons', visible=False)

    def long_press(self, obj):
        """long press on object because press_duration is not honored on touch
        see bug #1268782

        :parameter obj: the object to long press on
        """

        self.pointing_device.move_to_object(obj)
        self.pointing_device.press()
        time.sleep(3)
        self.pointing_device.release()

    def type_contact_phone_num(self, num_or_contact):
        """Select and type phone number or contact

        :parameter num_or_contact: number or contact to type
        """

        text_entry = self.get_newmessage_textfield()
        self.pointing_device.click_object(text_entry)
        text_entry.focus.wait_for(True)
        time.sleep(.3)
        text_entry.write(str(num_or_contact))
        self.keyboard.press_and_release("Enter")
        self.logger.info(
            'typed "{}" expected "{}"'.format(
                self.get_newmessage_textfield().text, num_or_contact))

    def click_send_button(self):
        """Click the send button on the message page"""

        button = self.get_send_button()
        button.enabled.wait_for(True)
        self.pointing_device.click_object(button)
        button.enabled.wait_for(False)

    def start_new_message(self):
        """Reveal the bottom edge page to start composing a new message"""
        self.reveal_bottom_edge_page()

    def enable_messages_selection_mode(self):
        """Enable the selection mode on the messages page by pressing and
        holding the first item"""
        message = self.wait_select_single("MessageDelegateFactory",
                                          objectName="message0")
        self.long_press(message)

        # FIXME: there should be a better way to detect when the popover is
        # gone
        time.sleep(2)

        # and now click the message again to start with it unselected
        self.pointing_device.click_object(message)

    def enable_threads_selection_mode(self):
        """Enable the selection mode on the threads page by pressing and
        holding the first item"""
        # FIXME: there is probably a better way to do this
        thread = self.select_many("ThreadDelegate")[0]
        self.long_press(thread)
        # and now click to unselect it
        self.pointing_device.click_object(thread)

    def close_osk(self):
        """Swipe down to close on-screen keyboard"""
        if get_init_system().is_running(MaliitServerJob()):
            get_keyboard()._keyboard.dismiss()

    def click_add_contact_icon(self):
        """Click the add contact icon"""
        header = self.get_header()
        header.click_action_button("contactList")

    def click_add_button(self):
        """Click add button from toolbar on messages page"""
        header = self.get_header()
        header.click_action_button("addContactAction")

    def click_call_button(self):
        """Click call button from toolbar on messages page"""
        header = self.get_header()
        header.click_action_button("contactCallAction")

    def click_threads_header_delete(self):
        """Click the header action 'Delete' on Messages view"""
        self.click_header_action('selectionModeDeleteAction')

    def click_threads_header_cancel(self):
        """Click the header action 'Cancel' on Messages view"""
        self.get_header().click_custom_back_button()

    def click_threads_header_settings(self):
        """Click the header action 'Settings' on main view"""
        self.click_header_action('settingsAction')

    def click_messages_header_delete(self):
        """Click the header action 'Delete' on Messages view"""
        self.click_header_action('selectionModeDeleteAction')

    def click_messages_header_cancel(self):
        """Click the header action 'Cancel' on Messages view"""
        self.get_header().click_custom_back_button()

    def delete_thread(self, phone_number):
        """Delete thread containing specified phone number.

        :parameter phone_number: phone number of thread to delete
        :parameter direction: right or left, the direction to swipe to delete
        """
        thread = self.get_thread_from_number(phone_number)
        thread.swipe_to_delete()
        # wait for the animation to end
        time.sleep(1)
        thread.confirm_removal()

    def send_message(self, numbers, message, check_sent=True):
        """Write a new message and send it.

        :param numbers: phone numbers of contacts to send message to.
        :param message: the message to be sent.

        """
        self.start_new_message()
        # type address number
        for number in numbers:
            self.type_contact_phone_num(number)
            self.keyboard.press_and_release("Enter")

        self.type_message(message)
        old_message_count = self.get_multiple_selection_list_view().count
        self.click_send_button()

        thread_bubble = None
        if (check_sent):
            self.get_multiple_selection_list_view().count.wait_for(
                old_message_count + 1)
            thread_bubble = self.get_message(message)

        return thread_bubble

    def _get_page(self, page_type, page_name):
        page = self.wait_select_single(
            page_type, objectName=page_name, active=True)
        return page

    def get_new_recipient_page(self):
        return self._get_page(NewRecipientPage, 'newRecipientPage')

    def open_settings_page(self):
        self.click_threads_header_settings()
        settings_page = self.wait_select_single(SettingsPage)
        settings_page.active.wait_for(True)
        return settings_page

    def get_swipe_item_demo(self):
        return self.wait_select_single(
            'SwipeItemDemo', objectName='swipeItemDemo', parentActive=True)

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MainView')

    def get_main_page(self):
        """Override base class method to use patched select method."""
        return patched_wait_select_single(
            root=self, class_object=MainPage, class_name='MainPage')

    def get_main_title(self):
        """ Retrieve the app title """
        # The title is hardcoded bacause it cannot be found in any attribute
        #  of the messaging app object
        return 'Messaging'

    @autopilot.logging.log_action(logger.info)
    def reveal_bottom_edge_page(self):
        """Bring the bottom edge page to top of screen."""
        start_x = (self.globalRect.x +
                   (self.globalRect.width * 0.5))
        start_y = (self.globalRect.y + self.height)
        stop_y = start_y - (self.height * 0.7)
        self.pointing_device.drag(start_x, start_y,
                                  start_x, stop_y)
        self.composingNewMessage.wait_for(True)

    @autopilot.logging.log_action(logger.info)
    def type_contact_name_and_select_num(self, name, number):
        self.type_contact_name(name)
        # Wait for the list to settle
        time.sleep(1)
        # Now select the right number from the list
        self.pointing_device.click_object(
            self.get_mobile_number_label(number)
        )

    def type_contact_name(self, name):
        text_entry = self.get_newmessage_multirecipientinput()
        text_entry.enter_text(name)

    def get_mobile_number_label(self, number):
        """Return first matching label displayed."""
        return self.wait_select_many(
            'UCLabel', ap_result_sort_keys=KEY_ORDER_BY_Y,
            text='Mobile {}'.format(number))[0]

    def detect_select_sim_prompt(self):
        """Detect if the select SIM dialog is shown."""
        try:
            self.wait_select_single('Dialog')
            return True
        except StateNotFoundError:
            return False

    def get_sim_label(self, sim_number):
        return self.wait_select_single(
            'UCLabel', text='SIM {}'.format(sim_number)
        )

    def get_thread_list(self):
        return self.select_single(objectName='threadList')

    def _get_attachment_button(self):
        """Returns the 'attach media' button."""
        return self.get_messages_page().select_single(
            objectName='attachButton')

    def attach_media(self, media_type):
        """Press the attach button on composing message and select the required
        media type. Return the content picker dialog.
        :param media_type: The type of media required, picture, video or
                           contact.
        :return: Content picker dialog proxy object.
        """
        self.pointing_device.click_object(self._get_attachment_button())
        panel = self._get_attachment_panel()
        panel.wait_until_not_moving()
        panel.choose_media_type(media_type)
        return self._get_content_picker()

    def _get_attachment_panel(self):
        """Return proxy object for attachment panel."""
        return self.wait_select_single(AttachmentPanel)

    def _get_content_picker(self):
        return self.wait_select_single(ContentPeerPicker13, visible=True)

    def type_message(self, message):
        """Select and type message in new message text area in messages page

        :parameter message: the message to type
        """
        text_entry = self.get_newmessage_textarea()
        text_entry.write(message)
        if text_entry.text != message:
            self.logger.info('typed: "{}" expected: "{}"'.format(
                text_entry.text, message
            ))

    def _get_thread_delegates(self):
        """Get all thread delegates in main screen."""
        return self.get_thread_list().select_many('ThreadDelegate')

    def get_visible_threads_count(self):
        """Return the number of visible threads."""
        return len([t for t in self._get_thread_delegates() if t.height != 0])

    def _is_search_mode(self):
        """Return bool indicating if the App header is in search mode."""
        return self.get_main_page().state == 'search'

    def _get_search_box(self):
        """Return the message searchbox."""
        """Return the contacts search box."""
        return self.select_single(TextField, objectName='searchField')

    def search(self, text):
        """Tap the search icon and write the requestion text."""
        if not self._is_search_mode():
            self.click_header_action('searchAction')
        search_box = self._get_search_box()
        search_box.clear()
        search_box.write(text)

    def clear_search(self):
        """Clear the search box."""
        if self._is_search_mode():
            self._get_search_box().clear()

    def is_thread_for_number_visible(self, number):
        """
        Return bool indicating the thread for `number` is visible on screen.
        """
        thread = self.get_thread_from_number(number)
        return thread.height > 0

    def _get_popup_paste_button(self):
        """
        Return the 'paste' button on the new message field' contextual popup.
        """
        paste_button = self.wait_select_single(iconName='edit-paste')
        paste_button.opacity.wait_for(1.0)
        return paste_button

    def paste_text_in_message_field(self):
        """
        Start new message and paste the copied text into the new message field.
        """
        self.start_new_message()
        text_entry = self.get_newmessage_textarea()
        # Double click the empty field for the copy popover to appear
        self.pointing_device.click_object(text_entry)
        self.pointing_device.click_object(text_entry)
        # Click the paste button on the popup
        self.pointing_device.click_object(self._get_popup_paste_button())
        return text_entry

    def _get_sim_confirmation_dialog(self, sim):
        """Return default confirmation dialog for specified SIM."""
        return patched_select_single(
            self, class_object=Dialog, class_name='Dialog',
            text=DEFAULT_SIM_MSG.format(sim=sim))

    def get_sim1_default_confirmation_dialog(self):
        """Return SIM 1 default confirmation dialog."""
        return self._get_sim_confirmation_dialog(SIM_1)

    def get_sim2_default_confirmation_dialog(self):
        """Return SIM 2 default confirmation dialog."""
        return self._get_sim_confirmation_dialog(SIM_2)

    def _get_sim_selector(self):
        """Return selector object which controls the SIM labels."""
        return self.select_single(objectName='headerSectionsItem')

    def get_selected_sim(self):
        """Return label of currently selected SIM either 'SIM 1' or 'SIM 2'."""
        index = self._get_sim_selector().selectedIndex
        button = self.select_single(
            objectName='section_button_label_{}'.format(index))
        return button.text

    @autopilot.logging.log_action(logger.info)
    def select_sim(self, sim):
        """Select the SIM to use using selection tabs.
        :param sim: SIM to use, either '1' or '2'.
        """
        selector = self._get_sim_selector()
        label = selector.select_single(
            text='SIM {}'.format(sim))
        label_index = int(label.objectName[-1])
        self.pointing_device.click_object(label)
        selector.selectedIndex.wait_for(label_index)

    def _get_back_button(self):
        """Return the button to return to main screen."""
        return self.select_single(objectName='apl_back_action_button')

    @autopilot.logging.log_action(logger.info)
    def go_back(self):
        """Press back button on header."""
        self.pointing_device.click_object(self._get_back_button())

    def delete_message(self, text):
        """Delete message and wait until the message is deleted."""
        message = self.get_message(text)
        super().delete_message(text)
        message.wait_until_destroyed()

    def get_newmessage_multirecipientinput(self):
        """Return MultiRecipientInput from the messages page"""
        return self.select_single(
            'MultiRecipientInput', objectName='multiRecipient', visible=True)

    def get_messages_page(self):
        """Return messages with objectName messagesPage"""
        return self.wait_select_single(
            'Messages', objectName='messagesPage', visible=True)


class MainPage(MainView):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MainPage')

    def get_thread_count(self):
        """Return the number of message threads."""
        return self.select_single(
            'MultipleSelectionListView', objectName='threadList').count

    def open_thread(self, participants):
        """Opens the thread with the given participants."""
        """FIXME: right now it only supports 1-1 conversations"""
        thread = self.select_single(
            ThreadDelegate, objectName='thread{}'.format(participants))
        self.pointing_device.click_object(thread)
        root = self.get_root_instance()
        return root.wait_select_single(Messages,
                                       firstParticipantId=participants)


class MultiRecipientInput(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MultiRecipientInput')

    def enter_text(self, text):
        contact_search_input = self.select_single(
            'TextField', objectName='contactSearchInput'
        )
        contact_search_input.write(text)


class ListItemWithActions(CustomProxyObjectBase):

    def confirm_removal(self):
        deleteButton = self.wait_select_single(name='delete')
        self.pointing_device.click_object(deleteButton)

    def swipe_to_delete(self):
        x, y, width, height = self.globalRect
        start_x = x + (width * 0.2)
        stop_x = x + (width * 0.8)
        start_y = stop_y = y + (height // 2)

        self.pointing_device.drag(start_x, start_y, stop_x, stop_y)

    def active_action(self, action_index):
        action_margin = ((self.actionWidth / 5) * 2)
        x_offset = ((self.actionWidth + action_margin) * action_index)
        x_offset += self.actionThreshold

        x, y, width, height = self.globalRect
        start_x = x + (width * 0.5)
        stop_x = start_x - x_offset
        start_y = stop_y = y + (height // 2)

        self.pointing_device.drag(start_x, start_y, stop_x, stop_y)


class ThreadDelegate(ListItemWithActions):
    """Autopilot helper for ThreadDelegate."""


class MessageDelegateFactory(ListItemWithActions):
    """Class to get all the messages from the app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MessageDelegateFactory')

    def get_mms_content_file_path(self):
        """Return MMSBase for MMS message."""
        file_url = self.select_single(MMSBase).get_media_file_url()
        return urlparse(file_url).path

    def get_message_bubble(self):
        return self.wait_select_single(MessageBubble, visible=True)


class MessageBubble(CustomProxyObjectBase):
    """Class to represent the message bubble for a displayed message."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MessageBubble')

    def wait_until_sent(self, timeout):
        sent_time = self.select_single(objectName='messageDate')
        sent_time.visible.wait_for(True, timeout=timeout)


class MMSBase(CustomProxyObjectBase):
    """Class to examine MMS content."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MMSBase')

    def get_media_file_url(self):
        """Return the file url of the media content."""
        image = self.get_media_container().select_single('QQuickImage')
        return image.get_properties()['source']

    def get_media_container(self):
        """Return container object for media content."""
        return self.select_single('UCUbuntuShape', visible='True')


class Messages(CustomProxyObjectBase):
    """Class to examine content of messages displayed in editor."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'Messages')

    def get_attachments(self):
        """Return list of displayed attachment objects."""
        attachment_types = [
            'ThumbnailImage', 'ThumbnailVideo', 'ThumbnailContact']
        attachments = []
        for attachment_type in attachment_types:
            for attachment in self.select_many(attachment_type, visible=True):
                if ('filePath' in attachment.get_properties() and
                        len(attachment.filePath) > 0):
                    attachments.append(attachment)
        return sort_by_keys(attachments, KEY_ORDER_BY_Y)

    def get_attachment_count(self):
        """Return a count of the displayed attachments in message editor."""
        return len(self.get_attachments())

    def get_matching_messages(self, text, sender):
        """Return matching messages for specified sender in thread."""
        return self.select_many(
            'RegularMessageDelegate', textMessage=text, senderId=sender,
            visible=True)

    def is_message_displayed(self, text, phone_number):
        """
        Check if message with phone_number and text is displayed in thread list

        :param text: The body of the message we are expecting
        :param number: The sending phone number
        :return: True if received, otherwise False

        """
        return bool(self.get_matching_messages(text, phone_number))

    def how_many_messages_displayed(self, text, phone_number):
        """
        Count how many messages are displayed with specified text and
        phone number in thread list.

        :param text: The body of the message we are expecting
        :param number: The sending phone number
        :return: Number of matching messages

        """
        return len(self.get_matching_messages(text, phone_number))

    @property
    def title(self):
        """Return title of currently displayed messages."""
        return self.select_single('PageHeader').title

    def get_messages_count(self):
        """Return the number of meesages."""
        return self.get_list_view().count

    def select_messages(self, *indexes):
        """Select messages.

        :param indexes: The indexes of the messages to select. The most
            recently received message has the 0 index, and the oldest message
            has the higher index.

        """
        for index in indexes:
            message_delegate = self._get_message_delegate(index)
            self.pointing_device.click_object(message_delegate)

    def scroll_list(self):
        list_view = self.get_list_view()
        x, y, width, height = list_view.globalRect
        start_x = stop_x = x + (width / 2)
        start_y = y + 10
        stop_y = y + height - 10
        self.pointing_device.drag(start_x, start_y, stop_x, stop_y)

    def _get_message_delegate(self, index):
        return self.wait_select_single(
            'MessageDelegateFactory', objectName='message{}'.format(index))

    def _long_press_to_select_message(self, message):
        self.pointing_device.move_to_object(message)
        self.pointing_device.press()
        self.selectionMode.wait_for(True)
        self.pointing_device.release()

    def get_messages(self):
        """Return a list with the information of the messages.

        Each item of the returned list is a tuple of (date, text).

        """
        messages = []
        for index in range(self.get_messages_count()):
            message_delegate = self._get_message_delegate(index)
            date = message_delegate.select_single(
                objectName='messageDate').text
            text = message_delegate.select_single(
                objectName='messageText').text
            messages.append((date, text))
        return messages

    def get_text_area_text(self):
        return self.wait_select_single(objectName='messageTextArea').text

    def get_list_view(self):
        """Returns the messages list view"""
        return self.wait_select_single(
            'MessagesListView', objectName='messageList')


class AttachmentPanel(CustomProxyObjectBase):
    """Class to select attachment type from panel."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'AttachmentPanel')

    def _get_picture_button(self):
        return self._get_button(MEDIA_TYPE_PICTURE)

    def _get_video_button(self):
        return self._get_button(MEDIA_TYPE_VIDEO)

    def _get_contact_button(self):
        return self._get_button(MEDIA_TYPE_CONTACT)

    def _get_button(self, media_type):
        button_name = media_type + 'Button'
        return self.select_single(objectName=button_name)

    def choose_media_type(self, media_type):
        """Select specified media type.
        :param media_type: Either picture, video or contact.
        """
        self.pointing_device.click_object(self._get_button(media_type))

    def choose_picture(self):
        """Select picture type attachment."""
        self.pointing_device.click_object(self._get_picture_button())

    def choose_video(self):
        """Select video type attachment."""
        self.pointing_device.click_object(self._get_video_button())

    def choose_contact(self):
        """Select contact type attachment."""
        self.pointing_device.click_object(self._get_contact_button())


class ContentPeerPicker13(CustomProxyObjectBase):
    """Class to pick a content source application."""

    def pick_app(self, app_name):
        """
        Taps the requested app's icon from the content picker screen
        :param app_name: name of the app to tap

        """
        app_icon = self.wait_select_many(
            ap_result_sort_keys=KEY_ORDER_BY_Y, text=app_name)[0].get_parent()
        self.pointing_device.click_object(app_icon)


class Dialog(CustomProxyObjectBase):
    """Class to represent default SIM selection dialog."""

    @autopilot.logging.log_action(logger.info)
    def confirm(self):
        self.pointing_device.click_object(self._get_yes_button())
        self.wait_until_destroyed()

    @autopilot.logging.log_action(logger.info)
    def deny(self):
        self.pointing_device.click_object(self._get_no_button())
        self.wait_until_destroyed()

    def get_dont_ask_again(self):
        return self._get_checkbox().checked

    def set_dont_ask_again(self, state):
        checkbox = self._get_checkbox()
        if checkbox.checked is not state:
            checkbox.change_state()

    def _get_checkbox(self):
        return self.select_single('CheckBox')

    def _get_yes_button(self):
        return self.select_single('Button', text='Yes')

    def _get_no_button(self):
        return self.select_single('Button', text='No')


class EmulatorException(Exception):
    """Exception raised when there is an error with the emulator."""


class NewRecipientPage(CustomProxyObjectBase):
    """Autopilot custom proxy object for NewRecipientPage components."""

    def _click_button(self, button):
        """Generic way to click a button"""
        self.visible.wait_for(True)
        button.visible.wait_for(True)
        self.pointing_device.click_object(button)
        return button

    def open_contact(self, index):
        contact_delegate = self._get_contact_delegate(index)
        self.pointing_device.click_object(contact_delegate)
        return self.get_root_instance().select_single(
            MessagingContactViewPage, objectName='contactViewPage')

    def _get_contact_delegate(self, index):
        contact_delegates = self._get_sorted_contact_delegates()
        return contact_delegates[index]

    def _get_sorted_contact_delegates(self):
        contact_delegates = self.select_many('ContactDelegate', visible=True)
        return sorted(
            contact_delegates, key=lambda delegate: delegate.globalRect.y)


class MessagingContactViewPage(ContactViewPage):
    """Autopilot custom proxy object for MessagingContactViewPage component."""

    def message_phone(self, index):
        phone_group = self.select_single(
            'ContactDetailGroupWithTypeView',
            objectName='phones')

        call_buttons = phone_group.select_many(
            objectName="message-contact")
        self.pointing_device.click_object(call_buttons[index])


class SettingsPage(CustomProxyObjectBase):
    """Autopilot helper for the settings page"""

    def get_mms_group_chat(self):
        return self.wait_select_single(objectName="mmsGroupChatEnabled")

    def toggle_mms_group_chat(self):
        self.pointing_device.click_object(self.get_mms_group_chat())
