# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from ubuntu_system_tests.helpers import context
from ubuntu_system_tests.helpers import ofono
from ubuntu_system_tests.helpers.observer import Observer
from ubuntu_system_tests.helpers.threads import StoppingThread

logger = logging.getLogger(__name__)


class SimPinObserver(Observer):
    """Observer object to dismiss sim pin entry screens."""

    def __init__(self, stop, result, sims):
        """
        :param stop: Event object to indicate when to stop observing.
        :param result: Event object to indicate when required sim pin screens
                       have been dismissed.
        :param sims: List of locked sim indexes that require sim pin
                     notification.
        """
        super().__init__(stop, result)
        self.sims = sims

    def start(self):
        """Start the observer."""
        super().start()
        logger.info('{} sim(s) locked.'.format(len(self.sims)))
        self.dismiss_sim_lock_screens()

    def _cleanup(self):
        """Cleanup actions before stopping."""
        if not self.result.is_set():
            logger.info(
                '{c} out of {e} expected sim pin entry dialogs '
                'diaplayed.'.format(c=self.count, e=len(self.sims)))

    def dismiss_sim_lock_screens(self):
        """Dismiss either sim pin or puk code screens."""
        self.count = 0
        for sim in self.sims:
            if self._stop.is_set():
                # Stop request is set, so just exit immediately.
                return
            self.dismiss_pin_type(sim)
        self.result.set()
        self.stop()

    def dismiss_pin_type(self, sim):
        """Dismiss requird pin or puk dialogs for sim."""
        n_list = context.shared.unity.main_window.get_notifications_list()
        if ofono.is_pin_type_pin(sim):
            logger.warning('sim {s} is pin locked.'.format(s=sim))
            self.dismiss_sim_pin_screens(sim, n_list)
        elif ofono.is_pin_type_puk(sim):
            logger.warning('sim {s} is puk locked.'.format(s=sim))
            self.dismiss_sim_puk_screens(sim, n_list)
        else:
            raise RuntimeError(
                'Unsupported sim lock type for sim {s}.'.format(s=sim))

    def dismiss_sim_puk_screens(self, sim, n_list):
        """Dismiss each required sim puk screen."""
        # Firstly dismiss the puk popup warning notification
        popup = context.shared.unity.main_window.get_info_popup()
        if popup:
            popup.dismiss()
        # Wait for sim puk notification to be displayed.
        notification = n_list.get_sim_puk_unlock_notification(
            sim, timeout=30, stop_event=self._stop)
        self.dimiss_notification(notification)

    def dismiss_sim_pin_screens(self, sim, n_list):
        """Dismiss each required sim pin screen."""
        # Wait for sim pin notification to be displayed.
        notification = n_list.get_sim_pin_unlock_notification(
            sim, timeout=30, stop_event=self._stop)
        self.dimiss_notification(notification)

    def dimiss_notification(self, notification):
        """Dismiss specified notification."""
        if notification:
            # notification is found, so dimiss it and move onto next
            notification.get_lockscreen().dismiss()
            self.count += 1


def _start_sim_pin_observer(stop, result, sims):
    """
    Start a sim pin notification observer.

    :param stop: Event object to signal when to stop observing.
    :param result: Event object to signal when required number of notifications
                   are observed.
    :param sims: List of locked sims requiring pin input.

    """
    observer = SimPinObserver(stop, result, sims)
    observer.start()


def start_sim_pin_observer(result, sims):
    """
    Start an observer thread to dismiss sim pin notifications.

    :param result: Event object to signal when required number of notifications
                   are observed.
    :param sims: List of locked sims requiring pin input.
    :return: StoppingThread object which should be stopped by caller when
             required.

    """
    launch_thread = StoppingThread(
        target=_start_sim_pin_observer, args=[result, sims])
    launch_thread.start()
    return launch_thread
