# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import signal
import subprocess

import psutil
from retrying import retry

logger = logging.getLogger(__name__)


class PidNotFoundError(RuntimeError):
    """Pid not found error."""
    def __init__(self, process_name, *args, **kwargs):
        message = 'No running process found for {}'.format(process_name)
        super().__init__(message, *args, **kwargs)


def is_process_running(pname, retries=True):
    if retries:
        try:
            return bool(get_process_id(pname))
        except PidNotFoundError:
            return False
    else:
        try:
            return bool(_get_process_id_no_retries(pname))
        except PidNotFoundError:
            return False


def _get_process_id_no_retries(pname):
    """
    Return PID for the requested process.

    :param pname: name of the process to get PID of.
    """
    try:
        pids = subprocess.check_output(
            ['pidof', pname], universal_newlines=True).split()
    except subprocess.CalledProcessError:
        raise PidNotFoundError(pname)
    for pid in pids:
        process = psutil.Process(int(pid))
        if process.status() == psutil.STATUS_RUNNING \
                or process.status() == psutil.STATUS_SLEEPING:
            return process.pid
    raise PidNotFoundError(pname)


@retry(stop_max_attempt_number=5,
       wait_fixed=1000,
       retry_on_exception=lambda exception: isinstance(
           exception, PidNotFoundError))
def get_process_id(pname):
    """
    Return PID for the requested process.

    :param pname: name of the process to get PID of.
    """
    return _get_process_id_no_retries(pname)


def stop_process(pname):
    if is_process_running(pname):
        [process.terminate() for process in psutil.process_iter() if
         process.name() == pname]


def start_application(app_name, *args):
    subprocess.call(['ubuntu-app-launch', app_name] + list(args))


def ensure_application_closed(app_id):
    subprocess.call(['ubuntu-app-stop', app_id])


def ensure_all_apps_closed():
    open_apps = list_open_applications()
    for app in open_apps:
        ensure_application_closed(app)


def list_open_applications():
    apps = subprocess.check_output(['ubuntu-app-list'],
                                   shell=True,
                                   universal_newlines=True)
    return apps.split()


def is_qmlscene_running_with_qmlfile(qmlfile):
    """ Indicate if the qmlscene is running the qml file qmlfile """
    pline = subprocess.check_output('ps -ef | grep qmlscene', shell=True,
                                    universal_newlines=True)
    return qmlfile in pline


def get_qmlscene_process_id(qmlfile):
    """ Retrieve the process id for the qmlscene running the file qmlfile """
    try:
        output = subprocess.check_output(['pgrep', 'qmlscene'])
    except subprocess.CalledProcessError:
        raise PidNotFoundError('qmlscene')
    else:
        for pid in output.decode().split():
            if qmlfile in get_process_args(pid):
                return int(pid)
    raise PidNotFoundError('qmlscene {}'.format(qmlfile))


def stop_qmlscene_process(qmlfile):
    """ Stop the qmlscene process running the file qmlfile """
    if is_qmlscene_running_with_qmlfile(qmlfile):
        pid = get_qmlscene_process_id(qmlfile)
        os.kill(pid, signal.SIGKILL)


def get_process_args(pid):
    """ Retrieve the command line used to start the process """
    return subprocess.check_output('xargs -0 < /proc/{}/cmdline'.format(pid),
                                   shell=True, universal_newlines=True)


def is_testability_enabled_for_process(pid):
    """Return True if testability is enabled for specified process."""
    proc_env = '/proc/{pid}/environ'.format(pid=pid)
    command = ['cat', proc_env]
    try:
        output = subprocess.check_output(
            command,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
    except subprocess.CalledProcessError as e:
        e.args += ('Failed to get environment for pid {}: {}.'.format(
            pid, e.output),)
        raise
    return output.find('QT_LOAD_TESTABILITY=1') >= 0
