# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014, 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from time import sleep

import autopilot.logging
import autopilot.exceptions as exceptions
from autopilot import introspection
from autopilot.introspection import is_element

from ubuntu_system_tests.helpers.ubuntuuitools._cpo._flickable import QQuickFlickable  # NOQA
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._mainview import MainView
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._checkbox import CheckBox
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA

from ubuntu_system_tests.helpers.autopilot import validate_dbus_object
from ubuntu_system_tests.helpers.autopilot.list import KEY_ORDER_BY_Y
from ubuntu_system_tests.helpers.brightness._cpo import BrightnessPageBase

logger = logging.getLogger(__name__)

APP_PATH_ROOT = b'ubuntu-system-settings'


class SystemSettings(CustomProxyObjectBase):

    """Helper class for System Settings application"""

    @classmethod
    def validate_dbus_object(cls, path, state):
        name = introspection.get_classname_from_path(path)
        return ((name == b'SystemSettings' and
                 state['applicationName'][1] == 'SystemSettings') or
                (name == b'ubuntu-system-settings' and
                 state['applicationName'][1] == 'ubuntu-system-settings'))

    @property
    def main_view(self):
        """Return main view"""
        return self.select_single(objectName='systemSettingsMainView')


class SystemSettingsMainWindow(MainView):
    """Autopilot helper for the Main Window."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MainWindow',
            objectName='systemSettingsMainView')

    @autopilot.logging.log_action(logger.debug)
    def click_item(self, object_name):
        """Click a system settings item.

        :param object_name: The objectName property of the item to click.

        """
        item = self.select_single(objectName=object_name)
        item.swipe_into_view()
        self.pointing_device.click_object(item)

    def go_to_reset_phone(self):
        return self._go_to_page('entryComponent-reset', 'resetPage')

    def go_to_language_page(self):
        return self._go_to_page('entryComponent-language', 'languagePage')

    def go_to_wifi_page(self):
        return self._go_to_page('entryComponent-wifi', 'wifiPage')

    def go_to_cellular_page(self):
        return self._go_to_page('entryComponent-cellular', 'cellularPage')

    def go_to_hotspot_page(self):
        return self._go_to_page('entryComponent-hotspot', 'hotspotPage')

    def go_to_bluetooth_page(self):
        return self._go_to_page('entryComponent-bluetooth', 'bluetoothPage')

    def go_to_phone_page(self):
        return self._go_to_page('entryComponent-phone', 'phonePage')

    def go_to_about_page(self):
        return self._go_to_page('entryComponent-about', 'aboutPage')

    def go_to_sound_page(self):
        return self._go_to_page('entryComponent-sound', 'soundPage')

    def go_to_security_page(self):
        return self._go_to_page('entryComponent-security-privacy',
                                'securityPrivacyPage')

    def go_to_notification_page(self):
        return self._go_to_page('entryComponent-notifications',
                                'systemNotificationsPage')

    def go_to_datetime_page(self):
        return self._go_to_page('entryComponent-time-date', 'timeDatePage')

    def go_to_vpn_page(self):
        return self._go_to_page('entryComponent-vpn', 'vpnPage')

    @autopilot.logging.log_action(logger.debug)
    def _go_to_page(self, item_object_name, page_object_name):
        self.click_item(item_object_name)
        page = self.wait_select_single(objectName=page_object_name)
        page.active.wait_for(True)
        return page

    def scroll_to_and_click(self, obj):
        self.scroll_to(obj)
        self.pointing_device.click_object(obj)

    def click_header_action(self, action):
        """Click the action 'action' on the header"""
        main_view = self.get_root_instance().select_single(
            objectName='systemSettingsMainView')
        header = main_view.select_single('AppHeader')
        header.click_action_button(action)

    @property
    def system_settings_page(self):
        return self.select_single(objectName='systemSettingsPage')

    @property
    def choose_page(self):
        """ Return 'Choose carrier' page """
        return self.select_single(objectName="chooseCarrierPage")

    @property
    def storage_page(self):
        """ Return 'Storage' page """
        return self.select_single(objectName='storagePage')

    @property
    def updates_page(self):
        """ Return 'System Update' page """
        return self.select_single(objectName='systemUpdatesPage')

    @property
    def background_page(self):
        """ Return 'Background' page """
        return self.select_single(objectName='backgroundPage')

    @property
    def _orientation_lock_switch(self):
        return self.wait_select_single(
            CheckBox,
            objectName='orientationLockSwitch')

    def enable_orientation_lock(self):
        self._orientation_lock_switch.check()

    def disable_orientation_lock(self):
        self._orientation_lock_switch.uncheck()

    def _get_battery_item(self):
        return self.select_single(objectName='entryComponent-battery')

    def _get_brightness_item(self):
        return self.select_single(objectName='entryComponent-brightness')

    def go_to_battery_page(self):
        self.scroll_to_and_click(self._get_battery_item())
        return self.wait_select_single('PageComponent', title='Battery')

    def go_to_brightness_page(self):
        self.scroll_to_and_click(self._get_brightness_item())
        return self.wait_select_single('ItemPage', title='Brightness')

    def scroll_to(self, obj):
        """Scroll to item on page and wait for it to stop moving."""

        def get_page_bottom():
            return page.globalRect[1] + page.globalRect[3]

        page = self.system_settings_page
        page_right = page.globalRect[0] + page.globalRect[2]
        page_bottom = get_page_bottom()
        page_center_x = int(page_right / 2)
        page_center_y = int(page_bottom / 2)
        while obj.globalRect[1] + obj.height > get_page_bottom():
            self.pointing_device.drag(
                page_center_x,
                page_center_y,
                page_center_x,
                page_center_y - obj.height * 2
            )
            # avoid a flick, this is very important
            sleep(0.3)
        obj.wait_until_not_moving()


class SimPin(QQuickFlickable):
    """Autopilot helper for the SimPin Page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'SimPin', objectName='simPinPage')

    def get_sim_pin_switch(self, sim_number):
        """Return the SIM PIN switch."""
        switches = self.select_many(objectName='simPinSwitch')
        switches = sorted(switches, key=lambda switch: (switch.globalRect.y))
        return switches[sim_number]

    @autopilot.logging.log_action(logger.debug)
    def click_sim_pin_switch(self, sim_number):
        """Click on the SIM PIN switch, return the SIM PIN dialog."""
        sim_pin_switch = self.get_sim_pin_switch(sim_number)
        self.pointing_device.click_object(sim_pin_switch)
        return self.get_root_instance().wait_select_single(
            objectName='lockDialogComponent'
        )

    @autopilot.logging.log_action(logger.debug)
    def enter_lock_pin(self, pin):
        """Enter the given pin into our dialog."""
        root_instance = self.get_root_instance()
        prev_input = root_instance.wait_select_single(
            objectName='prevInput'
        )
        prev_input.write(pin)
        lock_unlock_button = root_instance.select_single(
            objectName='lockButton'
        )
        self.pointing_device.click_object(lock_unlock_button)


class CellularPage(CustomProxyObjectBase):
    """Autopilot helper for the Cellular settings page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'PageComponent',
            objectName='cellularPage')

    def get_default_sim_for_messages_selector(self, value):
        return self.select_single(objectName='defaultForMessages' + value)

    @autopilot.logging.log_action(logger.debug)
    def enable_data(self):
        self._set_data(True)

    @autopilot.logging.log_action(logger.debug)
    def disable_data(self):
        self._set_data(False)

    @autopilot.logging.log_action(logger.debug)
    def _set_data(self, data):
        chk = self.select_single(objectName='data')
        if data:
            chk.check()
        else:
            chk.uncheck()

    @autopilot.logging.log_action(logger.debug)
    def get_data(self):
        return self.select_single(objectName='data').checked

    @autopilot.logging.log_action(logger.debug)
    def enable_roaming(self, timeout=10):
        self._set_roaming(True, timeout=timeout)

    @autopilot.logging.log_action(logger.debug)
    def disable_roaming(self, timeout=10):
        self._set_roaming(False, timeout=timeout)

    @autopilot.logging.log_action(logger.debug)
    def _set_roaming(self, roaming, timeout):
        chk = self.select_single(objectName='roaming')
        if roaming:
            chk.check(timeout=timeout)
        else:
            chk.uncheck(timeout=timeout)

    @autopilot.logging.log_action(logger.debug)
    def set_connection_type(self, radio_type, sim='/ril_0',
                            scroll_to_and_click=None):
        self._set_connection_type(radio_type, sim, scroll_to_and_click)

    @autopilot.logging.log_action(logger.debug)
    def _set_connection_type(self, radio_type, sim, scroll_to_and_click):
        t = self.wait_select_single(
            'OptionSelectorDelegate',
            objectName='%s_radio_%s' % (sim, radio_type))
        if scroll_to_and_click:
            scroll_to_and_click(t)
        else:
            t.swipe_into_view()

        self.pointing_device.click_object(t)

    @autopilot.logging.log_action(logger.debug)
    def change_carrier(self, carrier, sim=None):
        carrierApnPage = self._click_carrier_apn()
        chooseCarrierPage = carrierApnPage.open_carrier(sim)
        chooseCarrierPage.set_carrier(carrier)

    @autopilot.logging.log_action(logger.debug)
    def open_apn_editor(self, name, sim=None):
        carrierApnPage = self._click_carrier_apn()
        chooseApnPage = carrierApnPage.open_apn(sim)
        return chooseApnPage.open(name)

    @autopilot.logging.log_action(logger.debug)
    def open_apn_page(self, name, sim=None):
        carrierApnPage = self._click_carrier_apn()
        chooseApnPage = carrierApnPage.open_apn(sim)
        return chooseApnPage

    @autopilot.logging.log_action(logger.debug)
    def delete_apn(self, name, sim=None):
        carrierApnPage = self._click_carrier_apn()
        chooseApnPage = carrierApnPage.open_apn(sim)
        return chooseApnPage.delete(name)

    @autopilot.logging.log_action(logger.debug)
    def prefer_apn(self, name, sim=None):
        carrierApnPage = self._click_carrier_apn()
        chooseApnPage = carrierApnPage.open_apn(sim)
        return chooseApnPage.check(name)

    @autopilot.logging.log_action(logger.debug)
    def _click_carrier_apn(self):
        item = self.select_single(objectName='carrierApnEntry')
        self.pointing_device.click_object(item)
        return self.get_root_instance().wait_select_single(
            objectName='carrierApnPage')

    @autopilot.logging.log_action(logger.debug)
    def select_sim_for_data(self, sim):
        self._select_sim_for_data(sim)

    @autopilot.logging.log_action(logger.debug)
    def _select_sim_for_data(self, sim):
        item = self.select_single(objectName='use%s' % sim)
        self.pointing_device.click_object(item)

    @autopilot.logging.log_action(logger.debug)
    def select_sim_for_calls(self, sim):
        pass

    @autopilot.logging.log_action(logger.debug)
    def select_sim_for_messages(self):
        pass

    @autopilot.logging.log_action(logger.debug)
    def set_name(self, sim, name):
        self._set_name(sim, name)

    def get_name(self, sim):
        obj = self.select_single(
            objectName="label_%s" % sim)
        return obj.text

    def _set_name(self, sim, name):
        obj = self.select_single(
            objectName="edit_name_%s" % sim)
        self.pointing_device.click_object(obj)

        # wait for animation
        sleep(1)
        ok = self.select_single('Button', objectName="doRename")

        field = self.wait_select_single('TextField', objectName="nameField")
        field.write(name)
        self.pointing_device.click_object(ok)


class BatteryPage(CustomProxyObjectBase):
    """Autopilot helper for the Battery settings page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'PageComponent', title='Battery')

    def _get_wifi_switch_row(self):
        return self.wait_select_single('Standard', text='Wi-Fi')

    def get_battery_page_wifi_switch(self):
        """Return the wifi switch in battery settings page."""
        wifi_row = self._get_wifi_switch_row()
        return wifi_row.select_single('CheckBox', enabled=True)

    def is_wifi_switch_checked(self):
        return self.get_battery_page_wifi_switch().checked

    def turn_off_wifi(self):
        """Turn off wifi from battery settings page."""
        wifi_switch = self.get_battery_page_wifi_switch()
        wifi_switch.swipe_into_view()
        self.pointing_device.click_object(wifi_switch)
        wifi_switch.checked.wait_for(False)


class AboutPage(CustomProxyObjectBase):
    """Autopilot helper for the About page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'PageComponent',
            objectName='aboutPage')

    def get_device_name(self):
        device_label = self.select_single(objectName='deviceLabel')
        return device_label.text

    def is_serial_visible(self):
        serial_item = self._get_serial_item()
        return serial_item.visible

    def _get_serial_item(self):
        return self.select_single(objectName='serialItem')

    def get_serial(self):
        serial_item = self._get_serial_item()
        return serial_item.value

    def is_imei_visible(self):
        imei_item = self._get_imei_item()
        return imei_item.visible

    def _get_imei_item(self):
        return self.wait_select_single(objectName='imeiItem')

    def get_imei(self):
        imei_item = self._get_imei_item()
        return imei_item.value

    def get_os_information(self):
        os_item = self.select_single(objectName='osItem')
        return os_item.value

    def get_last_updated_date(self):
        last_updated_item = self.select_single(objectName='lastUpdatedItem')
        return last_updated_item.value

    @autopilot.logging.log_action(logger.info)
    def go_to_check_for_updates(self):
        check_for_updates_button = self.select_single(
            objectName='updateButton')
        check_for_updates_button.swipe_into_view()
        self.pointing_device.click_object(check_for_updates_button)
        system_updates_page = self.get_root_instance().wait_select_single(
            objectName='systemUpdatesPage')
        system_updates_page.active.wait_for(True)
        return system_updates_page

    @autopilot.logging.log_action(logger.info)
    def go_to_software_licenses(self):
        license_item = self.select_single(
            'StandardProgression', objectName='licenseItem')
        license_item.swipe_into_view()
        self.pointing_device.click_object(license_item)
        licenses_page = self.get_root_instance().wait_select_single(
            objectName='licensesPage')
        licenses_page.active.wait_for(True)
        return licenses_page

    def get_number(self, obj):
        number = self.select_single(objectName=obj)
        return number.value

    def get_wifi_address(self):
        """ Retrieve the wifi address value """
        return self.wait_select_single('SingleValue',
                                       text='Wi-Fi address').value

    def get_bluetooth_address(self):
        """ Retrieve the bluetooth address value """
        return self.wait_select_single('SingleValue',
                                       text='Bluetooth address').value

    def _get_storage(self):
        return self.wait_select_single('SingleValue', objectName='storageItem')

    def get_storage(self):
        """ Retrieve the storage item value """
        return self._get_storage().value

    def _go_to_page(self, itemname, pagename):
        item = self.select_single(objectName=itemname)
        item.swipe_into_view()
        self.pointing_device.click_object(item)
        page = self.get_root_instance().wait_select_single(
            objectName=pagename)
        page.active.wait_for(True)
        return page

    def go_to_os(self):
        """ Go to the OS version page and retrieve the page """
        return self._go_to_page('osItem', 'versionPage')

    def go_to_developer_mode(self):
        """ Go to the developer mode page and retrieve the page """
        return self._go_to_page('devmodeItem', 'devModePage')

    def go_to_storage(self):
        """ Go to the storage page and retrieve the page """
        return self._go_to_page('storageItem', 'storagePage')


class LicensesPage(CustomProxyObjectBase):
    """Autopilot helper for the Licenses page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ItemPage',
            objectName='licensesPage')

    def _get_licenses_list(self):
        return self.wait_select_single('QQuickListView')

    def _get_license(self, text):
        return self.select_single('Standard', text=text)

    def get_license(self, text):
        """ Swipe until the object appears in the tree, it is not possible to
        use swipe_child_into_view because the license objects are not part of
        the tree
        """
        list = self._get_licenses_list()
        while not list.atYEnd:
            list.swipe_to_show_more_below()
            if is_element(self._get_license, text):
                return self._get_license(text)

        raise exceptions.StateNotFoundError(class_name='Standard', text=text)


class VersionPage(CustomProxyObjectBase):
    """Autopilot helper for the Version page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'Version',
            objectName='versionPage')

    def get_build_number(self):
        return self.select_single('SingleValueStacked',
                                  objectName='versionBuildNumberItem').value

    def get_build_desc(self):
        return self.select_single('SingleValueStacked',
                                  objectName='ubuntuBuildIDItem').value


class DevModePage(CustomProxyObjectBase):
    """Autopilot helper for the developer mode page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ItemPage',
            objectName='devModePage')

    def _get_devmode_checkbox(self):
        return self.select_single('CheckBox',
                                  styleName='SwitchStyle')

    def is_devmode_on(self):
        return self._get_devmode_checkbox().checked

    def switch_devmode(self):
        switcher = self._get_devmode_checkbox()
        checked = switcher.checked
        self.pointing_device.click_object(switcher)
        switcher.checked.wait_for(not checked)


class StoragePage(CustomProxyObjectBase):
    """Autopilot helper for the developer mode page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'Storage',
            objectName='storagePage')

    def get_pictures_storage(self):
        return self.select_single('StorageItem',
                                  objectName='picturesItem').value

    def get_videos_storage(self):
        return self.select_single('StorageItem',
                                  objectName='moviesItem').value

    def get_audio_storage(self):
        return self.select_single('StorageItem',
                                  objectName='audioItem').value

    def _get_gallery_app(self):
        return self.select_single('SingleValue', objectName='appItemGallery')

    def get_gallery_app_size_in_disk(self):
        return self._get_gallery_app().value

    def get_gallery_app_icon(self):
        return self._get_gallery_app().icon


class BrightnessPage(BrightnessPageBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ItemPage',
            objectName='brightnessPage')

    def get_auto_adjust_checkbox(self):
        # TODO: Use select single once the checkbox gets an object name,
        # Until then, we need to get all the elements and filter them
        return self.select_many(
            'CheckBox', ap_result_sort_keys=KEY_ORDER_BY_Y)[0]

    def get_max_brightness(self):
        return 100
