# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from retrying import retry

from autopilot.exceptions import StateNotFoundError

from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA

from ubuntu_system_tests.helpers.autopilot import validate_dbus_object

APP_PATH_ROOT = b'comubuntutelegram'


def _retry_if_state_not_found_error(exception):
    return isinstance(exception, StateNotFoundError)


class TelegramApp(CustomProxyObjectBase):
    """Autopilot helper object for telegram app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, APP_PATH_ROOT)

    @property
    def main_view(self):
        return self.wait_select_single(MainView)


class MainView(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'MainView', objectName='appMainView')

    @property
    def intro_page(self):
        return self.wait_select_single(objectName='introPage')


class IntroPage(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'IntroPage', objectName='introPage')

    @retry(stop_max_attempt_number=3,
           retry_on_exception=_retry_if_state_not_found_error)
    def _get_auth_countries_page(self):
        countries_list_page = self.get_root_instance().wait_select_single(
            objectName='countriesListPage')
        countries_list_page.visible.wait_for(True)
        return countries_list_page

    def go_to_countries_list(self):
        start_button = self.select_single(objectName='startMessagingButton')
        self.pointing_device.click_object(start_button)
        return self._get_auth_countries_page()


class AuthCountriesPage(CustomProxyObjectBase):
    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'AuthCountriesPage',
            objectName='countriesListPage')

    def _swipe_to_country(self, list, country):
        while not list.atYEnd:
            try:
                item = list.select_single(objectName=country)
                item.swipe_into_view()
                return item
            except StateNotFoundError:
                list.swipe_to_show_more_below()

        raise StateNotFoundError

    def _convert_country_to_object_name(self, name):
        return name.lower().replace(' ', '_')

    @retry(stop_max_attempt_number=3,
           retry_on_exception=_retry_if_state_not_found_error)
    def _get_auth_number_page(self):
        countries_number_page = self.get_root_instance().wait_select_single(
            objectName='authPhonePage')
        countries_number_page.visible.wait_for(True)
        return countries_number_page

    def select_country(self, name):
        countries_list = self.select_single(objectName='countriesList')
        item_object_name = self._convert_country_to_object_name(name)
        country_item = self._swipe_to_country(countries_list, item_object_name)
        self.pointing_device.click_object(country_item)
        return self._get_auth_number_page()


class AuthNumberPage(CustomProxyObjectBase):
    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'AuthNumberPage',
            objectName='authPhonePage')

    def _get_country_code(self):
        return self.select_single(objectName='countryCode').text

    def _get_phone_number_entry(self):
        return self.select_single(objectName='phoneNumberEntry')

    def _get_done_button(self):
        return self.select_single(objectName='doneButton')

    def _enter_phone_number(self, number):
        self._get_phone_number_entry().write(number)

    def _confirm_number(self):
        confirmation_dialog = self.get_root_instance().wait_select_single(
            objectName='confirmationDialog')
        confirmation_dialog.visible.wait_for(True)
        confirmation_dialog.accept()

    @retry(stop_max_attempt_number=3,
           retry_on_exception=_retry_if_state_not_found_error)
    def _get_auth_code_page(self):
        auth_code_page = self.get_root_instance().wait_select_single(
            AuthCodePage)
        auth_code_page.visible.wait_for(True)
        return auth_code_page

    def _strip_country_code(self, number):
        country_code = self._get_country_code()
        if number.startswith(country_code):
            return number[len(country_code):]

        return number

    def send_phone_number(self, number, timeout=120):
        self._enter_phone_number(number=self._strip_country_code(number))
        self.pointing_device.click_object(self._get_done_button())
        self._confirm_number()
        progress_indicator = self.wait_select_single(
            objectName='processingIndicator')
        progress_indicator.wait_until_destroyed(timeout=timeout)
        return self._get_auth_code_page()


class ConfirmationDialog(CustomProxyObjectBase):
    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ConfirmationDialog',
            objectName='confirmationDialog')

    def _get_ok_button(self):
        return self.select_single(objectName='acceptButton')

    def accept(self):
        ok_button = self._get_ok_button()
        self.pointing_device.click_object(ok_button)
        self.wait_until_destroyed()


class AuthCodePage(CustomProxyObjectBase):
    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'AuthCodePage')

    def _get_verification_code_entry(self):
        return self.select_single(objectName='verificationCodeEntry')

    def _get_first_name_entry(self):
        return self.select_single(objectName='firstNameEntry')

    def _get_last_name_entry(self):
        return self.select_single(objectName='lastNameEntry')

    def _get_submit_button(self):
        return self.wait_select_single(objectName='signInButton')

    def _write_verification_code(self, code):
        self._get_verification_code_entry().write(code)

    def _write_first_name(self, first_name):
        self._get_first_name_entry().write(first_name)

    def _write_last_name(self, last_name):
        self._get_last_name_entry().write(last_name)

    def _sign_in(self, code):
        self._write_verification_code(code)
        self.pointing_device.click_object(self._get_submit_button())

    @retry(stop_max_attempt_number=3,
           retry_on_exception=_retry_if_state_not_found_error)
    def _get_account_page(self):
        account_page = self.get_root_instance().wait_select_single(
            AccountPage)
        account_page.visible.wait_for(True)
        return account_page

    def sign_in_or_signup(self, code, first_name, last_name, timeout=120):
        if not self.phoneRegistered:
            self._write_first_name(first_name=first_name)
            self._write_last_name(last_name=last_name)
        self._sign_in(code=code)

        progress_indicator = self.wait_select_single(
            objectName='processingIndicator')
        progress_indicator.wait_until_destroyed(timeout=timeout)

        return self._get_account_page()


class AccountPage(CustomProxyObjectBase):
    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'AccountPage',
            objectName='acountPage')
