# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import time

from autopilot import input
from gi.repository import Gio

from ubuntu_system_tests.helpers import context
from ubuntu_system_tests.helpers.init_systems import get_init_system
from ubuntu_system_tests.helpers.processes import is_testability_enabled_for_process  # NOQA
from ubuntu_system_tests.helpers import autopilot
from ubuntu_system_tests.helpers import processes
from ubuntu_system_tests.helpers.ubuntu_keyboard.job import MaliitServerJob


maliit = MaliitServerJob()

init_system = get_init_system()


def get_ubuntu_keyboard_proxy():
    from ubuntu_system_tests.helpers.ubuntu_keyboard import _cpo  # NOQA
    return autopilot.get_proxy_object(processes.get_process_id(maliit.name))


def is_keyboard_shifted():
    return get_keyboard()._keyboard.active_keypad_state == 'SHIFTED'


def is_maliit_server_running():
    """Check if maliit-server is running
    :return True if maliit-server is running, False otherwise
    """
    return processes.is_process_running(maliit.name, retries=False)


def get_keyboard():
    """Return the keyboard device."""
    if not hasattr(context.shared, 'keyboard') or not context.shared.keyboard:
        if init_system.is_running(maliit) and os.environ.get(
                'UITK_USE_MALIIT'):
            configure_osk_settings()
            restart_maliit_with_testability()
            context.shared.keyboard = input.Keyboard.create('OSK')
        else:
            context.shared.keyboard = input.Keyboard.create()
    return context.shared.keyboard


def restart_maliit_with_testability():
    """Restart maliit-server with testability enabled."""
    init_system.get_pid(maliit)
    pid = init_system.get_pid(maliit)
    if is_testability_enabled_for_process(pid):
        return
    init_system.stop(maliit)
    init_system.start(maliit, 'QT_LOAD_TESTABILITY=1')
    # This is needed to work around https://launchpad.net/bugs/1248913
    time.sleep(5)


def configure_osk_settings():
    """Configure OSK ready for testing by turning off all helpers."""
    gsettings = Gio.Settings.new("com.canonical.keyboard.maliit")
    gsettings.set_string("active-language", "en")
    gsettings.set_boolean("auto-capitalization", False)
    gsettings.set_boolean("auto-completion", False)
    gsettings.set_boolean("predictive-text", False)
    gsettings.set_boolean("spell-checking", False)
    gsettings.set_boolean("double-space-full-stop", False)


def get_osk_area():
    """Return globalRect of OSK."""
    return get_keyboard()._keyboard.active_keypad.globalRect


def is_osk_showing():
    """Return True if keyboard is showing on screen."""
    return get_keyboard()._keyboard.is_available()


def dismiss_keyboard():
    """Dismiss the OSK if it is showing."""
    if is_osk_showing():
        get_keyboard()._keyboard.dismiss()


def is_item_obscured_by_osk(item):
    """Return True if specified item is within OSK screen area."""
    if is_osk_showing():
        osk_area = get_osk_area()
        item_area = item.globalRect
        item_bottom_y = item_area.y + item_area.height
        return item_bottom_y >= osk_area.y
    return False
