# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014, 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import logging
import subprocess

import fixtures
import psutil
from dbus import DBusException
from retrying import retry

from ubuntu_system_tests.helpers import context
from ubuntu_system_tests.helpers import wait_until
from ubuntu_system_tests.helpers.autopilot import get_proxy_object
from ubuntu_system_tests.helpers.init_systems import get_init_system
from ubuntu_system_tests.helpers.processes import (
    PidNotFoundError,
    ensure_all_apps_closed)

from ubuntu_system_tests.helpers.unity8 import Unity8Job
from autopilot import (
    exceptions
)

from ubuntu_system_tests.helpers.unity8 import get_dash

logger = logging.getLogger(__name__)

UBUNTU_PLATFORM_API_TEST_OVERRIDE = 'UBUNTU_PLATFORM_API_TEST_OVERRIDE'


class SettingsWizard(fixtures.Fixture):

    WIZARD_FILE = '/home/phablet/.config/ubuntu-system-settings/wizard-has-run'

    def __init__(self, enable):
        super().__init__()
        self.enable = enable

    def setUp(self):
        super().setUp()
        original_state = self.is_settings_wizard_enabled()
        self.addCleanup(self.set_settings_wizard, original_state)
        if self.enable != original_state:
            self.set_settings_wizard(self.enable)

    def is_settings_wizard_enabled(self):
        return not os.path.exists(self.WIZARD_FILE)

    def set_settings_wizard(self, enabled):
        if enabled:
            self.enable_settings_wizard()
        else:
            self.disable_settings_wizard()

    def enable_settings_wizard(self):
        if os.path.exists(self.WIZARD_FILE):
            os.remove(self.WIZARD_FILE)

    def disable_settings_wizard(self):
        if not os.path.exists(self.WIZARD_FILE):
            open(self.WIZARD_FILE, 'a').close()


class SetupUnity8(fixtures.Fixture):

    def __init__(self, restart=True):
        self.unity8 = Unity8Job()
        self.init_system = get_init_system()
        self.restart = restart

    def setUp(self):
        super().setUp()
        self.setup_unity(restart=self.restart)

    # Retry added until fix for bug 1427946 in QtMir is landed
    @retry(stop_max_attempt_number=5,
           wait_fixed=2000,
           retry_on_exception=lambda exception: (
               isinstance(exception, psutil.NoSuchProcess) or
               isinstance(exception, ValueError) or
               isinstance(exception, DBusException) or
               isinstance(exception, TimeoutError) or
               isinstance(exception, PidNotFoundError) or
               isinstance(exception, exceptions.ProcessSearchError) or
               isinstance(exception, exceptions.BackendException)))
    def setup_unity(self, restart=True):
        """Setup unity8 environment, restarting if any error is observed.
        :param restart: Whether unity8 should be restarted.
        """
        from ubuntu_system_tests.helpers.unity8.shell import Unity8  # NOQA
        if restart:
            self.init_system.ensure_stopped(self.unity8)
            self.init_system.start(self.unity8)
        wait_until(self.init_system.is_running, self.unity8)
        context.shared.unity_pid = self.init_system.get_pid(self.unity8)
        context.shared.unity = get_proxy_object(pid=context.shared.unity_pid)
        self.unity8.wait_for_unity_ready()
        context.shared.unity.wait_for_greeter_content()
        context.shared.unity.unlock()
        context.shared.dash = get_dash()
        self.addCleanup(ensure_all_apps_closed)


class EnableUnity8FakeSensors(fixtures.Fixture):
    """Fixture to enable fake sensors for unity8."""
    def __init__(self):
        self.unity8 = Unity8Job()
        self.init_system = get_init_system()

    def setUp(self):
        super().setUp()
        self.init_system.ensure_stopped(self.unity8)
        self.init_system.set_environment_variable(
            UBUNTU_PLATFORM_API_TEST_OVERRIDE, 'sensors',
            global_scope=False)
        self.addCleanup(self.init_system.unset_environment_variable,
                        UBUNTU_PLATFORM_API_TEST_OVERRIDE)
        args = ['QT_LOAD_TESTABILITY=1', '--no-wait']
        self.init_system.start(self.unity8, *args)
        self.unity8.sensors.validate_fifo_ready(self.unity8.name)
        self.unity8.sensors.setup_sensors(self.unity8.name)
        self.useFixture(SetupUnity8(restart=False))


class Tutorial(fixtures.Fixture):

    def __init__(self, enable):
        super().__init__()
        self.enable = enable

    def setUp(self):
        super().setUp()
        original_state = self._is_tutorial_enabled()
        if self.enable != original_state:
            self.addCleanup(self._set_tutorial, original_state)
            self._set_tutorial(self.enable)

    def _is_tutorial_enabled(self):
        command = [
            'dbus-send', '--system', '--print-reply',
            '--dest=org.freedesktop.Accounts',
            '/org/freedesktop/Accounts/User32011',
            'org.freedesktop.DBus.Properties.Get',
            'string:com.canonical.unity.AccountsService',
            'string:demo-edges'
        ]
        output = subprocess.check_output(command, universal_newlines=True)
        return output.count('true') > 0

    def _set_tutorial(self, value):
        value_string = 'true' if value else 'false'
        command = [
            'dbus-send', '--system', '--print-reply',
            '--dest=com.canonical.PropertyService',
            '/com/canonical/PropertyService',
            'com.canonical.PropertyService.SetProperty',
            'string:edge', 'boolean:{}'.format(value_string)
        ]
        subprocess.check_output(command)
