# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import time

from collections import deque
from importlib import import_module
from inspect import stack

logger = logging.getLogger(__name__)

DUMMY_IMPORT_ERR = ('DummyImportObject being used because try_import failed. '
                    'Check all required dependencies are installed.')


def try_import(name, package=None):
    """
    Import named module from package and return it or DummyImportObject if
    not found.

    This method is used in cases where an import is made but should not raise
    an ImportError if the module is not found. The main usage for this is
    where an autopilot package is not installed on the host, but it is still
    required to list all availble tests on the host. In this case the test
    module must import successfully, but won't actually be executed so it is
    not neccessary to have the dependency installed.

    If a DummyImportObject is returned this allows any attributes or methods
    to be imported from the DummyImportObject without error.

    :param name: Absolute or relative name of module to import.
    :param package: If name is relative, package must specify the name of the
                    package to import from.
    :return: Imported module or DummyImportObject if it doesn't exist.

    """
    try:
        return import_module(name, package)
    except ImportError:
        return DummyImportObject()


def wait_until(predicate, *args, timeout=10, period=1, stop_event=None,
               validator=None, **kwargs):
    """
    Waits either until the timeout is reached, the predicate evaluation is
    True, or the stop_event is set.

    To evaluate the predicate
     :param: predicate: The method to be evaluated
     :param: *args: The predicate positional parameters
     :param: timeout: The timeout for the polling
     :param: period: The retry period of time
     :param: stop_event: Event object to indicate whether to stop waiting
                         before the timeout period is over. This will be
                         ignored if not set.
     :param: validator: Validation method used to evaluate result of predicate.
                        This will be ignored if not set.
     :param: **kwargs: The predicate optional parameters
     :return: Result of predicate if evaluated as True before the
              timeout is reached, False otherwise.
    """
    end = time.time() + timeout
    while time.time() < end:
        if stop_event and stop_event.is_set():
            return False
        result = predicate(*args, **kwargs)
        if _is_match(result, validator):
            return result
        time.sleep(period)
    return False


def wait_until_stable(predicate, *args, timeout=10, period=1, count=3,
                      stop_event=None, **kwargs):
    """
    Waits either until the timeout is reached, the predicate result returns
    the same value for the specified number of times, or the stop_event is
    set.

     :param: predicate: The method to be evaluated which must return a value
                        to be compared.
     :param: *args: The predicate positional parameters
     :param: timeout: The timeout for the polling
     :param: period: The retry period of time
     :param: count: The number of results to compare which must all be equal
                    for result to be set True.
     :param: stop_event: Event object to indicate whether to stop waiting
                         before the timeout period is over. This will be
                         ignored if not set.
     :param: **kwargs: The predicate optional parameters
     :return: True if the result of predicate is stable within the timeout
              period, False otherwise.
    """
    results = deque(maxlen=count)
    end = time.time() + timeout
    while time.time() < end:
        if stop_event and stop_event.is_set():
            return False
        results.append(predicate(*args, **kwargs))
        if results.count(results[0]) == results.maxlen:
            return True
        time.sleep(period)
    return False


def _is_match(result, validator):
    """
    Return True if the the result is evaluated as valid.

    If a validator method is specified then return result of validation.
    Otherwise return boolean evaluation of result.

    :param result: Result to evaluate
    :param validator: Validation method to execute using result as parameter.
    :return: True if result is evaluated as True, False otherwise.

    """
    if validator:
        return validator(result)
    else:
        return bool(result)


def _is_import():
    """Return True if being imported during test discovery."""
    stack_trace = stack()
    stack_trace.reverse()
    for frame in stack_trace:
        if frame[3] == 'load_test_suite_from_name':
            return True
    return False


class AttrDict(dict):
    """ Class used to access to the dict keys as parameters """

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class DummyImportObject:
    """
    Class used to create objects with dummy attributes and methods that can
    be accessed during import operations only.

    This is used when an object is required to return a dummy attribute value
    or method result during import rather than raise a AttributeError or
    TypeError exception.

    This class can be used to create objects that support nested attributes
    and methods.

    """

    def __init__(self, *args, **kwargs):
        pass

    def __getattribute__(self, name):
        """Return a refernce to self for every attribute value to support
        nested attributes."""
        if _is_import():
            return self
        raise AttributeError(DUMMY_IMPORT_ERR)

    def __call__(self, *args, **kwargs):
        """Return a reference to self for every function call to support
        nested function calls."""
        if _is_import():
            return self
        raise AttributeError(DUMMY_IMPORT_ERR)
