# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2015, 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

import time

from autopilot import introspection
from autopilot.exceptions import StateNotFoundError
from autopilot.platform import model
from autopilot import input
from retrying import retry

from ubuntu_system_tests.helpers.ubuntuuitools._cpo._textfield import TextField  # NOQA
from ubuntu_system_tests.helpers.ubuntuuitools._cpo._common import CustomProxyObjectBase  # NOQA

from ubuntu_system_tests.helpers.autopilot import (
    patched_wait_select_single, validate_dbus_object)
from ubuntu_system_tests.helpers.autopilot.list import KEY_ORDER_BY_Y

logger = logging.getLogger(__name__)

APP_PATH_ROOT = b'webbrowser-app'


class Webbrowser(CustomProxyObjectBase):

    """Autopilot custom proxy object for the webbrowser app."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        name = introspection.get_classname_from_path(path)
        if name == b'webbrowser-app':
            if state['applicationName'][1] == 'webbrowser-app':
                return True
        return False

    @property
    def main_window(self):
        return self.select_single(Browser)

    def get_windows(self, **kwargs):
        return self.select_many(Browser, **kwargs)


class Browser(CustomProxyObjectBase):

    def __init__(self, *args):
        super().__init__(*args)
        self.chrome = self._get_chrome()
        self.address_bar = self.chrome.address_bar
        self.keyboard = input.Keyboard.create()

    def _get_chrome(self):
        return self.select_single(Chrome)

    def wait_until_page_loaded(self, url):
        webview = self.wait_select_single("WebViewImpl", current=True, url=url)
        # loadProgress == 100 ensures that a page has actually loaded
        webview.loadProgress.wait_for(100, timeout=20)
        webview.loading.wait_for(False)

    def go_back(self):
        self.chrome.go_back()

    def go_forward(self):
        self.chrome.go_forward()

    # Since the NewPrivateTabView does not define any new QML property in its
    # extended file, it does not report itself to autopilot with the same name
    # as the extended file. (See http://pad.lv/1454394)
    def is_new_private_tab_view_visible(self):
        try:
            self.get_new_private_tab_view()
            return True
        except StateNotFoundError:
            return False

    def get_window(self):
        return self.get_parent()

    def get_current_webview(self):
        return self.select_single("WebViewImpl", current=True)

    def get_webviews(self):
        return self.select_many("WebViewImpl", incognito=False)

    def get_incognito_webviews(self):
        return self.select_many("WebViewImpl", incognito=True)

    def get_error_sheet(self):
        return self.select_single("ErrorSheet")

    def get_sad_tab(self):
        return self.wait_select_single(SadTab)

    def get_suggestions(self):
        return self.select_single(Suggestions)

    def get_geolocation_dialog(self):
        return self.wait_select_single(GeolocationPermissionRequest)

    def get_http_auth_dialog(self):
        return self.wait_select_single(HttpAuthenticationDialog)

    def get_media_access_dialog(self):
        return self.wait_select_single(MediaAccessDialog)

    def get_tabs_view(self):
        return self.wait_select_single(TabsList, visible=True)

    def get_recent_view_toolbar(self):
        return self.wait_select_single(Toolbar, objectName="recentToolbar",
                                       state="shown")

    def get_new_tab_view(self):
        if self.wide:
            return self.wait_select_single("NewTabViewWide", visible=True)
        else:
            return self.wait_select_single("NewTabView", visible=True)

    # Since the NewPrivateTabView does not define any new QML property in its
    # extended file, it does not report itself to autopilot with the same name
    # as the extended file. (See http://pad.lv/1454394)
    def get_new_private_tab_view(self):
        return self.wait_select_single("QQuickItem",
                                       objectName="newPrivateTabView",
                                       visible=True)

    def get_settings_page(self):
        return self.wait_select_single(SettingsPage, visible=True)

    def get_downloads_page(self):
        return self.wait_select_single(DownloadsPage, visible=True)

    def get_content_picker_dialog(self):
        # only on devices
        return self.wait_select_single("PopupBase",
                                       objectName="contentPickerDialog")

    def get_download_dialog(self):
        return self.wait_select_single("PopupBase",
                                       objectName="downloadDialog")

    def get_peer_picker(self):
        return self.wait_select_single(objectName="contentPeerPicker")

    def click_cancel_download_button(self):
        button = self.select_single("Button",
                                    objectName="cancelDownloadButton")
        self.pointing_device.click_object(button)

    def click_choose_app_button(self):
        button = self.select_single("Button",
                                    objectName="chooseAppButton")
        self.pointing_device.click_object(button)

    def click_download_file_button(self):
        button = self.select_single("Button",
                                    objectName="downloadFileButton")
        self.pointing_device.click_object(button)

    def get_bottom_edge_handle(self):
        return self.select_single(objectName="bottomEdgeHandle")

    def get_bottom_edge_bar(self):
        return self.select_single(objectName="bottomEdgeBar", visible=True)

    def get_bookmark_options(self):
        return self.select_single(BookmarkOptions)

    def get_new_bookmarks_folder_dialog(self):
        return self.wait_select_single("Dialog",
                                       objectName="newFolderDialog")

    # The bookmarks view is dynamically created, so it might or might not be
    # available
    def get_bookmarks_view(self):
        try:
            if self.wide:
                return self.select_single("BookmarksViewWide")
            else:
                return self.select_single("BookmarksView")
        except StateNotFoundError:
            return None

    # The history view is dynamically created, so it might or might not be
    # available
    def get_history_view(self):
        try:
            if self.wide:
                return self.wait_select_single(HistoryViewWide)
            else:
                return self.wait_select_single(HistoryView)
        except StateNotFoundError:
            return None

    def get_expanded_history_view(self):
        return self.wait_select_single(ExpandedHistoryView, visible=True)

    def press_key(self, key):
        self.keyboard.press_and_release(key)

    def get_context_menu(self):
        if self.wide:
            return self.wait_select_single(ContextMenuWide)
        else:
            return self.wait_select_single(ContextMenuMobile)

    def open_item_context_menu_on_item(self, item, menuClass):
        cx = item.globalRect.x + item.globalRect.width // 2
        cy = item.globalRect.y + item.globalRect.height // 2
        self.pointing_device.move(cx, cy)
        if model() == 'Desktop':
            self.pointing_device.click(button=3)
        else:
            self.pointing_device.press()
            time.sleep(1.5)
            self.pointing_device.release()
        return self.wait_select_single(menuClass)

    def open_context_menu(self):
        webview = self.get_current_webview()
        chrome = self.chrome
        x = webview.globalRect.x + webview.globalRect.width // 2
        y = webview.globalRect.y + \
            (webview.globalRect.height + chrome.height) // 2
        self.pointing_device.move(x, y)
        if model() == 'Desktop':
            self.pointing_device.click(button=3)
        else:
            self.pointing_device.press()
            time.sleep(1.5)
            self.pointing_device.release()
        return self.get_context_menu()

    def dismiss_context_menu(self, menu):
        if self.wide:
            # Dismiss by clicking outside of the menu
            webview_rect = self.get_current_webview().globalRect
            actions = menu.get_visible_actions()
            outside_x = (webview_rect.x + actions[0].globalRect.x) // 2
            outside_y = webview_rect.y + webview_rect.height // 2
            self.pointing_device.move(outside_x, outside_y)
            self.pointing_device.click()
        else:
            # Dismiss by clicking the cancel action
            menu.click_cancel_action()
        menu.wait_until_destroyed()

    def get_alert_dialog(self):
        return AlertDialog(
            self.wait_select_single("Dialog", objectName="alertDialog")
        )

    def get_before_unload_dialog(self):
        return BeforeUnloadDialog(
            self.wait_select_single("Dialog", objectName="beforeUnloadDialog")
        )

    def get_confirm_dialog(self):
        return ConfirmDialog(
            self.wait_select_single("Dialog", objectName="confirmDialog")
        )

    def get_prompt_dialog(self):
        return PromptDialog(
            self.wait_select_single("Dialog", objectName="promptDialog")
        )

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'Browser')

    def get_page_loaded(self, url):
        """Wait until the page is fully loaded and returns the text in the
        address bar
        :param url The url to wait
        :return the text displayed in the address bar
        """

        self.wait_until_page_loaded(url)
        return self.address_bar.text

    def get_current_url(self):
        return self.address_bar.text_field.text

    @retry(stop_max_attempt_number=3,
           wait_fixed=1000,
           retry_on_exception=lambda exception: (
               isinstance(exception, ValueError) or
               isinstance(exception, AssertionError)))
    def go_to_url(self, url):
        """Make the browser go to the url passed as parameter
        :param url: The url to go to.
        :raises: ValueError when the url could not be correctly written.
        :raises: AssertionError when the existing url cannot be cleared.
        """
        self.address_bar.wait_until_not_moving()
        if self.get_current_url() != url:
            self.address_bar.clear()
            self.address_bar.write(url)
            if self.get_current_url() != url:
                raise ValueError(
                    'The url couldn\'t be correctly written into address bar')
        self.address_bar.text_field.keyboard.press_and_release('Enter')

    def open_drawer(self):
        """ Open the chrome bar drawer """
        drawer_button = self.chrome.get_drawer_button()
        self.pointing_device.click_object(drawer_button)

    def open_history_view(self):
        """Open the history page where the visited pages are shown"""
        self.open_drawer()
        history_action = self.chrome.get_drawer_action('history')
        self.pointing_device.click_object(history_action)
        history_action.wait_until_destroyed()

    def open_bookmarks_view(self):
        """Open the bookmarks page where bookmarks are shown"""
        self.open_drawer()
        bookmarks_action = self.chrome.get_drawer_action('bookmarks')
        self.pointing_device.click_object(bookmarks_action)
        bookmarks_action.wait_until_destroyed()
        return self.get_bookmarks_view()

    def open_downloads_page(self):
        """Open the downloads page where all downloads are shown"""
        self.open_drawer()
        downloads_action = self.chrome.get_drawer_action('downloads')
        self.pointing_device.click_object(downloads_action)
        downloads_action.wait_until_destroyed()
        return self.wait_select_single(DownloadsPage)

    def open_find_in_page(self):
        """ Open the find in page option """
        self.open_drawer()
        find_action = self.chrome.get_drawer_action('findinpage')
        self.pointing_device.click_object(find_action)
        find_action.wait_until_destroyed()

    def find_in_page(self, text):
        """ Write the text in the find in page textfield """
        self._get_find_in_page_textfield().clear()
        self._get_find_in_page_textfield().write(text)

    def _get_find_in_page_textfield(self):
        return self.wait_select_single(TextField,
                                       objectName='addressBarTextField',
                                       visible=True)

    def get_find_in_page_counter_text(self):
        return self.select_single(objectName='findInPageCounter').text

    def is_find_in_page_enabled(self):
        return self._get_chrome().findInPageMode

    def is_find_in_page_disable(self):
        return not self.is_find_in_page_enabled()

    def is_find_in_page_cursor_displayed(self):
        """ Indicates if the cursor is being displayed in the find in page
        text bar
        """
        textfield = self._get_find_in_page_textfield()
        return textfield.cursorVisible and textfield.cursorPosition >= 0

    def get_find_in_page_placeholder_text(self):
        return self._get_find_in_page_textfield().placeholderText

    def get_find_in_page_previous_button(self):
        return self.select_single(objectName='findPreviousButton')

    def get_find_in_page_next_button(self):
        return self.select_single(objectName='findNextButton')

    def get_history_urls(self):
        """Return all the pages (urls) that are shown in the history page"""
        return self.wait_select_many('UrlDelegate', visible=True)

    def get_tab_preview_urls(self):
        """Return all the pages (urls) that are shown in the history page"""
        return self._get_urls_delegates()

    def _get_urls_delegates(self):
        return self.select_many('UrlDelegate')

    def open_new_tab(self):
        """Open a new tab from the history view and return the tab"""
        self.open_history_view()
        new_tab_button = self.select_single("ToolbarAction",
                                            iconName="tab-new")
        self.pointing_device.click_object(new_tab_button)
        return self.get_new_tab_view()

    def get_popover_foreground_dialog(self):
        self.select_single('UCStyledItemBase', objectName='popover_foreground')

    def get_popover_foreground_dialog_item(self, text):
        elems = self.select_many('Empty', visible=True)
        return next(elem for elem in elems if elem.text == text)

    def get_download_options_dialog(self):
        return patched_wait_select_single(
            self, Dialog, class_name='Dialog',
            objectName='downloadOptionsDialog')

    def get_open_with_dialog(self):
        return patched_wait_select_single(
            self, ContentPeerPicker13, class_name='ContentPeerPicker13',
            objectName='contentPeerPicker')

    def open_page_and_accept_location_permission_prompt(self, url):
        """Open the provided url in the browser and accept the location
        access prompt."""
        self.go_to_url(url)
        dialog = self.get_geolocation_dialog()
        accept_button = dialog.get_allow_button()
        self.pointing_device.click_object(accept_button)
        self.wait_page_loaded()

    def wait_page_loaded(self):
        """Wait for the webbrowser to stop loading the current page."""
        webview = self.get_current_webview()
        # loadProgress == 100 ensures that a page has actually loaded
        webview.loadProgress.wait_for(100, timeout=20)
        webview.loading.wait_for(False)

    def _get_popup_copy_button(self):
        """Return the 'copy' button on the address bar's contextual popup."""
        copy_button = self._get_popover().select_single(iconName='edit-copy')
        copy_button.opacity.wait_for(1.0)
        return copy_button

    def _get_popover(self):
        """Return text edit popover menu."""
        return self.wait_select_single(objectName='text_input_contextmenu')

    def _select_url_text(self):
        """Long press on the address bar to show the contextual popup."""
        self.pointing_device.click_object(self.address_bar.text_field)
        self.address_bar.text_field.activeFocus.wait_for(True)
        self.pointing_device.click_object(
            self.address_bar.text_field, press_duration=1)

    def copy_url_text(self):
        """Click the 'copy' button on the address bar's contextual popup."""
        self._select_url_text()
        copy_button = self._get_popup_copy_button()
        self.pointing_device.click_object(copy_button)

    def get_html_section_dimensions(self):
        """ Retrieve the (x, y. width, height) for the html section, not
        including the chrome bar section. When the keyboard is displayed it is
        not included as part these dimensions.
        """
        (x1, y1, w1, h1) = self.get_current_webview().globalRect
        (x2, y2, w2, h2) = self._get_chrome().globalRect

        return x1, y2 + h2, w1, h1 - h2

    def _get_keyboard_area(self):
        return self.wait_select_single('KeyboardRectangle')

    def is_keyboard_displayed(self):
        return self._get_keyboard_area().state == 'shown'

    def is_keyboard_hidden(self):
        return self._get_keyboard_area().state == 'hidden'

    def _get_bookmark_icon(self):
        """Return the bookmark icon on the address bar."""
        bookmark_toggle = self.address_bar.wait_select_single(
            objectName='bookmarkToggle')
        return bookmark_toggle.wait_select_single('Icon', visible=True)

    def is_current_page_bookmarked(self):
        """Return boolean representing if current loaded page is bookmarked."""
        return self._get_bookmark_icon().name == 'starred'

    def _bookmark_current_page(self):
        self.pointing_device.click_object(self._get_bookmark_icon())
        bookmark_options = self.get_bookmark_options()
        bookmark_options.click_dismiss_button()
        bookmark_options.wait_until_destroyed()

    def bookmark_current_page(self):
        """Bookmark the currently loaded page, if its not already."""
        if not self.is_current_page_bookmarked():
            self._bookmark_current_page()


class Chrome(CustomProxyObjectBase):

    def __init__(self, *args):
        super().__init__(*args)
        self.address_bar = self._get_address_bar()

    def _get_address_bar(self):
        return self.select_single(AddressBar)

    def go_back(self):
        back_button = self._get_back_button()
        back_button.enabled.wait_for(True)
        self.pointing_device.click_object(back_button)

    def _get_back_button(self):
        return self.select_single("ChromeButton", objectName="backButton")

    def is_back_button_enabled(self):
        back_button = self._get_back_button()
        return back_button.enabled

    def go_forward(self):
        forward_button = self._get_forward_button()
        forward_button.enabled.wait_for(True)
        self.pointing_device.click_object(forward_button)

    def _get_forward_button(self):
        return self.select_single("ChromeButton", objectName="forwardButton")

    def is_forward_button_enabled(self):
        forward_button = self._get_forward_button()
        return forward_button.enabled

    def get_drawer_button(self):
        return self.select_single("ChromeButton", objectName="drawerButton")

    def get_drawer(self):
        return self.wait_select_single("QQuickItem", objectName="drawer",
                                       clip=False)

    def get_drawer_action(self, actionName):
        drawer = self.get_drawer()
        return drawer.select_single(objectName=actionName, visible=True)

    def get_tabs_bar(self):
        return self.select_single(TabsBar)

    def get_find_next_button(self):
        return self.select_single("ChromeButton",
                                  objectName="findNextButton")

    def get_find_prev_button(self):
        return self.select_single("ChromeButton",
                                  objectName="findPreviousButton")


class BookmarksView(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'BookmarksView')

    def get_bookmarks_urls(self):
        """Return all the pages (urls) that are shown in the bookmarks page"""
        return self.wait_select_many('UrlDelegate', visible=True)

    def delete_bookmark(self, url):
        bookmark_delegate = self.wait_select_single(UrlDelegate, url=url)
        bookmark_delegate.swipe_to_delete()

    def get_done_button(self):
        """Get the 'Done' button."""
        return self.wait_select_single('Button', objectName='doneButton')

    def done(self):
        """Press 'Done' button."""
        done_button = self.get_done_button()
        self.pointing_device.click_object(done_button)


class UrlPreviewGrid(CustomProxyObjectBase):
    def get_delegates(self):
        return sorted(self.select_many("UrlPreviewDelegate"),
                      key=lambda delegate: delegate.globalRect.y)

    def get_urls(self):
        return [delegate.url for delegate in self.get_delegates()]


class BookmarksFoldersView(CustomProxyObjectBase):

    def get_delegates(self):
        return sorted(self.select_many(objectName="bookmarkFolderDelegate"),
                      key=lambda delegate: delegate.globalRect.y)

    def get_folder_delegate(self, folder):
        return self.select_single(objectName="bookmarkFolderDelegate",
                                  folderName=folder)

    def get_urls_from_folder(self, folder):
        return sorted(folder.select_many(UrlDelegate),
                      key=lambda delegate: delegate.globalRect.y)

    def get_header_from_folder(self, folder):
        return folder.wait_select_single(objectName="bookmarkFolderHeader")


class NewTabView(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'NewTabView')

    def get_bookmarks_more_button(self):
        """Return the 'More' button.

        NOTE the webbrowser-app emulator's objectName doesn't change,
        but its label does--we query the button's 'text' to be thorough.

        """
        return self.select_single(
            objectName='bookmarks.moreButton', text='More')

    def get_bookmarks_less_button(self):
        """Return the 'Less' button.

        NOTE the webbrowser-app emulator's objectName doesn't change,
        but its label does--we query the button's 'text' to be thorough.

        """
        return self.select_single(
            objectName='bookmarks.moreButton', text='Less')

    def show_more(self):
        """Tap the 'More' button."""
        self.pointing_device.click_object(self.get_bookmarks_more_button())

    def show_less(self):
        """Tap the 'Less' button."""
        self.pointing_device.click_object(self.get_bookmarks_less_button())

    def get_bookmark_urls(self):
        return [delegate.url for delegate in self.get_bookmark_delegates()]

    def get_homepage_bookmark(self):
        return self.select_single(UrlDelegate, objectName="homepageBookmark")

    def get_bookmarks_list(self):
        return self.select_single(objectName="bookmarksList")

    def get_bookmark_delegates(self):
        list = self.get_bookmarks_list()
        return sorted(list.select_many(UrlDelegate),
                      key=lambda delegate: delegate.globalRect.y)

    def get_top_sites_list(self):
        return self.select_single(UrlPreviewGrid, objectName="topSitesList")

    def get_notopsites_label(self):
        return self.select_single(objectName="notopsites")

    def get_top_site_items(self):
        return self.get_top_sites_list().get_delegates()

    def get_bookmarks_folder_list_view(self):
        return self.wait_select_single(BookmarksFoldersView)

    def get_bookmarks(self, folder_name):
        # assumes that the "more" button has been clicked
        folders = self.get_bookmarks_folder_list_view()
        folder_delegate = folders.get_folder_delegate(folder_name)
        return folders.get_urls_from_folder(folder_delegate)

    def get_folder_names(self):
        folders = self.get_bookmarks_folder_list_view().get_delegates()
        return [folder.folderName for folder in folders]


class UrlDelegate(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'UrlDelegate')

    def select(self):
        self.pointing_device.click_object(self)

    def _drag_pointing_device_to_delete(self):
        x, y, width, height = self.globalRect
        start_x = x + (width * 0.2)
        stop_x = x + (width * 0.8)
        start_y = stop_y = y + (height // 2)

        self.pointing_device.drag(start_x, start_y, stop_x, stop_y)
        self.dragging.wait_for(False)

    def _get_delete_button(self):
        """Return the delete button revealed on slide to right."""
        return self.wait_select_single(
            objectName='actionbutton_leadingAction.delete'
        )

    def swipe_to_delete(self):
        """Delete a bookmark by swipe."""
        self._drag_pointing_device_to_delete()
        self.pointing_device.click_object(self._get_delete_button())
        self.wait_until_destroyed()


class ContentPeerPicker13(CustomProxyObjectBase):
    """CPO to represent content app selection dialog."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'ContentPeerPicker13',
            objectName='contentPeerPicker')

    def _get_button(self, text):
        return self.select_many(
            ap_result_sort_keys=KEY_ORDER_BY_Y, objectName='label',
            text=text)[0].get_parent()

    def _get_gallery_button(self):
        return self._get_button('Gallery')

    def _get_notes_button(self):
        return self._get_button('Notes')

    def _get_tagger_button(self):
        return self._get_button('Tagger')

    def _get_media_player_button(self):
        return self._get_button('Media Player')

    def pick_gallery(self):
        self.pointing_device.click_object(self._get_gallery_button())
        self.wait_until_destroyed()

    def pick_notes(self):
        self.pointing_device.click_object(self._get_notes_button())
        self.wait_until_destroyed()

    def pick_tagger(self):
        self.pointing_device.click_object(self._get_tagger_button())
        self.wait_until_destroyed()

    def pick_media_player(self):
        self.pointing_device.click_object(self._get_media_player_button())
        self.wait_until_destroyed()


class Dialog(CustomProxyObjectBase):
    """CPO to represent download options dialog."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(
            path, state, APP_PATH_ROOT, b'Dialog',
            objectName='downloadOptionsDialog')

    def _get_cancel_button(self):
        return self.select_single(objectName='cancelDownloadButton')

    def _get_choose_application_button(self):
        return self.select_single(objectName='chooseAppButton')

    def _get_download_button(self):
        return self.select_single(objectName='downloadFileButton')

    def cancel(self):
        self.pointing_device.click_object(self._get_cancel_button())
        self.wait_until_destroyed()

    def choose_application(self):
        self.pointing_device.click_object(
            self._get_choose_application_button())
        self.wait_until_destroyed()

    def download(self):
        self.pointing_device.click_object(self._get_download_button())
        self.wait_until_destroyed()


class BrowserPage(CustomProxyObjectBase):

    def get_header(self):
        return self.select_single(PageHeader)


class DownloadsPage(BrowserPage):
    """CPO to represent downloads page."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(path, state, APP_PATH_ROOT,
                                    b'DownloadsPage')

    @property
    def header(self):
        return self.get_header()

    def get_download_delegate(self, url):
        """Return list delegate by url."""
        return self.select_single(DownloadDelegate, url=url)

    def cancel_download(self, url):
        """Cancel the requested download by url"""
        download_item = self.get_download_delegate(url=url)
        download_item.cancel()

    def delete_history_item(self, url):
        """Delete a downloads history item by associated url"""
        self.header.click_edit_button()
        item = self.get_download_delegate(url=url)
        self.pointing_device.click_object(item.get_checkbox())
        self.header.click_delete_button()
        item.wait_until_destroyed()

    def clear_history(self):
        """Clear all downloads history"""
        self.header.click_edit_button()
        self.header.click_select_all_button()
        self.header.click_delete_button()

    def get_visible_downloads_count(self):
        return len(self.select_many(DownloadDelegate))

    def is_download_finished(self, url):
        """Return bool representing if the download completed"""
        download_item = self.get_download_delegate(url=url)
        return not download_item.incomplete

    def go_back_to_main_screen(self):
        """Go back to the main browser screen"""
        self.header.click_back_button()
        self.wait_until_destroyed()


class DownloadDelegate(CustomProxyObjectBase):
    """CPO to represent a single Download delegate."""

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(path, state, APP_PATH_ROOT,
                                    b'DownloadDelegate')

    def _get_cancel_button(self):
        return self.select_single('Button', text='Cancel')

    def cancel(self):
        """Cancel the download"""
        self.pointing_device.click_object(self._get_cancel_button())
        self.wait_until_destroyed()

    def get_checkbox(self):
        """Return the checkbox item."""
        return self.wait_select_single(objectName='listitem_select')


class PageHeader(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(path, state, APP_PATH_ROOT,
                                    b'PageHeader')

    def _get_button(self, icon_name):
        return self.wait_select_single(iconName=icon_name, visible=True)

    def click_button(self, icon_name):
        self.pointing_device.click_object(self._get_button(icon_name))

    def click_edit_button(self):
        self.click_button('edit')

    def click_delete_button(self):
        self.click_button('delete')

    def click_select_all_button(self):
        self.click_button('select')

    def click_back_button(self):
        button = self.select_single(objectName="back_button")
        self.pointing_device.click_object(button)


class BookmarkOptions(CustomProxyObjectBase):

    @classmethod
    def validate_dbus_object(cls, path, state):
        return validate_dbus_object(path, state, APP_PATH_ROOT,
                                    b'BookmarkOptions')

    def _get_ok_button(self):
        return self.select_single(objectName='bookmarkOptions.okButton')

    def add_bookmark(self):
        button = self._get_ok_button()
        self.pointing_device.click_object(button)
        button.wait_until_destroyed()

    def get_title_text_field(self):
        return self.select_single(TextField, objectName="titleTextField")

    def get_save_in_option_selector(self):
        return self.select_single("OptionSelector", currentlyExpanded=False)

    def click_new_folder_button(self):
        button = self.select_single("Button",
                                    objectName="bookmarkOptions.newButton")
        self.pointing_device.click_object(button)

    def click_dismiss_button(self):
        button = self.select_single("Button",
                                    objectName="bookmarkOptions.okButton")
        self.pointing_device.click_object(button)


class Suggestions(CustomProxyObjectBase):

    def get_ordered_entries(self):
        return sorted(self.select_many("Suggestion"),
                      key=lambda item: item.globalRect.y)


class SadTab(CustomProxyObjectBase):

    def click_close_tab_button(self):
        button = self.select_single("Button", objectName="closeTabButton")
        self.pointing_device.click_object(button)

    def click_reload_button(self):
        button = self.select_single("Button", objectName="reloadButton")
        self.pointing_device.click_object(button)


class GeolocationPermissionRequest(CustomProxyObjectBase):

    def get_deny_button(self):
        return self.select_single("Button", objectName="deny")

    def get_allow_button(self):
        return self.select_single("Button", objectName="allow")


class HttpAuthenticationDialog(CustomProxyObjectBase):

    def get_deny_button(self):
        return self.select_single("Button", objectName="deny")

    def get_allow_button(self):
        return self.select_single("Button", objectName="allow")

    def get_username_field(self):
        return self.select_single("TextField", objectName="username")

    def get_password_field(self):
        return self.select_single("TextField", objectName="password")


class MediaAccessDialog(CustomProxyObjectBase):

    def click_deny_button(self):
        button = self.select_single("Button",
                                    objectName="mediaAccessDialog.denyButton")
        self.pointing_device.click_object(button)

    def click_allow_button(self):
        button = self.select_single(
            "Button",
            objectName="mediaAccessDialog.allowButton")
        self.pointing_device.click_object(button)


class TabPreview(CustomProxyObjectBase):

    def select(self):
        area = self.select_single("QQuickMouseArea", objectName="selectArea")
        # click towards the top of the area to ensure we’re not selecting
        # the following preview that might be overlapping
        ca = area.globalRect
        self.pointing_device.move(ca.x + ca.width // 2, ca.y + ca.height // 4)
        self.pointing_device.click()

    def close(self):
        button = self.select_single(objectName="closeButton")
        self.pointing_device.click_object(button)


class TabsList(CustomProxyObjectBase):

    def get_previews(self):
        previews = self.select_many(TabPreview)
        previews.sort(key=lambda tab: tab.globalRect.y)
        return previews


class AddressBar(CustomProxyObjectBase):

    def __init__(self, *args):
        super().__init__(*args)
        self.text_field = self.select_single(
            TextField, objectName='addressBarTextField')

    def focus(self):
        self.pointing_device.click_object(self)
        self.activeFocus.wait_for(True)

    def clear(self):
        self.text_field.clear()

    def go_to_url(self, url):
        self.write(url)
        self.press_key('Enter')

    def write(self, text, clear=True):
        self.text_field.write(text, clear)

    def press_key(self, key):
        self.text_field.keyboard.press_and_release(key)

    def click_action_button(self):
        button = self.select_single("QQuickMouseArea",
                                    objectName="actionButton")
        self.pointing_device.click_object(button)

    def get_bookmark_toggle(self):
        return self.select_single("QQuickMouseArea",
                                  objectName="bookmarkToggle")

    def get_find_in_page_counter(self):
        return self.select_single(objectName="findInPageCounter")


class TabsBar(CustomProxyObjectBase):

    def click_new_tab_button(self):
        button = self.select_single("QQuickMouseArea",
                                    objectName="newTabButton")
        self.pointing_device.click_object(button)

    def get_tabs(self):
        return self.select_many("QQuickMouseArea", objectName="tabDelegate")

    def get_tab(self, index):
        return self.select_single("QQuickMouseArea", objectName="tabDelegate",
                                  tabIndex=index)

    def select_tab(self, index):
        self.pointing_device.click_object(self.get_tab(index))

    def close_tab(self, index):
        tab = self.get_tab(index)
        close_button = tab.select_single(objectName="closeButton")
        self.pointing_device.click_object(close_button)


class Toolbar(CustomProxyObjectBase):

    def click_button(self, name):
        self.isFullyShown.wait_for(True)
        button = self.select_single("Button", objectName=name)
        self.pointing_device.click_object(button)

    def click_action(self, name):
        self.isFullyShown.wait_for(True)
        action = self.select_single("ToolbarAction", objectName=name)
        self.pointing_device.click_object(action)


class BrowserPage(CustomProxyObjectBase):

    def get_header(self):
        return self.select_single(PageHeader)


class SettingsPage(BrowserPage):

    def get_searchengine_entry(self):
        return self.select_single(objectName="searchengine")

    def get_searchengine_page(self):
        return self.wait_select_single(objectName="searchEnginePage")

    def get_homepage_entry(self):
        return self.select_single(objectName="homepage")

    def get_restore_session_entry(self):
        return self.select_single(objectName="restoreSession")

    def get_privacy_entry(self):
        return self.select_single(objectName="privacy")

    def get_privacy_page(self):
        return self.wait_select_single(objectName="privacySettings")

    def get_reset_settings_entry(self):
        return self.select_single(objectName="reset")


class DialogWrapper(object):
    def __init__(self, dialog):
        self.dialog = dialog

        self.text = self.dialog.text
        self.wait_until_destroyed = self.dialog.wait_until_destroyed
        self.visible = self.dialog.visible


class AlertDialog(DialogWrapper):
    def get_ok_button(self):
        return self.dialog.select_single("Button", objectName="okButton")


class HistoryView(CustomProxyObjectBase):

    def get_domain_entries(self):
        return sorted(self.select_many("UrlDelegate"),
                      key=lambda item: item.globalRect.y)


class HistoryViewWide(CustomProxyObjectBase):

    def get_entries(self):
        return sorted(self.select_many("UrlDelegate"),
                      key=lambda item: item.globalRect.y)

    def get_search_field(self):
        return self.select_single(objectName="searchQuery")


class ExpandedHistoryView(CustomProxyObjectBase):

    def get_header(self):
        return self.select_single(objectName="header")

    def get_entries(self):
        return sorted(self.select_many("UrlDelegate",
                                       objectName="entriesDelegate"),
                      key=lambda item: item.globalRect.y)


class ContextMenuBase(CustomProxyObjectBase):

    def get_title_label(self):
        return self.select_single(objectName="titleLabel")

    def get_visible_actions(self):
        return self.select_many("Empty", visible=True)

    def get_action(self, objectName):
        name = objectName + "_item"
        return self.select_single("Empty", objectName=name)

    def click_action(self, objectName):
        name = objectName + "_item"
        action = self.select_single("Empty", visible=True,
                                    enabled=True, objectName=name)
        self.pointing_device.click_object(action)
        self.wait_until_destroyed()


class ContextMenuWide(ContextMenuBase):
    pass


class ContextMenuMobile(ContextMenuBase):

    def click_cancel_action(self):
        action = self.select_single("Empty", objectName="cancelAction")
        self.pointing_device.click_object(action)


class BeforeUnloadDialog(DialogWrapper):
    def get_leave_button(self):
        return self.dialog.select_single("Button", objectName="leaveButton")

    def get_stay_button(self):
        return self.dialog.select_single("Button", objectName="stayButton")


class ConfirmDialog(DialogWrapper):
    def get_cancel_button(self):
        return self.dialog.select_single("Button", objectName="cancelButton")

    def get_ok_button(self):
        return self.dialog.select_single("Button", objectName="okButton")


class PromptDialog(DialogWrapper):
    def get_cancel_button(self):
        return self.dialog.select_single("Button", objectName="cancelButton")

    def get_input_textfield(self):
        return self.dialog.select_single("TextField",
                                         objectName="inputTextField")

    def get_ok_button(self):
        return self.dialog.select_single("Button", objectName="okButton")
