# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import logging
import requests

from math import ceil
from requests.exceptions import ConnectionError
from retrying import retry

URL_PROD_PRACTITEST = 'https://prod.practitest.com/api'
URL_ALL_TESTS = URL_PROD_PRACTITEST + '/tests.json'
URL_TEST = URL_PROD_PRACTITEST + '/tests/{}.json'

logger = logging.getLogger(__name__)


class PractitestSession:

    def __init__(self, api_token, project_id=None):
        """
        Initialise the PractitestSession object.
        :param api_token: Token to use for api authentication.
        :param project_id: Project ID to use. If no value is given then the
                           first project of the account will be used.
        """
        if not api_token:
            raise ValueError('Must specify a Practitest api token.')
        self.project_id = project_id
        self.headers = self._compose_headers(api_token)

    @staticmethod
    def _compose_headers(api_token):
        """Return auth headers including api token."""
        auth_header = {
            'Authorization': 'custom api_token={}'.format(api_token),
            'Content-Type': 'application/json'
        }
        return auth_header

    @staticmethod
    def _validate_response_status(response):
        """Validate response is ok and raise custom exception if not."""
        if not response.ok:
            err_str = json.loads(response.text)['error']
            raise ValueError('Practitest error: {e}'.format(e=err_str))

    @retry(retry_on_exception=lambda e: isinstance(e, ConnectionError),
           stop_max_attempt_number=4,
           wait_exponential_max=10000,
           wait_exponential_multiplier=1000)
    def _get(self, url, params=None):
        """Get the given url, retrying on failure."""
        params = {} if params is None else params
        if self.project_id:
            params['project_id'] = self.project_id
        response = requests.get(
            url=url,
            headers=self.headers,
            data=json.dumps(params))
        self._validate_response_status(response)
        return response

    def _get_all(self, url, params=None):
        """Get all pages of results for the given url."""
        params = {} if params is None else params
        response = self._get(url, params).json()
        data = response['data']
        pagination = response['additional_data']['pagination']
        current_page = 1
        total_pages = self._get_total_pages(pagination)
        while current_page < total_pages:
            current_page += 1
            params['page'] = str(current_page)
            response = self._get(url, params)
            data.extend(response.json()['data'])
        return data

    @staticmethod
    def _get_total_pages(pagination):
        """Return the total number of pages required for all results."""
        return ceil(
            int(pagination['total_entities'])/int(pagination['limit']))

    def get_tests(self, filter_id):
        """Return a list of tests for specified filter_id.

        Retrieves all tests associated with account.

        See http://support.practitest.com/customer/portal/articles/1108202-tests-api .  # NOQA

        :param filter_id: ID of the Practitest filter to use.
        :return: List of JSON formatted test descriptions.

        """
        params = {}
        if filter_id:
            params['filter_id'] = filter_id
        return self._get_all(URL_ALL_TESTS, params)

    def get_test(self, test_id):
        """Return a single test description.

        See http://support.practitest.com/customer/portal/articles/1108202-tests-api .  # NOQA

        :param test_id: Practitest test id.
        :return: JSON formatted test description.

        """
        return self._get(URL_TEST.format(test_id)).json()
