#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import shlex
import shutil
import stat
import subprocess

from fixtures import TempDir

KERNEL_FILTER = 'casper/vmlinuz*'
# initial ram disk
CASPER_INITRD = 'casper/initrd'
CASPER_INITRDLZ = 'casper/initrd.lz'

KERNEL_OPTS = (
    'boot=casper DEBCONF_DEBUG=developer -- debconf/priority=critical '
    'locale=en_US console-setup/ask_detect=false console-setup/layoutcode=us '
    'noprompt automatic-ubiquity'
)
QEMU_LAUNCH_OPTS = (
    '{qemu} -m {ram} -smp {cpu} -localtime '
    '-net nic,model=virtio -net user -cpu core2duo -enable-kvm '
    '-drive file={disk_img},if=virtio '
)
QEMU_INSTALL_OPTS = (
    '-cdrom {iso} -boot d -initrd {initrd} -kernel {kernel} '
    '-append "{cmdline}" -pidfile {workdir}/qemu.pid -boot menu=off '
    '-no-reboot '
)
QEMU_DISPLAY_OPTS = (
    '-display sdl -vga qxl '
)
QEMU_SOUND_OPTS = (
    '-soundhw all '
)
QEMU_DISPLAY_HEADLESS = (
    '-display none '
)
QEMU_SYSTEM_AMD64 = 'qemu-system-x86_64'
QEMU_SYSTEM_I386 = 'qemu-system-i386'
ARCH_AMD64 = 'amd64'
ARCH_I386 = 'i386'

PRESEED = 'preseed.cfg'

SERIAL_CONFIG_SHELL = '-serial unix:{workdir}/qemu.ttyS{port},server,nowait '
SERIAL_CONFIG_LOG = '-serial file:{workdir}/qemu.ttyS{port} '
SERIAL_SHELL = [SERIAL_CONFIG_SHELL, SERIAL_CONFIG_SHELL]
SERIAL_LOG = [SERIAL_CONFIG_LOG, SERIAL_CONFIG_LOG]


def get_architecture():
    """Return architecture string for system."""
    return subprocess.check_output(
        ['dpkg', '--print-architecture']).decode().strip()


def get_qemu_path():
    """Return path of qemu-system executable for system."""
    if get_architecture() == ARCH_AMD64:
        target = QEMU_SYSTEM_AMD64
    else:
        target = QEMU_SYSTEM_I386
    return subprocess.check_output(['which', target]).decode().strip()


def get_qemu_launch_command(work_dir, disk_img, ram, cpu, headless,
                            serial_config=None, iso_path=None, kernel=None,
                            kernel_opts=None):
    """Return command to launch qemu process using optional install parameters.
    :param work_dir: Working directory to use.
    :param disk_img: Path of the disk image file used for installation.
    :param ram: Amount of ram allocated to qemu.
    :param cpu: Number of cpus allocated to qemu.
    :param headless: Whether to run installer in headless mode or not.
    :param serial_config: List of serial parameters to specify.
    :param iso_path: Path of iso file to use for installation.
    :param kernel: Path of kernel to use.
    :param kernel_opts: Options to launch kernel with.
    :return: Qemu launch command string.
    """
    # Create command base with resource parameters
    cmd = QEMU_LAUNCH_OPTS.format(
        qemu=get_qemu_path(), ram=ram, cpu=cpu, disk_img=disk_img,
        workdir=work_dir)
    # Add display parameters
    cmd += get_display_args(headless)
    # Add serial port configuration
    for port, options in enumerate(serial_config or []):
        cmd += options.format(workdir=work_dir, port=port)
    # Add install options if specified
    if iso_path and kernel and kernel_opts:
        initrd_path = os.path.join(work_dir, CASPER_INITRD)
        kernel_path = os.path.join(work_dir, kernel)
        cmd += QEMU_INSTALL_OPTS.format(
            iso=iso_path, initrd=initrd_path, kernel=kernel_path,
            cmdline=kernel_opts, workdir=work_dir)
    return shlex.split(cmd)


def get_kernel(iso_path):
    """Return path of kernel from specified iso file."""
    return subprocess.check_output(
        ['bsdtar', '-tf', iso_path, KERNEL_FILTER]).decode().strip()


def get_qemu_pid(working_dir):
    """Return process id of current qemu. In case of multiple qemus running
    on a single system, read the pid value from file in working directory to
    ensure the correct one is used.
    """
    path = os.path.join(working_dir, 'qemu.pid')
    with open(path) as pidfile:
        pid = pidfile.read().strip()
    if not pid.isdigit():
        raise ValueError('No valid pid for qemu process in {}.'.format(path))
    return pid


def wait_for_qemu_stopped(working_dir):
    """Wait for required qemu process to stop."""
    pid = get_qemu_pid(working_dir)
    subprocess.check_call(['wait', pid], shell=True)


def build_qemu(img_path, iso_path, cpu, ram, img_disk_size, username, password,
               display_name, headless, verbose):
    """
    Build a qemu image file using specified ISO file as installer.
    :param img_path: Path of img file to use.
    :param iso_path: Path of iso file to use for installation.
    :param cpu: Number of cpus to allocate to qemu process.
    :param ram: Amount of ram to allocate to qemu process.
    :param img_disk_size: Maximum size of disk for img file.
    :param username: Username to setup during installation.
    :param password: Password of user account.
    :param display_name: Display name of user account.
    :param headless: Whether to show installer UI during installation process.
    """
    create_disk_image(img_path, img_disk_size)
    with TempDir() as working_dir:
        kernel = get_kernel(iso_path)
        extract_kernel(iso_path, kernel, working_dir.path)
        prepare_initrd(working_dir.path, username, password, display_name)
        # launch qemu with installer options to start install process
        launch_qemu(
            working_dir.path, img_path, ram, cpu, headless, SERIAL_LOG,
            iso_path, kernel, KERNEL_OPTS)
        wait_for_qemu_stopped(working_dir.path)
        if verbose:
            print_logs(working_dir.path, label=' installer logs ')
        # Boot the qemu again to run the first-run installer script
        launch_qemu(working_dir.path, img_path, ram, cpu, headless, SERIAL_LOG)
        if verbose:
            print_logs(working_dir.path, label=' qemu setup logs ')


def extract_kernel(iso_path, kernel_path, dst_path):
    """Extract kernel from iso to specified path."""
    subprocess.check_call(
        ['bsdtar', 'xf', iso_path, '-C', dst_path, CASPER_INITRDLZ,
         kernel_path])


def create_disk_image(disk_img, disk_size):
    """Create a disk image of specified name and size."""
    subprocess.check_call(
        ['qemu-img', 'create', '-f', 'qcow2', disk_img, disk_size])


def create_preseed(path, username, password, display_name):
    """Create a preseed file to use for automated installation.
    :param path: Path of preseed file to create.
    :param username: Username to setup during installation.
    :param password: Password of user account.
    :param display_name: Display name of user account.
    """
    with open(path, 'w+') as preseed:
        preseed.write(
            get_preseed_content(username, password, display_name))


def get_preseed_content(username, password, display_name):
    """Return preseed content for automated installation and create required
    user account.
    :param username: Username to setup during installation.
    :param password: Password of user account.
    :param display_name: Display name of user account.
    """
    preseed_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), PRESEED)
    with open(preseed_path, 'r') as f:
        content = f.read()
    return content.format(
        display_name=display_name, username=username, password=password)


def get_qemu_scripts_dir():
    """Return path of the source directory containing qemu setup scripts."""
    src_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(src_dir, 'qemu-scripts')


def copy_qemu_scripts_to_path(path):
    """Copy all qemu scripts to required path and ensure correct permissions.
    :param path: Target path to copy scripts to.
    """
    src_dir = get_qemu_scripts_dir()
    mask = (stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH |
            stat.S_IXOTH)
    for src_name in os.listdir(src_dir):
        src = os.path.join(src_dir, src_name)
        dst = os.path.join(path, src_name)
        shutil.copy2(src, dst)
        os.chmod(dst, mask)


def prepare_initrd(working_dir, username, password, display_name):
    """
    Prepare the initial ram disk. This is used to add all required custom test
    actions to the installation process and adds the preseed file to the root
    of initrd to allow installation to be automated.
    :param working_dir: Working directory to use.
    :param username: Username to setup during installation.
    :param password: Password of user account.
    :param display_name: Display name of user account.
    :param auto_login: Whether to enable auto login for user.
    :return: Path of initrd file with installer scripts added.
    """
    initrd_path = os.path.join(working_dir, CASPER_INITRD)
    initrdlz_path = os.path.join(working_dir, CASPER_INITRDLZ)
    initrd_work_dir = os.path.join(working_dir, 'initrd')
    # create dir to extract initrd, change to it and then extract files
    os.mkdir(initrd_work_dir)
    os.chdir(initrd_work_dir)
    ps = subprocess.Popen(['xzcat', initrdlz_path], stdout=subprocess.PIPE)
    subprocess.check_call(['cpio', '--quiet', '-id'], stdin=ps.stdout)
    ps.wait()
    # copy the qemu scripts into initrd
    copy_qemu_scripts_to_path(initrd_work_dir)
    casper_prov = os.path.join(initrd_work_dir, 'qemu-casper-provision')
    add_script_to_casper_setup(initrd_work_dir, casper_prov)
    # create the preseed file
    preseed_path = os.path.join(initrd_work_dir, PRESEED)
    create_preseed(preseed_path, username, password, display_name)
    # Copy all files back to the archive
    with open(initrd_path, 'w') as initrd:
        ps = subprocess.Popen(['find', '.'], stdout=subprocess.PIPE)
        subprocess.check_call(
            ['cpio', '--quiet', '-o', '-H', 'newc'],
            stdin=ps.stdout, stdout=initrd)
        ps.wait()
    return initrd_path


def add_script_to_casper_setup(initrd_path, script_path):
    """Add specified script into the casper setup phase of installation.
    :param initrd_path: Path of initial ram disk to add the scripts.
    :param script_path: Path of script file to add.
    """
    casper_dir = os.path.join(initrd_path, 'scripts', 'casper-bottom')
    script_name = os.path.basename(script_path)
    script_dst = os.path.join(casper_dir, script_name)
    shutil.copy2(script_path, script_dst)
    order_file = os.path.join(casper_dir, 'ORDER')
    with open(order_file, 'r+') as f:
        content = f.read()
        f.seek(0)
        f.write(
            '/scripts/casper-bottom/{} "$@"\n[ -e /conf/param.conf ] && . '
            '/conf/param.conf\n'.format(script_name) + content
        )


def get_display_args(headless):
    """Return qemu-system display arguments based on headless parameter.
    :param headless: Whether qemu-system should run in headless mode or not.
    :return: Display parameters for required display state.
    """
    if headless:
        return QEMU_DISPLAY_HEADLESS
    else:
        return QEMU_DISPLAY_OPTS + QEMU_SOUND_OPTS


def launch_qemu(working_dir, disk_image_path, ram, cpu, headless,
                serial_config=None, iso_path=None, kernel=None,
                kernel_opts=None):
    """Launch qemu-system to install the iso file into the disk image.
    :param working_dir: Working directory to use.
    :param disk_image_path: Path of the disk image file used for installation.
    :param ram: Amount of ram allocated to qemu.
    :param cpu: Number of cpus allocated to qemu.
    :param headless: Whether to run installer in headless mode or not.
    :param serial_config: List of serial parameters to specify.
    :param iso_path: Path of iso file to use for installation.
    :param kernel: Path of kernel to use.
    :param kernel_opts: Options to launch kernel with.

    """
    cmd = get_qemu_launch_command(
        working_dir, disk_image_path, ram, cpu, headless, serial_config,
        iso_path, kernel, kernel_opts)
    print(' '.join(cmd))
    subprocess.check_call(cmd)


def print_logs(working_dir, label=''):
    """Print the logs recorded in the working directory to stdout."""
    with open(os.path.join(working_dir, 'qemu.ttyS0'), 'rb') as log_file:
        lines = log_file.readlines()
    print('******{}*****'.format(label))
    for line in lines:
        print(line.decode().strip())
    print('***********')
