#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import fnmatch
import glob
import hashlib
import os
import re
import requests
import sys


def get_download_url(iso_url, arch):
    """Get the ISO file download url.
    :param iso_url: The url to examine. This could be spefified as:
        - A url to a specific ISO file, in which case the same url will
          be returned. e.g.
        http://cdimage.ubuntu.com/daily-live/current/yakkety-desktop-amd64.iso
        - A url to a page displaying ISOs for different architectures, in
          which case the file of required architecture will be selected. e.g.
          http://cdimage.ubuntu.com/daily-live/current/
    :param arch: The required architecture of ISO to download. This will be
        used to select an ISO from a download page when a specific ISO file
        is not provided.
    :return: ISO file download URL.
    """
    if iso_url.endswith('.iso'):
        # A specific iso url has been specified so use that
        return iso_url
    # Otherwise, calculate the url based on release name and architecture
    iso_arch = '-desktop-{}.iso'.format(arch)
    iso_request = requests.get(iso_url)
    match = re.search(r'"(.+){}"'.format(iso_arch), iso_request.text)
    name = match.group(1)
    iso_name = name + iso_arch
    return os.path.join(iso_url, iso_name)


def get_download_sha256(iso_url, arch):
    """Return sha256 value of ISO file of specified architecture.
    The sha256 value will be downloaded from the SHA256SUMS file on the ISO
    page.
    :param iso_url: URL for a specific ISO file or ISO page.
    :param arch: The required architecture, used to select ISO from page.
    :return: SHA256 value for required ISO.
    """
    if iso_url.endswith('.iso'):
        iso_page_url = os.path.dirname(iso_url)
    else:
        iso_page_url = iso_url
    sha256_url = os.path.join(iso_page_url, 'SHA256SUMS')
    sha_request = requests.get(sha256_url)
    match = re.search(r'(.+) \*.+{}'.format(arch), sha_request.text)
    return match.group(1)


def download_file(url, target_file):
    """Download file at url and save to target_file."""
    with open(target_file, 'wb') as f:
        print('Downloading {} to {}'.format(url, target_file))
        response = requests.get(url, stream=True)
        total_length = response.headers.get('content-length')
        if total_length is None:
            f.write(response.content)
        else:
            dl = 0
            total_length = int(total_length)
            for data in response.iter_content(chunk_size=1024):
                dl += len(data)
                f.write(data)
                done = int(50 * dl / total_length)
                sys.stdout.write('\r[{}{}]'.format('='*done, ' '*(50-done)))
                sys.stdout.flush()


def get_sha256_file_path(file_path):
    """Return path of file used to save SHA256 value of specified file."""
    return file_path + '.sha256'


def get_sha256_from_file(file_path):
    """Return SHA256 value for specified file. First attempt to read the value
    from cached file. If the saved value is not available then calculate the
    SHA256 value and save it for future reference.
    """
    if file_path:
        try:
            with open(get_sha256_file_path(file_path), 'r') as f:
                sha = f.readline().strip()
        except TypeError:
            return None
        except FileNotFoundError:
            # .sha256 file is not present for some reason
            # re-cacluate the sha256 value and save to file
            sha = calculate_file_sha256(file_path)
            save_sha256_to_file(file_path, sha)
        return sha
    return None


def save_sha256_to_file(file_path, sha256):
    """Save the specified SHA256 value to file with .sha256 extension."""
    with open(get_sha256_file_path(file_path), 'w') as f:
        f.write(sha256)


def get_local_img_file_path(img_path, max_count):
    """Return the path of a local img file to use.
    :param img_path: This could be:
        - Path to a local img file, in which case an existing file will be
          deleted if it already exists.
        - Path to a local directory, in which case a unique name will be given
          to the file in that directory.
    :param max_count: If img_path is a local directory, max_count is the max
        number of files to keep in that directory. The oldest files above this
        value will be deleted. -1 indicates no files should be automatically
        deleted.
    :return: Path name for img file.
    """
    if os.path.isfile(img_path):
        # Delete the existing image file to create a new one of same name
        os.remove(img_path)
        return img_path
    # Make sure root dir exists
    img_dir = os.path.dirname(img_path)
    if not os.path.exists(img_dir):
        os.makedirs(img_dir)
    if '.' in os.path.basename(img_path):
        # A file name has been specified, but it doesn't exist yet which is ok
        return img_path
    else:
        # Path is a dir, so make sure it exists and get a unique filename
        return get_img_file_name_for_dir(img_path, max_count)


def get_img_file_name_for_dir(dir_path, max_count):
    """Return img file name for specified directory, cleaning any old files
    if required.
    :param dir_path: Path to a local directory
    :param max_count: Max number of files to keep in directory. The oldest
        files above this value will be deleted. -1 indicates no files should be
        automatically deleted.
    :return: Path name for img file.
    """
    if os.path.exists(dir_path):
        # dir already exists, so clean any old images required.
        # max_count is 1 less, to account for the image that is about
        # to be created.
        if max_count > 0:
            max_count = max_count - 1
        clean_folder(dir_path, '*.img', max_count)
    else:
        os.mkdir(dir_path)
    return get_unique_file_name(dir_path, 'disk.img')


def clean_cache_folder(cache_dir, max_count):
    """Clean the iso and img files from cache dir."""
    clean_iso_cache_folder(cache_dir, max_count)
    clean_img_cache_folder(cache_dir, max_count)


def clean_iso_cache_folder(cache_dir, max_count):
    """Clean the iso files from cache dir."""
    clean_folder(os.path.join(cache_dir, 'iso'), '*.iso', max_count)


def clean_img_cache_folder(cache_dir, max_count):
    """Clean the img files from cache dir."""
    clean_folder(os.path.join(cache_dir, 'img'), '*.img', max_count)


def clean_folder(folder_path, filter, max_count):
    """Clean oldest files in folder_path matching filter string so that only
    max_count files remain.
    :param folder_path: Folder to clean.
    :param filter: Filter to identify required files to clean.
    :param max_count: Number of files to leave in place. -1 indicates no files
                      should be deleted.
    """
    # Firstly delete any .fuse_hidden* files which can be left behind
    delete_hidden_files(folder_path)
    # Now clean any remaining files
    if max_count < 0:
        # No files are to be deleted automatically
        return
    files = get_sorted_files(folder_path, filter)
    del_count = len(files) - max_count
    if del_count > 0:
        for del_file in files[:del_count]:
            delete_file_and_sha(del_file)


def delete_hidden_files(folder_path):
    """Remove any temporary files from folder which can get left behind."""
    for del_file in get_sorted_files(folder_path, '.fuse_hidden*'):
        os.remove(del_file)


def delete_file_and_sha(file_path):
    os.remove(file_path)
    sha_file = get_sha256_file_path(file_path)
    if os.path.isfile(sha_file):
        os.remove(sha_file)


def get_local_iso_path(iso, arch, data_folder, auto_download):
    """Return path of local ISO file to use. If a download is required to get
    a newer version, then return None.
    :param iso: Either of the following:
        - A url to a specific iso file, e.g.
        http://cdimage.ubuntu.com/daily-live/current/yakkety-desktop-amd64.iso
        - A url to a page displaying ISOs for different architectures e.g.
          http://cdimage.ubuntu.com/daily-live/current/
        - A path to a local ISO file, e.g.
          /home/user/Downloads/yakkety-desktop-amd64.iso
    :param arch: Architecture for required ISO. Either i386 or amd64. This is
        used to select the correct ISO from a download page if a specific ISO
        is not given.
    :param data_folder: Path to where downloaded ISO files are stored locally.
        This is used to check if any previous downloads are already up-to-date
        or whether a new file needs to be downloaded.
    :param auto_download: Whether downloads should be attempted automatically.
        If False and no suitable ISO file can be found then a FileNotFoundError
        assertion will be raised.
    :returns: Path of local file to use, or None if a download is required to
        get a newer version.
    :raises: FileNotFoundError if no suitable ISO file can be found and
        auto_download is set False.
    """
    # Check if iso is a local iso file
    if os.path.isfile(iso):
        return iso

    iso_data_folder = os.path.join(data_folder, 'iso')
    newest_local_iso = get_newest_iso_file(iso_data_folder, arch)
    if auto_download:
        # ISO download is enabled, so check if a new ISO needs downloading
        return new_download_available(iso, arch, newest_local_iso)
    elif newest_local_iso:
        # Auto download is disabled, just use the latest local iso
        return newest_local_iso
    # Download is disabled and no local iso is found.
    raise FileNotFoundError(
        'No ISO file exists for architecture {} at {} and auto-download '
        'is disabled.'.format(arch, iso))


def new_download_available(iso_page, arch, local_iso):
    """Return local_iso path if no new download is available, None if a newer
    download is available.
    :param iso_page: Url for iso download page.
    :param arch: Required architecture for iso.
    :param local_iso: Path to local iso file for comparison.
    :return: True if new download available, False otherwise.
    """
    download_sha = get_download_sha256(iso_page, arch)
    local_sha = get_sha256_from_file(local_iso)
    if download_sha == local_sha:
        # No new download is availble, so just return local_iso path
        return local_iso
    # A newer download is available, return None to indicate that.
    return None


def download_iso(url, arch, data_folder, max_count):
    """Download and save ISO from specified url.
    :param url: Either of the following:
        - A url to a specific iso file, e.g.
        http://cdimage.ubuntu.com/daily-live/current/yakkety-desktop-amd64.iso
        - A url to a page displaying ISOs for different architectures e.g.
          http://cdimage.ubuntu.com/daily-live/current/
    :param arch: Architecture for required ISO. Either i386 or amd64. This is
        used to select the correct ISO from a download page if a specific ISO
        is not given.
    :param data_folder: Path to save downloaded ISO file.
    :param max_count: Maximum number of ISO files to keep. Any old files above
        this will be deleted. -1 indicates no files should be automatically
        deleted.
    :return: Path to downloaded ISO file.
    """
    iso_folder = os.path.join(data_folder, 'iso')
    if url.endswith('.iso'):
        # A specific iso file has been specified to download
        # so find the parent page to check the sha256 value
        download_url = url
        download_folder = os.path.dirname(download_url)
    else:
        # A download folder page has been specified,
        # so find required iso url for architecture
        download_url = get_download_url(url, arch)
        download_folder = url
    download_name = download_url.split('/')[-1]
    target_path = get_unique_file_name(iso_folder, download_name)
    if max_count > 0:
        # Decrease max count by 1 to account for the iso about to be downloaded
        max_count = max_count - 1
    clean_iso_cache_folder(data_folder, max_count)
    # download the file and only save the sha256 file after the download has
    # completed in case download fails
    download_sha = get_download_sha256(download_folder, arch)
    download_file(download_url, target_path)
    save_sha256_to_file(target_path, download_sha)
    return target_path


def get_newest_iso_file(folder_path, arch):
    """Return path of newest iso file for required architecture in folder."""
    files = get_sorted_files(folder_path, '*{}*.iso'.format(arch))
    if len(files) > 0:
        return files[-1]
    return None


def get_sorted_files(folder_path, name_filter):
    """Return list of files sorted by creation time."""
    return sorted(
        glob.iglob(os.path.join(folder_path, name_filter)),
        key=os.path.getctime)


def calculate_file_sha256(file_path):
    """Calculate SHA256 value for file at specified path."""
    BUF_SIZE = 65536
    sha256 = hashlib.sha256()
    with open(file_path, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha256.update(data)
    return sha256.hexdigest()


def get_unique_file_name(folder_path, file_name):
    """Return unique name for file within directory."""
    if not os.path.isdir(folder_path):
        os.makedirs(folder_path)
    file_root, file_ext = os.path.splitext(file_name)
    ext_filter = '*' + file_ext
    file_name = file_root.split('/')[-1]
    count = len(fnmatch.filter(os.listdir(folder_path), ext_filter)) + 1
    while True:
        path = os.path.join(
            folder_path,
            '{n}-{c}{e}'.format(n=file_name, c=count, e=file_ext))
        if not os.path.exists(path):
            return path
        count += 1
