# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import testscenarios

from unittest import mock

from ubuntu_system_tests.host import commands
from ubuntu_system_tests.common import config
from ubuntu_system_tests.selftests import (
    check_any_item_in_list_contains_value,
    check_any_item_in_list_startswith_value,
    ConfigBaseTestCase,
    DummyArgs,
)


class RunCommandTestCase(ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.args = DummyArgs(target='adb')

    def test_setup_network_command_contains_credentials(self):
        self.set_config_values()

        command = commands._get_device_network_command(
            'test_ssid', 'test_password')

        self.assertTrue(command.startswith(
            'nmcli device wifi connect \'test_ssid\' '
            'password \'test_password\''))


@mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
            return_value=['dep1'])
@mock.patch('ubuntu_system_tests.common.ssh.get_ssh_config_commands')
class BuildAdtRunCommandTestCase(ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.args = DummyArgs(target='adb')

    def test_adt_run_must_be_called_with_short_timeout(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        command = commands.build_adt_run_test_command(
            self.config_stack, self.args)
        command = ' '.join(command)

        run_arguments, _ = command.split('---')
        self.assertIn('--timeout-short=600', run_arguments)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_command_no_rw_file_mount(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        run_cmd = ' '.join(commands.build_adt_run_test_command(
            self.config_stack, self.args))
        self.assertNotIn('mount -o remount,rw /', run_cmd)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_command_ro_file_mount(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        run_cmd = ' '.join(commands.build_adt_run_test_command(
            self.config_stack, self.args))
        self.assertIn('mount -o remount,ro /', run_cmd)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_with_serial(
            self, mock_ssh_config_commands, mock_get_dependencies):
        serial = 'test_serial'
        self.set_config_values(device_serial=serial)

        command = commands.build_adt_run_test_command(
            self.config_stack, self.args)
        command = ' '.join(command)

        _, adb_arguments = command.split('ssh -s adb --')
        self.assertIn('-s test_serial', adb_arguments)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_without_serial(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values(device_serial='')

        command = commands.build_adt_run_test_command(
            self.config_stack, self.args)
        command = ' '.join(command)

        _, adb_arguments = command.split('ssh -s adb --')
        self.assertNotIn('-s test_serial', adb_arguments)
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_command_with_network_ready_creates_no_network(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        self.args.network_ready = True
        command = commands.build_adt_run_test_command(
            self.config_stack, self.args)

        self.assertFalse(check_any_item_in_list_startswith_value(
            command, 'nmcli device wifi connect'))
        mock_ssh_config_commands.assert_called_once_with()

    def test_adt_run_command_with_default_parameters_creates_network(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()

        command = commands.build_adt_run_test_command(
            self.config_stack, self.args)

        self.assertTrue(check_any_item_in_list_startswith_value(
            command, 'nmcli device wifi connect'))
        mock_ssh_config_commands.assert_called_once_with()

    def test_enabling_verbose_option(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        self.args.verbose = True
        command = commands.build_adt_run_test_command(
            self.config_stack, self.args)
        self.assertTrue(check_any_item_in_list_contains_value(command, '-ddd'))
        mock_ssh_config_commands.assert_called_once_with()

    def test_verbose_disabled_by_default(
            self, mock_ssh_config_commands, mock_get_dependencies):
        self.set_config_values()
        command = commands.build_adt_run_test_command(
            self.config_stack, self.args)
        self.assertFalse(
            check_any_item_in_list_contains_value(command, '-ddd'))
        mock_ssh_config_commands.assert_called_once_with()


@mock.patch('ubuntu_system_tests.host.commands._validate_ppas',
            return_value=[commands.Ppa(
                'ppa:user/name',
                'http://ppa.launchpad.net/owner/name/dist/dists')])
class SetupCommandTestCase(ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()
        self.args = DummyArgs(target='adb')

    def test_ppa_setup_commands_set_image_readwrite(
            self, mock_validate_ppas):
        self.set_config_values()
        self.assertIn(
            'mount -o remount,rw /',
            ' '.join(commands.build_adt_run_setup_command(
                self.config_stack, self.args))
        )

    def test_ppa_setup_commands_used_in_adt_setup(
            self, mock_validate_ppas):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, self.args)
        command = ' '.join(command)

        run_arguments, _ = command.split('---')
        ppa_cmd = 'apt-add-repository -y {}'.format(
            mock_validate_ppas.return_value[0].id)
        self.assertIn(ppa_cmd, run_arguments)

    def test_setup_commands_install_testability_libs(
            self, mock_validate_ppas):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, self.args)
        self.assertTrue(
            check_any_item_in_list_contains_value(
                command,
                'apt-get -y --no-install-recommends install '
                'qttestability-autopilot')
        )

    def test_setting_fs_rw_is_conditional_on_current_state(
            self, mock_validate_ppas):
        self.assertIn(
            commands._get_is_fs_ro_command(), commands._get_fs_rw_command())

    def test_setting_fs_ro_is_conditional_on_current_state(
            self, mock_validate_ppas):
        command = commands._get_fs_ro_setup_command()
        check_any_item_in_list_contains_value(
            command, commands._get_is_fs_rw_command())

    def test_enabling_verbose_option(self, mock_validate_ppas):
        self.set_config_values()
        self.args.verbose = True
        command = commands.build_adt_run_setup_command(
            self.config_stack, self.args)
        self.assertTrue(check_any_item_in_list_contains_value(command, '-ddd'))

    def test_verbose_disabled_by_default(self, mock_validate_ppas):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, self.args)
        self.assertFalse(
            check_any_item_in_list_contains_value(command, '-ddd'))

    @mock.patch('ubuntu_system_tests.host.debian.get_dependencies',
                return_value=['dep1'])
    def test_cache_deps_installs_apt_cacher(
            self, mock_get_dependencies, mock_validate_ppas):
        self.set_config_values()
        self.args.cache_deps = True
        command = commands.build_adt_run_setup_command(
            self.config_stack, self.args)
        self.assertTrue(
            check_any_item_in_list_contains_value(
                command,
                'apt-get -y --no-install-recommends install apt-cacher-ng'))
        mock_get_dependencies.assert_called_once_with(
            commands.TESTS_UST_DEFAULT)

    def test_default_no_apt_cacher_install(self, mock_validate_ppas):
        self.set_config_values()
        command = commands.build_adt_run_setup_command(
            self.config_stack, self.args)
        self.assertFalse(
            check_any_item_in_list_contains_value(command, 'apt-cacher-ng'))

    @mock.patch('ubuntu_system_tests.host.debian.get_dependencies')
    def test_install_tests_adds_extra_dependencies(
            self, mock_get_dependencies, mock_validate_ppas):
        self.set_config_values()
        self.args.install_tests = True
        mock_get_dependencies.return_value = ['test-dep-1', 'test-dep-2']
        command = commands.build_adt_run_setup_command(
            self.config_stack, self.args)
        self.assertTrue(
            check_any_item_in_list_contains_value(
                command, 'install test-dep-1 test-dep-2'))


class TestPPAArgsTestCase(ConfigBaseTestCase,
                          testscenarios.TestWithScenarios):

    scenarios = [
        ('single ppa', {
            'ppa_args': 'ppa:ci-train-ppa-service/landing-001',
            'expected_ppas': ['ppa:ci-train-ppa-service/landing-001']
        }),
        ('single silo', {
            'ppa_args': 'silo:landing-001',
            'expected_ppas': ['ppa:ci-train-ppa-service/landing-001']
        }),
        ('no ppa prefix', {
            'ppa_args': 'ci-train-ppa-service/landing-010',
            'expected_ppas': ['ppa:ci-train-ppa-service/landing-010']
        }),
        ('combined ppa silo and no prefix', {
            'ppa_args': (
                'ppa:ci-train-ppa-service/landing-001,'
                'silo:002,'
                'ci-train-ppa-service/003'),
            'expected_ppas': [
                'ppa:ci-train-ppa-service/landing-001',
                'ppa:ci-train-ppa-service/002',
                'ppa:ci-train-ppa-service/003'
            ]
        }),
        ('silo number format 100', {
            'ppa_args': 'silo:100',
            'expected_ppas': ['ppa:ci-train-ppa-service/100']
        }),
        ('ppa none', {
            'ppa_args': None,
            'expected_ppas': []
        }),
        ('ppa zero str', {
            'ppa_args': '',
            'expected_ppas': []
        }),
    ]

    def test_get_ppa_ids_for_args(self):
        fmt_ppas = commands._get_ppa_ids(self.ppa_args)
        self.assertEqual(fmt_ppas, self.expected_ppas)
