# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import json
import os
import testscenarios

from fixtures import TempDir
from os.path import join
from unittest import mock

from ubuntu_system_tests import selftests
from ubuntu_system_tests.host import debian
from ubuntu_system_tests.common import config


class GetDependenciesTestCase(testscenarios.TestWithScenarios):

    dep_map = {
        'app1': ['dep1'],
        'app1.tests.module1': ['dep1.1.1', 'dep1.1.2'],
        'app1.tests.module2': ['dep1.2.1', 'dep1.2.2'],
        'app2': ['dep2'],
        'app2.tests.module1': ['dep2.1.1', 'dep2.1.2'],
        'app2.tests.module2': ['dep2.2.1', 'dep2.2.2'],
    }

    scenarios = [
        ('single test', {
            'tests': ['app1.tests.module1.class1.test1'],
            'expected_deps': ['dep1', 'dep1.1.1', 'dep1.1.2']
        }),
        ('mutliple tests', {
            'tests': ['app1.tests.module1.class1.test1',
                      'app2.tests.module1.class1.test1'],
            'expected_deps': ['dep1', 'dep1.1.1', 'dep1.1.2',
                              'dep2', 'dep2.1.1', 'dep2.1.2']
        }),
        ('single module', {
            'tests': ['app1'],
            'expected_deps': ['dep1', 'dep1.1.1', 'dep1.1.2', 'dep1.2.1',
                              'dep1.2.2']
        }),
        ('multiple modules', {
            'tests': ['app1', 'app2'],
            'expected_deps': ['dep1', 'dep1.1.1', 'dep1.1.2', 'dep1.2.1',
                              'dep1.2.2', 'dep2', 'dep2.1.1', 'dep2.1.2',
                              'dep2.2.1', 'dep2.2.2']
        }),
        ('single test class', {
            'tests': ['app1.tests.module1.class1'],
            'expected_deps': ['dep1', 'dep1.1.1', 'dep1.1.2']
        }),
        ('mutliple test class', {
            'tests': ['app1.tests.module1.class1',
                      'app2.tests.module1.class1'],
            'expected_deps': ['dep1', 'dep1.1.1', 'dep1.1.2',
                              'dep2', 'dep2.1.1', 'dep2.1.2']
        }),
    ]

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    def test_get_dependencies(self):
        with mock.patch(
            'ubuntu_system_tests.host.debian._read_dependency_files'
        ) as mock_read_depends:
            mock_read_depends.return_value = self.dep_map
            depends = debian.get_dependencies(self.tests)
        self.assertCountEqual(depends, self.expected_deps)


class ReadDependenciesFilesTestCase(testscenarios.TestWithScenarios):

    deps_json1 = {
        'dependencies': [
            {
                'id': 'dep1',
                'version': '',
                'tmpinstall': 'false'
            }
        ],
        'tests': [
            {
                'ids': ['app1.tests'],
                'depends': ['dep1']
            }
        ]
    }
    deps_json2 = {
        'dependencies': [
            {
                'id': 'dep2',
                'version': '',
                'tmpinstall': 'false'
            }
        ],
        'tests': [
            {
                'ids': ['app2.tests'],
                'depends': ['dep2']
            }
        ]
    }

    scenarios = [
        ('single file', {
            'files': [
                {'name': 'deps1.json', 'content': deps_json1},
            ],
            'expected_dep': {'app1.tests': ['dep1,']},
        }),
        ('multiple file', {
            'files': [
                {'name': 'deps1.json', 'content': deps_json1},
                {'name': 'deps2.json', 'content': deps_json2},
            ],
            'expected_dep': {'app1.tests': ['dep1,'], 'app2.tests': ['dep2,']},
        })
    ]

    def test_read_files(self):
        with TempDir() as deps_dir:
            for f in self.files:
                deps_file = open(join(deps_dir.path, f['name']), 'w')
                deps_file.write(json.dumps(f['content']))
                deps_file.close()

            actual = debian._read_dependency_files(deps_dir.path, True, True)
            self.assertEqual(self.expected_dep, actual)


@mock.patch('ubuntu_system_tests.host.debian._read_dependency_files',
            return_value={'app1.tests': ['dep1']})
class GetDependenciesForTestIDTestCase(selftests.ConfigBaseTestCase):

    def test_valid_test_id(self, mock_read_depends):
        depends = debian.get_dependencies(['app1.tests'])
        self.assertCountEqual(depends, ['dep1'])

    def test_invalid_single_test_id_raises_exception(self, mock_read_depends):
        with self.assertRaises(RuntimeError):
            debian.get_dependencies(['app2.tests'])

    def test_invalid_mutli_test_id_raises_exception(self, mock_read_depends):
        with self.assertRaises(RuntimeError):
            debian.get_dependencies(['app1.tests', 'app2.tests'])


class ReadDependenciesDataTestCase(selftests.ConfigBaseTestCase):

    deps_correct = {
        'dependencies': [
            {
                'id': 'dep1',
                'version': '',
                'tmpinstall': 'false'
            }
        ],
        'tests': [
            {
                'ids': ['app1.tests'],
                'depends': ['dep1']
            }
        ]
    }
    deps_duplicate_id = {
        'dependencies': [
            {
                'id': 'dep1',
                'version': '1.0',
                'tmpinstall': 'false'
            },
            {
                'id': 'dep1',
                'version': '2.0',
                'tmpinstall': 'false'
            }
        ],
        'tests': [
            {
                'ids': ['app1.tests'],
                'depends': ['dep1']
            }
        ]
    }

    def test_read_files_no_duplicate_dep_id(self):
        with TempDir() as td, open(join(td.path, 'deps.json'), 'w') as deps:
            deps.write(json.dumps(self.deps_correct))
            deps.close()

            actual = debian._read_dependency_files(td.path, True, True)
            self.assertEqual({'app1.tests': ['dep1,']}, actual)

    def test_read_files_duplicated_dep_id(self):
        with TempDir() as td, open(join(td.path, 'deps.json'), 'w') as deps:
            deps.write(json.dumps(self.deps_duplicate_id))
            deps.close()

            with self.assertRaises(ValueError):
                debian._read_dependency_files(td.path, True, True)


class ReadCommonDependenciesFilesTestCase(testscenarios.TestWithScenarios):

    json_deps_no_common = {
        'dependencies': [
            {
                'id': 'dep1',
                'version': '',
                'tmpinstall': 'false'
            },
            {
                'id': 'dep2',
                'version': '',
                'tmpinstall': 'false'
            }
        ],
        'tests': [
            {
                'ids': ['app1.tests'],
                'depends': ['dep1']
            }
        ]
    }
    json_deps_common = {
        'dependencies': [
            {
                'id': 'dep1',
                'version': '',
                'tmpinstall': 'false'
            },
            {
                'id': 'dep2',
                'version': '',
                'tmpinstall': 'false'
            }
        ],
        'dependencies-common': [
            'dep2'
        ],
        'tests': [
            {
                'ids': ['app1.tests'],
                'depends': ['dep1']
            }
        ]
    }

    scenarios = [
        ('deps_no_common', {
            'files': [
                {'name': 'deps1.json', 'content': json_deps_no_common},
            ],
            'expected_dep': {'app1.tests': ['dep1,']},
        }),
        ('deps_common', {
            'files': [
                {'name': 'deps1.json', 'content': json_deps_common},
            ],
            'expected_dep': {'app1.tests': ['dep1,', 'dep2,']},
        })
    ]

    def test_read_files(self):
        with TempDir() as deps_dir:
            for f in self.files:
                deps_file = open(join(deps_dir.path, f['name']), 'w')
                deps_file.write(json.dumps(f['content']))
                deps_file.close()

            actual = debian._read_dependency_files(deps_dir.path, True, True)
            self.assertEqual(self.expected_dep, actual)


class ControlFileTestCase(selftests.ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    def test_create_temp_control(self):
        temp_control = debian.create_temp_control_file(
            tests=['test1', 'test2', 'test3'],
            dependencies=['dep1,', 'dep2,', 'dep3,'],
            restrictions=['res1', 'res2', 'res3']
        )
        with open(temp_control) as temp:
            tmp_control_content = temp.read()
            expected_content = """Tests: test1, test2, test3
Restrictions: res1, res2, res3
Depends: dep1,
         dep2,
         dep3,
"""
            self.assertEqual(tmp_control_content, expected_content)


class DependenciesForTestsTestCase(selftests.ConfigBaseTestCase):

    def setUp(self):
        super().setUp()
        self.config_stack = config.UbuntuSystemTestsConfig()

    def test_dependencies_for_test_module(self):
        dep_map = {'app1.tests': ['dep1,'], 'app2.tests': ['dep2,']}
        deps = debian._dependencies_for_test('app1.tests', dep_map)
        self.assertCountEqual(deps, ['dep1,'])

    def test_dependencies_for_test(self):
        dep_map = {'app1.tests': ['dep1,'], 'app2.tests': ['dep2,']}
        deps = debian._dependencies_for_test('app1.tests.test1', dep_map)
        self.assertCountEqual(deps, ['dep1,'])

    def test_dependencies_for_module(self):
        dep_map = {'apps.app1': ['dep1,'], 'apps.app2': ['dep2,']}
        deps = debian._dependencies_for_test('apps', dep_map)
        self.assertCountEqual(deps, ['dep1,', 'dep2,'])

    def test_dependencies_for_module_and_test(self):
        dep_map = {
            'apps': ['dep,'],
            'apps.app1': ['dep1,'],
            'apps.app2': ['dep2,']}
        deps = debian._dependencies_for_test('apps.app1', dep_map)
        self.assertCountEqual(deps, ['dep,', 'dep1,'])

    def test_dependencies_for_module_and_sub_test(self):
        dep_map = {
            'apps': ['dep,'],
            'apps.app1': ['dep1,'],
            'apps.app2': ['dep2,']}
        deps = debian._dependencies_for_test('apps.app1.test1', dep_map)
        self.assertCountEqual(deps, ['dep,', 'dep1,'])

    def test_dependency_control_format_no_version(self):
        dep = {'id': 'dep1', 'version': ''}
        dep_str = debian._get_dependency_str(dep, control_fmt=True)
        self.assertEqual(dep_str, 'dep1,')

    def test_dependency_control_format_with_version(self):
        dep = {'id': 'dep1', 'version': '1.2.3'}
        dep_str = debian._get_dependency_str(dep, control_fmt=True)
        self.assertEqual(dep_str, 'dep1 (>=1.2.3),')

    def test_dependency_cmdline_format_no_version(self):
        dep = {'id': 'dep1', 'version': ''}
        dep_str = debian._get_dependency_str(dep, control_fmt=False)
        self.assertEqual(dep_str, 'dep1 ')

    def test_dependency_cmdline_format_with_version(self):
        dep = {'id': 'dep1', 'version': '1.2.3'}
        dep_str = debian._get_dependency_str(dep, control_fmt=False)
        # version is not supported for cmdline format
        self.assertEqual(dep_str, 'dep1 ')

    def test_is_dependency_required_no_tmpinstall_no_include(self):
        dep = {'id': 'dep1'}
        self.assertTrue(debian.is_dependency_required(dep, tmp_install=False))

    def test_is_dependency_required_no_tmpinstall_include(self):
        dep = {'id': 'dep1'}
        self.assertTrue(debian.is_dependency_required(dep, tmp_install=True))

    def test_is_dependency_required_tmpinstall_no_include(self):
        dep = {'id': 'dep1', 'tmpinstall': 'true'}
        self.assertFalse(debian.is_dependency_required(dep, tmp_install=False))

    def test_is_dependency_required_tmpinstall_include(self):
        dep = {'id': 'dep1', 'tmpinstall': 'true'}
        self.assertTrue(debian.is_dependency_required(dep, tmp_install=True))

    @mock.patch('ubuntu_system_tests.host.debian._is_installed')
    def test_is_dependency_required_install_condition_true(
            self, mock_is_installed):
        mock_is_installed.return_value = True
        dep = {'id': 'dep1', 'condition': 'installed'}
        self.assertTrue(debian.is_dependency_required(dep, tmp_install=False))
        mock_is_installed.assert_called_once_with()

    @mock.patch('ubuntu_system_tests.host.debian._is_installed')
    def test_is_dependency_required_install_condition_false(
            self, mock_is_installed):
        mock_is_installed.return_value = False
        dep = {'id': 'dep1', 'condition': 'installed'}
        self.assertFalse(debian.is_dependency_required(dep, tmp_install=False))
        mock_is_installed.assert_called_once_with()

    @mock.patch('ubuntu_system_tests.host.debian._is_installed')
    def test_is_dependency_required_tmpinstall_overrules_condition(
            self, mock_is_installed):
        mock_is_installed.return_value = True
        dep = {'id': 'dep1', 'condition': 'installed', 'tmpinstall': 'true'}
        self.assertFalse(debian.is_dependency_required(dep, tmp_install=False))
        mock_is_installed.assert_not_called()


class ReadDependenciesJSONTestCase(unittest.TestCase):

    def test_dependencies_json_file_valid(self):
        deps_path = os.path.realpath('debian/tests/dependencies.json')
        with open(deps_path) as deps_file:
            json.loads(deps_file.read())
