# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import testscenarios

from unittest.mock import (
    call,
    mock_open,
    patch,
)

from ubuntu_system_tests import selftests
from ubuntu_system_tests.host import qemu
from ubuntu_system_tests.selftests import DummyArgs


class DummyPopenArgs(DummyArgs):

    def wait(self):
        pass


class DummyTempDir():

    def __init__(self):
        self.path = '/tmp/12345'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class QemuTestCase(selftests.ConfigBaseTestCase):

    popen_files_1 = DummyPopenArgs(stdout='/tmp/file_1\n/tmp/file_2')
    popen_files_2 = DummyPopenArgs(stdout='/tmp/file_3\n/tmp/file_4')

    @patch('subprocess.check_output')
    def test_get_architecture(self, mock_check_output):
        qemu.get_architecture()
        mock_check_output.assert_called_once_with(
            ['dpkg', '--print-architecture'])

    def test_get_qemu_pid_valid(self):
        with patch('builtins.open', mock_open(read_data='1234')) as o:
            pid = qemu.get_qemu_pid('/tmp/test_dir')
            o.assert_called_once_with('/tmp/test_dir/qemu.pid')
        self.assertEqual(pid, '1234')

    def test_get_qemu_pid_non_digit(self):
        with patch('builtins.open', mock_open(read_data='abcd')):
            with self.assertRaises(ValueError):
                qemu.get_qemu_pid('/tmp/test_dir')

    @patch('subprocess.check_call')
    @patch('ubuntu_system_tests.host.qemu.get_qemu_pid')
    def test_wait_for_qemu_stopped(self, mock_get_qemu_pid, mock_check_call):
        mock_get_qemu_pid.return_value = '1234'
        qemu.wait_for_qemu_stopped('/tmp/test_dir')
        mock_check_call.assert_called_once_with(['wait', '1234'], shell=True)

    @patch('subprocess.check_output')
    def test_get_kernel(self, mock_check_output):
        qemu.get_kernel('/tmp/test_dir')
        mock_check_output.assert_called_once_with(
            ['bsdtar', '-tf', '/tmp/test_dir', qemu.KERNEL_FILTER])

    @patch('subprocess.check_call')
    def test_extract_kernel(self, mock_check_call):
        qemu.extract_kernel('/tmp/test_iso', '/tmp/kernel', '/tmp/dst')
        mock_check_call.assert_called_once_with(
            ['bsdtar', 'xf', '/tmp/test_iso', '-C', '/tmp/dst',
             qemu.CASPER_INITRDLZ, '/tmp/kernel'])

    @patch('subprocess.check_call')
    def test_create_disk_image(self, mock_check_call):
        qemu.create_disk_image('/tmp/disk_img', '10G')
        mock_check_call.assert_called_once_with(
            ['qemu-img', 'create', '-f', 'qcow2', '/tmp/disk_img', '10G'])

    @patch('ubuntu_system_tests.host.qemu.get_preseed_content',
           return_value='abc')
    def test_create_preseed(self, mock_get_preseed_content):
        with patch('builtins.open', mock_open()) as o:
            qemu.create_preseed('/tmp/preseed.txt', '', '', '')
            expected_calls = [
                call('/tmp/preseed.txt', 'w+'),
                call().__enter__(),
                call().write('abc'),
                call().__exit__(None, None, None)
            ]
            o.assert_has_calls(expected_calls)

    @patch('shutil.copy2')
    @patch('os.chmod')
    @patch('os.listdir', return_value=['file_1', 'file_2'])
    @patch('ubuntu_system_tests.host.qemu.get_qemu_scripts_dir',
           return_value='/tmp/src_dir')
    def test_copy_qemu_scripts_to_path(
            self, mock_get_qemu_scripts_dir, mock_listdir, mock_chmod,
            mock_copy2):
        qemu.copy_qemu_scripts_to_path('/tmp/tgt_dir')
        expected_chmod = [
            call('/tmp/tgt_dir/file_1', 493),
            call('/tmp/tgt_dir/file_2', 493),
        ]
        expected_copy2 = [
            call('/tmp/src_dir/file_1', '/tmp/tgt_dir/file_1'),
            call('/tmp/src_dir/file_2', '/tmp/tgt_dir/file_2'),
        ]
        mock_chmod.assert_has_calls(expected_chmod)
        mock_copy2.assert_has_calls(expected_copy2)

    @patch('os.mkdir')
    @patch('os.chdir')
    @patch('subprocess.Popen', side_effect=[popen_files_1, popen_files_2])
    @patch('subprocess.check_call')
    @patch('ubuntu_system_tests.host.qemu.copy_qemu_scripts_to_path')
    @patch('ubuntu_system_tests.host.qemu.add_script_to_casper_setup')
    @patch('ubuntu_system_tests.host.qemu.create_preseed')
    def test_prepare_initrd(
            self, mock_create_preseed, mock_add_script, mock_copy_script,
            mock_check_call, mock_popen, mock_chdir, mock_mkdir):
        with patch('builtins.open', mock_open()) as o:
            initrd_path = qemu.prepare_initrd(
                '/tmp/work_dir', 'user1', 'pass1', 'disp1')
            self.assertEqual(
                '/tmp/work_dir/{}'.format(qemu.CASPER_INITRD), initrd_path)
            expected_check_calls = [
                call(['cpio', '--quiet', '-id'],
                     stdin='/tmp/file_1\n/tmp/file_2'),
                call(['cpio', '--quiet', '-o', '-H', 'newc'],
                     stdin='/tmp/file_3\n/tmp/file_4', stdout=o())
            ]
            mock_check_call.assert_has_calls(expected_check_calls)

            expected_popen_calls = [
                call(['xzcat', '/tmp/work_dir/casper/initrd.lz'], stdout=-1),
                call(['find', '.'], stdout=-1)
            ]
            mock_popen.assert_has_calls(expected_popen_calls)

            mock_copy_script.assert_called_once_with('/tmp/work_dir/initrd')
            mock_add_script.assert_called_once_with(
                '/tmp/work_dir/initrd',
                '/tmp/work_dir/initrd/qemu-casper-provision')
            mock_create_preseed.assert_called_once_with(
                '/tmp/work_dir/initrd/preseed.cfg', 'user1', 'pass1', 'disp1')

    @patch('shutil.copy2')
    def test_add_script_to_casper_setup(self, mock_copy2):
        with patch('builtins.open', mock_open(read_data='abcd')) as o:
            qemu.add_script_to_casper_setup('/tmp/initrd', '/tmp/test_script')
            expected_file_calls = [
                call('/tmp/initrd/scripts/casper-bottom/ORDER', 'r+'),
                call().__enter__(),
                call().read(),
                call().seek(0),
                call().write(
                    '/scripts/casper-bottom/test_script "$@"\n[ -e '
                    '/conf/param.conf ] && . /conf/param.conf\nabcd'),
                call().__exit__(None, None, None),
            ]
            o.assert_has_calls(expected_file_calls)

    @patch('ubuntu_system_tests.host.qemu.get_qemu_path',
           return_value='/tmp/qemu')
    @patch('builtins.print')
    @patch('subprocess.check_call')
    def test_launch_qemu_install(self, mock_check_call, mock_print,
                                 mock_get_qemu_path):
        qemu.launch_qemu(
            '/tmp/work_dir', '/tmp/disk.img', 4096, 3, True, None,
            '/tmp/disk.iso', '/tmp/kernel', 'kenel_opts')
        expected_args = [
            '/tmp/qemu', '-m', '4096', '-smp', '3', '-localtime', '-net',
            'nic,model=virtio', '-net', 'user', '-cpu', 'core2duo',
            '-enable-kvm', '-drive', 'file=/tmp/disk.img,if=virtio',
            '-display', 'none', '-cdrom', '/tmp/disk.iso', '-boot', 'd',
            '-initrd', '/tmp/work_dir/casper/initrd', '-kernel', '/tmp/kernel',
            '-append', 'kenel_opts', '-pidfile', '/tmp/work_dir/qemu.pid',
            '-boot', 'menu=off', '-no-reboot',
        ]
        mock_check_call.assert_called_once_with(expected_args)

    @patch('ubuntu_system_tests.host.qemu.get_qemu_path',
           return_value='/tmp/qemu')
    @patch('builtins.print')
    @patch('subprocess.check_call')
    def test_launch_qemu(self, mock_check_call, mock_print,
                         mock_get_qemu_path):
        qemu.launch_qemu(
            '/tmp/work_dir', '/tmp/disk.img', 4096, 3, True, None)
        expected_args = [
            '/tmp/qemu', '-m', '4096', '-smp', '3', '-localtime', '-net',
            'nic,model=virtio', '-net', 'user', '-cpu', 'core2duo',
            '-enable-kvm', '-drive', 'file=/tmp/disk.img,if=virtio',
            '-display', 'none',
        ]
        mock_check_call.assert_called_once_with(expected_args)

    @patch('ubuntu_system_tests.host.qemu.TempDir',
           new=DummyTempDir)
    @patch('ubuntu_system_tests.host.qemu.create_disk_image')
    @patch('ubuntu_system_tests.host.qemu.get_kernel',
           return_value='/tmp/kernel')
    @patch('ubuntu_system_tests.host.qemu.extract_kernel')
    @patch('ubuntu_system_tests.host.qemu.prepare_initrd')
    @patch('ubuntu_system_tests.host.qemu.wait_for_qemu_stopped')
    @patch('ubuntu_system_tests.host.qemu.get_qemu_path',
           return_value='/tmp/qemu')
    @patch('subprocess.check_call')
    def test_build_qemu(
            self, mock_check_call, mock_get_qemu_path, mock_wait,
            mock_prepare_initrd, mock_extract_kernel, mock_get_kernel,
            mock_create_disk_image):
        qemu.build_qemu(
            '/tmp/disk.img', '/tmp/disk.iso', 3, 4096, '8G',
            'user1', 'pass1', 'User 1', False, False)
        mock_create_disk_image.assert_called_once_with('/tmp/disk.img', '8G')
        mock_get_kernel.assert_called_once_with('/tmp/disk.iso')
        mock_extract_kernel.assert_called_once_with(
            '/tmp/disk.iso', '/tmp/kernel', '/tmp/12345')
        mock_prepare_initrd.assert_called_once_with(
            '/tmp/12345', 'user1', 'pass1', 'User 1')
        mock_wait.assert_called_once_with('/tmp/12345')
        launch_qemu_expected_calls = [
            call(['/tmp/qemu', '-m', '4096', '-smp', '3', '-localtime', '-net',
                  'nic,model=virtio', '-net', 'user', '-cpu', 'core2duo',
                  '-enable-kvm', '-drive', 'file=/tmp/disk.img,if=virtio',
                  '-display', 'sdl', '-vga', 'qxl', '-soundhw', 'all',
                  '-serial', 'file:/tmp/12345/qemu.ttyS0', '-serial',
                  'file:/tmp/12345/qemu.ttyS1', '-cdrom', '/tmp/disk.iso',
                  '-boot', 'd', '-initrd', '/tmp/12345/casper/initrd',
                  '-kernel', '/tmp/kernel', '-append', 'boot=casper '
                  'DEBCONF_DEBUG=developer -- debconf/priority=critical '
                  'locale=en_US console-setup/ask_detect=false '
                  'console-setup/layoutcode=us noprompt automatic-ubiquity',
                  '-pidfile', '/tmp/12345/qemu.pid', '-boot', 'menu=off',
                  '-no-reboot']),
            call(['/tmp/qemu', '-m', '4096', '-smp', '3', '-localtime', '-net',
                  'nic,model=virtio', '-net', 'user', '-cpu', 'core2duo',
                  '-enable-kvm', '-drive', 'file=/tmp/disk.img,if=virtio',
                  '-display', 'sdl', '-vga', 'qxl', '-soundhw', 'all',
                  '-serial', 'file:/tmp/12345/qemu.ttyS0', '-serial',
                  'file:/tmp/12345/qemu.ttyS1']),
        ]
        mock_check_call.assert_has_calls(launch_qemu_expected_calls)


class QemuPreseedContentTestCase(selftests.ConfigBaseTestCase,
                                 testscenarios.TestWithScenarios):

    scenarios = [
        ('User 1', {
            'username': 'testuser_1',
            'password': 'userpass_1',
            'display_name': 'Test User 1',
        }),
        ('User 2', {
            'username': 'testuser_2',
            'password': 'userpass_2',
            'display_name': 'Test User 2',
        }),
    ]

    def test_get_preseed_content(self):
        content = qemu.get_preseed_content(
            self.username, self.password, self.display_name)
        self.assertGreaterEqual(
            content.find(
                'd-i passwd/username string {}'.format(self.username)),
            0)
        self.assertGreaterEqual(
            content.find(
                'd-i passwd/user-password password {}'.format(self.password)),
            0)
        self.assertGreaterEqual(
            content.find(
                'd-i passwd/user-password-again password {}'.format(
                    self.password)),
            0)
        self.assertGreaterEqual(
            content.find(
                'd-i passwd/user-fullname string {}'.format(
                    self.display_name)),
            0)


class QemuPathTestCase(selftests.ConfigBaseTestCase,
                       testscenarios.TestWithScenarios):

    scenarios = [
        ('amd64', {
            'arch': qemu.ARCH_AMD64,
            'name': qemu.QEMU_SYSTEM_AMD64,
        }),
        ('i386', {
            'arch': qemu.ARCH_I386,
            'name': qemu.QEMU_SYSTEM_I386,
        }),
    ]

    @patch('subprocess.check_output')
    @patch('ubuntu_system_tests.host.qemu.get_architecture')
    def test_get_qemu_path(self, mock_get_architecture,
                           mock_check_output):
        mock_get_architecture.return_value = self.arch
        qemu.get_qemu_path()
        mock_check_output.assert_called_once_with(['which', self.name])


class DisplayOptionsTestCase(selftests.ConfigBaseTestCase,
                             testscenarios.TestWithScenarios):

    scenarios = [
        ('headless', {
            'headless': True,
            'expected_options': '-display none ',
        }),
        ('display', {
            'headless': False,
            'expected_options': '-display sdl -vga qxl -soundhw all ',
        }),
    ]

    def test_get_display_args(self):
        options = qemu.get_display_args(self.headless)
        self.assertEqual(options, self.expected_options)
