# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014, 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import shutil
import tempfile
import unittest
from argparse import Namespace
from unittest import mock

from ubuntu_system_tests.common import config

DEFAULT_SECTION = '[{}]\n'.format(config.KEY_DEFAULT)
ISOLATED_ENVIRON = {'HOME': None}


class DummyArgs:

    def __init__(self, **kwargs):
        for arg in kwargs.keys():
            setattr(self, arg, kwargs[arg])

    def __getattr__(self, name):
        return None


class ConfigBaseTestCase(unittest.TestCase):

    def setUp(self):
        super().setUp()

        # Patch the atexit.register method so nothing is executed after the
        # test exits. As we are using a tmp directory that will be removed
        # on test cleanUp, then saving the config file to that unexisting
        # location at exit would fail.
        patcher = mock.patch('atexit.register')
        patcher.start()
        self.addCleanup(patcher.stop)

        set_uniq_cwd(self)
        isolate_from_env(
            self, dict(HOME=self.uniq_dir, XDG_CONFIG_HOME=None))

    def set_config_values(self, **kwargs):
        config_string = DEFAULT_SECTION
        for option, value in kwargs.items():
            config_string += '{} = {}\n'.format(option, value)
        # Fill with dummy values the rest of options.
        for option in config.CONFIG_OPTIONS:
            if option.name not in kwargs:
                config_string += '{} = dummy\n'.format(option.name)

        self.config_stack.get_config_from_string(config_string)
        self.config_stack.save()


def check_any_item_in_list_contains_value(list, value):
    return any(item.find(value) >= 0 for item in list)


def check_any_item_in_list_startswith_value(list, value):
    return any(item.startswith(value) for item in list)


def get_cmdline_list_params(
        command='list', config=config.DEFAULT_CONF,
        config_section=config.KEY_DEFAULT, silent=False, suites=None,
        filters=None):
    suites = suites or []
    return Namespace(
        command=command, config=config, config_section=config_section,
        silent=silent, suites=suites, filters=filters)


def get_cmdline_run_params(
        command='run', config=config.DEFAULT_CONF,
        config_section=config.KEY_DEFAULT, network_ready=False, verbose=False,
        silent=False, target=None, headless=False, img=None, suites=None,
        touch_visualization=False, filters=None):
    suites = suites or []
    return Namespace(
        command=command, config=config, config_section=config_section,
        network_ready=network_ready, verbose=verbose, silent=silent,
        suites=suites, touch_visualization=touch_visualization,
        filters=filters, target=target, headless=headless, img=img)


def get_cmdline_setup_params(
        command='setup', config=config.DEFAULT_CONF,
        config_section=config.KEY_DEFAULT, network_ready=False, verbose=False,
        cache_deps=False, dist_upgrade=False, install_tests=False,
        update_apt=False, silent=False, target=None, headless=False, img=None,
        suites=None, filters=None, ppa=None):
    suites = suites or []
    return Namespace(
        command=command, config=config, config_section=config_section,
        network_ready=network_ready, verbose=verbose, cache_deps=cache_deps,
        dist_upgrade=dist_upgrade, install_tests=install_tests,
        update_apt=update_apt, silent=silent, suites=suites, filters=filters,
        ppa=ppa, target=target, headless=headless, img=img)


def set_uniq_cwd(test, prefix=None):
    """Create a temporary directory an cd into it for the test duration.

    :param test: The test to isolate in a temporary directory.

    :param prefix: A prefix to use to name the created temporary directory. It
       defaults to the test method name which help debug leaks under '/tmp'.

    This is generally called during a test setup. It requires a test providing
    the 'addCleanup' method and will set the 'uniq_dir' attribute.

    Note that it doesn't prohibit the test to write anywhere it has access to,
    but it changes its current directory so that relative paths end in a
    temporary directory. In practice, this means using simple strings paths
    instead of os.path.join'ing them which is less readable and clutter the
    test.

    This can also be used to create a home directory (as long as $HOME is also
    set appropriately) when the test needs to access '~/' files for example.

    """
    if prefix is None:
        prefix = '{}-'.format(test._testMethodName)
    test.uniq_dir = tempfile.mkdtemp(prefix=prefix, suffix='.tmp')
    test.addCleanup(shutil.rmtree, test.uniq_dir, True)
    current_dir = os.getcwd()
    test.addCleanup(os.chdir, current_dir)
    os.chdir(test.uniq_dir)


def override_env_variable(name, value):
    """Modify the environment, setting or removing the env_variable.

    :param name: The environment variable to set.

    :param value: The value to set the environment to. If None, then
        the variable will be removed.

    :return: The original value of the environment variable.
    """
    orig = os.environ.get(name)
    if value is None:
        if orig is not None:
            del os.environ[name]
    else:
        os.environ[name] = value
    return orig


def override_env(test, name, new):
    """Set an environment variable, and reset it after the test.

    :param name: The environment variable name.

    :param new: The value to set the variable to. If None, the variable is
        deleted from the environment.

    :returns: The actual variable value.
    """
    value = override_env_variable(name, new)
    test.addCleanup(override_env_variable, name, value)
    return value


def isolate_from_env(test, env=None):
    """Isolate test from the environment variables.

    This is usually called in setUp for tests that needs to modify the
    environment variables and restore them after the test is run.

    :param test: A test instance

    :param env: A dict containing variable definitions to be installed. Only
        the variables present there are protected. They are initialized with
        the provided values.
    """
    if env is None:
        env = ISOLATED_ENVIRON
    for name, value in env.items():
        override_env(test, name, value)
