# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014-2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os

from tempfile import mkstemp
from autopilot import (
    exceptions,
    input,
    testcase
)
from autopilot import platform
from autopilot.introspection import _object_registry as object_registry
from retrying import retry

from ubuntu_system_tests.helpers import (
    context,
    images,
    mir
)
from ubuntu_system_tests.helpers import screen
from ubuntu_system_tests.helpers import wait_until
from ubuntu_system_tests.helpers.autopilot import get_proxy_object
from ubuntu_system_tests.helpers.init_systems import get_init_system
from ubuntu_system_tests.helpers.job_fixtures import JobEnvironmentVariable
from ubuntu_system_tests.helpers.location.fixture_setup import (
    SetLocationAccessRequests)
from ubuntu_system_tests.helpers.unity8 import Unity8Job
from ubuntu_system_tests.helpers.timeout import timeout
from ubuntu_system_tests.helpers.unity8.fixture_setup import SetupUnity8
from ubuntu_system_tests.helpers.webapp import (
    DEFAULT_WEBVIEW_INSPECTOR_IP,
    DEFAULT_WEBVIEW_INSPECTOR_PORT,
)

from ubuntu_system_tests.tests.data.definitions import DESKTOP_PLATFORM_NAME

# Any test name ending with this will leave device locked
TEST_ID_DEVICE_LOCKED = 'device_locked'

logger = logging.getLogger(__name__)

unity8 = Unity8Job()


PLATFORMS_DEFAULT_SETUP = [DESKTOP_PLATFORM_NAME]
DEFAULT_FIXTURES = [SetLocationAccessRequests('unity8-dash', True),
                    JobEnvironmentVariable(QT_LOAD_TESTABILITY='1')]


class BaseUbuntuSystemTestCase(testcase.AutopilotTestCase):

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        unity8.ensure_unity_running_and_greeter_hidden()

    def setUp(self, custom_fixtures=[SetupUnity8()]):
        super().setUp()
        self._reset_autopilot_registry()
        self._init_shared_context()
        DEFAULT_FIXTURES.extend(custom_fixtures)
        for fixture in DEFAULT_FIXTURES:
            self.useFixture(fixture)
        if platform.model() not in PLATFORMS_DEFAULT_SETUP:
            # Workaround to remove the UInput keyboard created by autopilot.
            # This stops the OSK from being displayed, so must remove it.
            self._setup_keyboard()
        if self.device_unlock_required():
            from ubuntu_system_tests.helpers.unity8.shell import Unity8  # NOQA
            context.shared.unity_pid = get_init_system().get_pid(Unity8Job())
            context.shared.unity = get_proxy_object(
                pid=context.shared.unity_pid)
            context.shared.unity.unlock()

    def _setup_keyboard(self):
        self._remove_autopilot_keyboard()
        self._reset_keyboard()
        self._enable_osk_for_uitk()

    @property
    def shell(self):
        return context.shared.unity.main_window

    @staticmethod
    def _init_shared_context():
        # The shared context is reset for each test
        context.shared.clear()
        context.shared.timers = []
        context.shared.retry_time = 10000

    def _enable_osk_for_uitk(self):
        """Enable use of the OSK with uitk helpers."""
        os.environ['UITK_USE_MALIIT'] = '1'

    def _remove_autopilot_keyboard(self):
        """Remove the UInput simulated keyboard automatically created by
        autopilot which stops the OSK from appearing."""
        if self._kb:
            if not self._kb._device._device.device:
                self.assertTrue(
                    wait_until(self._find_keyboard),
                    'Could not find the UInput keyboard created by autopilot')
            try:
                self._kb._device._device.close()
            except OSError:
                pass
            self._kb = None

    def _find_keyboard(self):
        """Find the autopilot keyboard if it is available and save instace
        in the self._kb member variable.
        :return: True if keyboard found, False otherwise."""
        kb = self._kb._device._device._find_device()
        if kb:
            self._kb._device._device.device = kb
        return kb

    def _reset_autopilot_registry(self):
        """Reset autopilot's registry between tests to avoid CPO errors."""
        object_registry.patch_registry({})

    # Retry added until fix for bug 1511895 in ubuntu-keyboard is landed
    @retry(stop_max_attempt_number=3,
           retry_on_exception=lambda exception: (
               isinstance(exception, exceptions.ProcessSearchError) or
               isinstance(exception, exceptions.BackendException) or
               isinstance(exception, TimeoutError)))
    @timeout(30)
    def _reset_keyboard(self):
        """Reset the Autopilot connection to the OSK"""
        input.Keyboard.create('OSK')._keyboard.reset()

    def set_touch_visualization(self):
        """Set touch visualization as required and disable on cleanup."""
        enable = os.getenv('TOUCH_VISUALIZATION') == 'true'
        screen.set_visualization(enable)
        if enable:
            self.addCleanup(screen.disable_touch_visualization)

    def device_unlock_required(self):
        """Return True if test requires device unlocking, False otherwise."""
        test_id = self.id().split('.')[-1]
        return not test_id.endswith(TEST_ID_DEVICE_LOCKED)

    def _get_screenshot_size(self, x, y, width, height):
        """ Calculate the screenshot size """
        unity = context.shared.unity
        if width and height:
            return width, height
        else:
            return unity.get_width(), unity.get_height()

    def get_screenshot_image(self, x=0, y=0, width=0, height=0):
        """ Take an screenshot and return a PIL Image file. The screenshot
        will be automatically deleted after the test execution. Passing the
        parameters for the rectangle coordinates, the screenshot captured will
        be just for the section selected
        :return: A PIL Image file which contains the screenshot
        """
        file_path = mkstemp()[1]
        self.addCleanup(os.remove, file_path)
        mir.get_rgba_screenshot(file_path, x, y, width, height)

        return images.get_image_from_rgba(
            file_path, self._get_screenshot_size(x, y, width, height))

    def setup_webapp_environment(self):
        """ Setup the environment for the webapps. This method has to be
        called before launch the first web app in the test.
        """
        self.useFixture(JobEnvironmentVariable(
            global_=True,
            UBUNTU_WEBVIEW_DEVTOOLS_HOST=DEFAULT_WEBVIEW_INSPECTOR_IP,
            UBUNTU_WEBVIEW_DEVTOOLS_PORT=str(DEFAULT_WEBVIEW_INSPECTOR_PORT)))
