# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2014, 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Steps
# --------
# 1. Navigate to the Apps Scope
# * apps scope is displayed
# 2. Scroll to the bottom of the scope
# * Ubuntu apps store is displayed
# 3. Tap on The ubuntu store button
# * Ubuntu apps store opens
# 4. Tap on an uninstalled app
# * Main view for app appears
# 5. Tap on install
# * U1 login page appears
# 6. Login to u1
# * Login window goes app starts to download
# 7. Tap on Open
# * Application opens

import logging

from autopilot.matchers import Eventually
from testtools.matchers import Equals
from ubuntu_system_tests.helpers import (
    click,
    context,
    google_maps,
    processes,
    ubuntuone_credentials as uone_cred
)
from ubuntu_system_tests.helpers.uid import b64_uuid
from ubuntu_system_tests.helpers.scopes import go_to_apps_scope
from ubuntu_system_tests.helpers.scopes.apps import APP_STORE_TITLE
from ubuntu_system_tests.tests import base

logger = logging.getLogger(__name__)


class AppStoreTestCase(base.BaseUbuntuSystemTestCase):

    def setUp(self):
        super().setUp()
        existing_account_ids = uone_cred.get_existing_account_ids()
        self.addCleanup(
            uone_cred.remove_accounts, ids_to_keep=existing_account_ids)

        uone_cred.restart_online_services_with_testability()
        self.dash = context.shared.dash

    def test_go_to_dash_from_app_store(self):
        """
        1. From the dash, open the Apps scope and go to the Apps store.
        2. Swipe from left edge to reveal launcher.
        3. Press dash icon on launcher to return to dash.
        """
        dash_title = self.dash.get_current_page_header().title
        go_to_apps_scope().go_to_store()
        header = self.dash.get_current_page_header(wait_for_dash_loaded=False)
        self.assertEqual(header.title, APP_STORE_TITLE)
        self.shell.show_dash_from_launcher()
        header = self.dash.get_current_page_header()
        self.assertEqual(header.title, dash_title)

    def test_install_free_app(self):
        """Create U1 account and install application from store."""
        click.ensure_package_uninstalled(google_maps.PACKAGE_NAME)
        store_scope = go_to_apps_scope().go_to_store()
        store_scope.enter_search_query(google_maps.APP_TITLE)

        preview = store_scope.open_preview(
            'appstore', google_maps.APP_TITLE,
            subtitle=google_maps.APP_PUBLISHER)
        preview.install()
        self.addCleanup(
            click.ensure_package_uninstalled, google_maps.PACKAGE_NAME)

        self.log_into_u1_account()

        # Wait for the application to start installing.
        self.dash.wait_for_processing_to_complete()
        preview.wait_until_package_installed()
        app_id = click.get_click_app_identifier(google_maps.PACKAGE_NAME)
        preview.open()
        self.addCleanup(processes.ensure_application_closed, app_id)

        self.assertThat(
            self.shell.get_current_focused_app_id, Eventually(
                Equals(google_maps.APP_ID)))

    def log_into_u1_account(self):
        """Log into u1 signon-u1 dialog that pops up."""
        account_id = b64_uuid()
        email_address = 'u1test+{}@canonical.com'.format(account_id)
        user_name = 'Ubuntu Tester {}'.format(account_id)
        password = 'Hola123*'

        proxy = uone_cred.get_ubuntuone_credentials_proxy()
        self.assertThat(
            self.unity_proxy.get_number_of_surfaces, Eventually(Equals(2)))
        account_form = proxy.accounts_creation_page
        account_form.register_new_account(
            email_address, user_name, password, password, True)

        self.assertThat(
            self.unity_proxy.get_number_of_surfaces,
            Eventually(Equals(1), timeout=30))
