# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-

#
# Ubuntu System Tests
# Copyright (C) 2016 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from unittest import skipUnless

from autopilot import platform
from autopilot.matchers import Eventually
from testtools.matchers import Equals, LessThan, GreaterThan

from ubuntu_system_tests.helpers.brightness.sensors_utils import (
    RepowerdSensors)
from ubuntu_system_tests.helpers.indicators.battery import (
    BatteryIndicatorPage)

from ubuntu_system_tests.helpers import system_settings
from ubuntu_system_tests.helpers.indicators.indicators import IndicatorsFactory
from ubuntu_system_tests.helpers import display
from ubuntu_system_tests.helpers.power import fixture_setup as power_fixture
from ubuntu_system_tests.helpers.power.fixture_setup import (
    EnableRepowerdFakeSensors)
from ubuntu_system_tests.helpers.unity8.fixture_setup import SetupUnity8
from ubuntu_system_tests.tests import base

DEVICE_ARALE = 'arale'
DEVICE_KRILLIN = 'krillin'
DEVICE_MAKO = 'aosp_mako'
SUPPORTED_DEVICES = [DEVICE_ARALE, DEVICE_KRILLIN, DEVICE_MAKO]


class BrightnessTestCase(base.BaseUbuntuSystemTestCase):

    def setUp(self, custom_fixtures=[]):
        if not custom_fixtures:
            custom_fixtures.append(SetupUnity8())
        super().setUp(custom_fixtures=custom_fixtures)
        max_brightness = display.get_max_brightness_level()

        self.useFixture(
            power_fixture.BrightnessEnvironmentFixture(brightness_value=1))

        self.assertEquals(
            display.get_stable_brightness_value(), max_brightness,
            'Brightness Fixture failed to set the brightness indicator to '
            'its maximum level')

    @staticmethod
    def open_brightness_from_system_settings():
        settings = system_settings.launch_system_settings()
        return settings.go_to_brightness_page()

    def open_brightness_from_battery_indicator(self):
        indicator_page = IndicatorsFactory.get_indicators_page(
            BatteryIndicatorPage.name)
        self.addCleanup(self.shell.close_indicator_page)
        return indicator_page


class AutoBrightnessTestCase(BrightnessTestCase):

    sensors = RepowerdSensors()

    @skipUnless(sensors.is_brightness_sensor_present(),
                'Brightness sensor support required.')
    def setUp(self):
        super().setUp(custom_fixtures=[EnableRepowerdFakeSensors(),
                                       SetupUnity8()])

    def _test_automatic_screen_brightness(self, settings_page):
        """
        Test the automatic screen brightness setting on the specified
        settings page.
        1. Enable automatic screen brightness setting
        2. Set light conditions dark and ensure screen dims
        3. Set light conditions bright and ensure screen brightens
        4. Disable automatic screen brightness setting
        5. Set light conditions dark and ensure brightness doesn't change

        :param settings_page: The page that contains auto-brightness settings.
        """
        self.assertFalse(settings_page.is_auto_adjust_enabled())
        brightness_default = display.get_stable_brightness_value()

        settings_page.toggle_auto_adjust()
        self.assertTrue(settings_page.is_auto_adjust_enabled())
        self.sensors.set_light_level_dark()
        brightness_auto_dark = display.get_stable_brightness_value()
        self.assertLess(brightness_auto_dark, brightness_default)

        self.sensors.set_light_level_bright()
        brightness_auto_bright = display.get_stable_brightness_value()
        self.assertGreater(brightness_auto_bright, brightness_auto_dark)

        settings_page.toggle_auto_adjust()
        self.assertFalse(settings_page.is_auto_adjust_enabled())
        brightness_manual_bright = display.get_stable_brightness_value()
        self.sensors.set_light_level_dark()
        brightness_manual_dark = display.get_stable_brightness_value()
        self.assertEqual(brightness_manual_bright, brightness_manual_dark)

    def test_battery_indicator_automatic_screen_brightness(self):
        """
        1. Drag down the battery indicator
        2. Enable automatic screen brightness setting
            3. Set light conditions dark and ensure screen dims
        4. Set light conditions bright and ensure screen brightens
        5. Disable automatic screen brighness setting
        6. Set light conditions dark and ensure brightness doesn't change
        """
        self._test_automatic_screen_brightness(
            self.open_brightness_from_battery_indicator())

    def test_system_settings_automatic_screen_brightness(self):
        """
        1. Open system settings and go to brightness page
        2. Enable automatic screen brightness setting
        3. Set light conditions dark and ensure screen dims
        4. Set light conditions bright and ensure screen brightens
        5. Disable automatic screen brighness setting
        6. Set light conditions dark and ensure brightness doesn't change
        """
        self._test_automatic_screen_brightness(
            self.open_brightness_from_system_settings())


class ManualBrightnessTestCase(BrightnessTestCase):

    def _test_brightness_slider(self, brightness_page):
        max_brightness = display.get_max_brightness_level()
        min_brightness = display.get_min_brightness_level()

        slider = brightness_page.get_brightness_slider_widget()
        while brightness_page.get_current_brightness() != 0.0:
            brightness_slider = brightness_page.get_current_brightness()
            brightness_backend = display.get_current_brightness()
            slider.slide_one_step_left()

            self.assertThat(
                lambda: brightness_page.get_current_brightness(),
                Eventually(LessThan(brightness_slider)))
            self.assertThat(
                lambda: display.get_current_brightness(),
                Eventually(LessThan(brightness_backend)))

        self.assertThat(lambda: display.get_current_brightness(),
                        Eventually(Equals(min_brightness)))

        while brightness_page.get_current_brightness() != \
                brightness_page.get_max_brightness():
            brightness_slider = brightness_page.get_current_brightness()
            brightness_backend = display.get_current_brightness()
            slider.slide_one_step_right()
            self.assertThat(
                lambda: brightness_page.get_current_brightness(),
                Eventually(GreaterThan(brightness_slider)))
            self.assertThat(
                lambda: display.get_current_brightness(),
                Eventually(GreaterThan(brightness_backend)))

        self.assertThat(lambda: display.get_current_brightness(),
                        Eventually(Equals(max_brightness)))

    @skipUnless(platform.image_codename() in SUPPORTED_DEVICES,
                'Supported devices: {}'.format(', '.join(SUPPORTED_DEVICES)))
    def test_brightness_slider_from_the_battery_indicator(self):
        """
        Prerequisite: ensure auto brightness is disabled and screen brightness
                      is maximum
        1. Drag down the battery indicator
        2. Drag the brightness handle to the left till the end
        3. Drag the brightness handle to the right till the end
        """
        brightness_page = self.open_brightness_from_battery_indicator()
        self._test_brightness_slider(brightness_page)

    def test_brightness_slider_the_system_settings(self):
        """
        Prerequisite: ensure auto brightness is disabled and screen brightness
                      is maximum
        1. Open the brightness page from the system settings page
        2. Drag the brightness handle to the left till the end
        3. Drag the brightness handle to the right till the end
        """
        brightness_page = self.open_brightness_from_system_settings()
        self._test_brightness_slider(brightness_page)
